/*
 *
 *      This code is loosely based on the Linux serial driver, written by
 *      Linus Torvalds, Theodore T'so and others.
 *
 *      This program is free software; you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 2 of the License, or
 *      (at your option) any later version.
 *
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *
 *      You should have received a copy of the GNU General Public License
 *      along with this program; if not, write to the Free Software
 *      Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 *
 */

#ifdef MODVERSIONS
#ifndef MODULE
#define MODULE
#endif
#endif

#define ADV_BAUDBASE	921600

#include <linux/version.h>
#define VERSION_CODE(ver,rel,seq) ((ver << 16) | (rel << 8) | seq)
#ifdef MODULE
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,18))
#include <linux/config.h>
#endif
#ifdef MODVERSIONS
#include <linux/modversions.h>
#endif
#include <linux/module.h>
#else
#define	MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif
#include <linux/init.h>
//#include <linux/autoconf.h>
#include <linux/errno.h>
#include <linux/signal.h>
#include <linux/sched.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/tty.h>
#include <linux/tty_flip.h>
#include <linux/serial.h>
#include <linux/serial_reg.h>
#include <linux/major.h>
#include <linux/string.h>
#include <linux/fcntl.h>
#include <linux/ptrace.h>
#include <linux/ioport.h>
#include <linux/mm.h>
//#include <linux/smp_lock.h>
#include <linux/proc_fs.h>

#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
#include <linux/kthread.h>
#endif

#include <linux/netlink.h>
#include "sock_k.h"
#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))

#define copy_from_user memcpy_fromfs
#define copy_to_user memcpy_tofs
#define put_to_user(arg1, arg2) put_fs_long(arg1, (unsigned long *)arg2)
#define get_from_user(arg1, arg2) arg1 = get_fs_long((unsigned long *)arg2)
#define schedule_timeout(x) {current->timeout = jiffies + (x); schedule();}
#define signal_pending(x) ((x)->signal & ~(x)->blocked)
#else
#include <asm/uaccess.h>
#include <linux/poll.h>
#define put_to_user(arg1, arg2) put_user(arg1, arg2)
#define get_from_user(arg1, arg2) get_user(arg1, arg2)
#endif

#include "advtty.h"

#define	ADVTTY_EVENT_TXLOW          1
#define	ADVTTY_EVENT_HANGUP         2

#define SERIAL_DO_RESTART

#define SERIAL_TYPE_NORMAL          1
#define SERIAL_TYPE_CALLOUT         2

#define WAKEUP_CHARS                256

#ifndef MAX_SCHEDULE_TIMEOUT
#define	MAX_SCHEDULE_TIMEOUT        ((long)(~0UL>>1))
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
#define PORTNO(x) (MINOR((x)->device) - (x)->driver.minor_start)
#else
#define PORTNO(x) ((x)->index)
#endif

#define RELEVANT_IFLAG(iflag)	(iflag & (IGNBRK|BRKINT|IGNPAR|PARMRK|INPCK))

#ifndef MIN
#define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif

#define ADVTTYMAJOR		            39
#define ADVTTYCUMAJOR	            38

static int _tty_major = ADVTTYMAJOR;
static int _cout_major = ADVTTYCUMAJOR;
static int _verbose = 1;

int	_DebugLevel = AD_DEBUG_OFF;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
static struct task_struct *kvcom_task[CF_MAXPORTS];
#else
static pid_t kvcom_pid[CF_MAXPORTS];
#endif
static  char nl_pseudo;

#ifdef MODULE

#if (LINUX_VERSION_CODE > VERSION_CODE(2,1,11))
MODULE_DESCRIPTION("TTY Virtual Driver");

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,6))
MODULE_PARM(_tty_major, "i");
MODULE_PARM(_cout_major, "i");
MODULE_PARM(_verbose, "i");
#else
module_param(_tty_major, int, 0);
module_param(_cout_major, int, 0);
module_param(_verbose, int, 0644);
MODULE_VERSION(CF_VERSION);
#endif // 2.6.6

#if (LINUX_VERSION_CODE > VERSION_CODE(2,4,0))
#ifdef MODULE_LICENSE
MODULE_LICENSE("GPL");
#endif
#endif // 2.4.0

#endif // 2.1.11

#endif // MODULE

#ifndef B921600
#define	B921600	(B460800 + 1)
#endif

#define	ADVTTY_CMD_TIMEOUT              (10 * HZ)

#define	ADVTTY_NET_NODE_OPENED			0x01
#define	ADVTTY_NET_NODE_CONNECTED		0x02
#define	ADVTTY_NET_NODE_DISCONNECTED	0x04
#define	ADVTTY_NET_DO_SESSION_RECOVERY	0x08
#define	ADVTTY_NET_DO_INITIALIZE		0x10
#define	ADVTTY_NET_TTY_INUSED			0x20

struct advtty_struct
{
    int			port;
    int			flags;
    int			type;
    struct tty_struct *	tty;
    int			xmit_fifo_size;
    int			custom_divisor;
    int			x_char;
    int			close_delay;
    unsigned short		closing_wait;
    int			modem_control;
    int			modem_status;
    unsigned long		event;
    int			count;
    pid_t	    session;
    pid_t       pgrp;
    unsigned char * xmit_buf;
    int			xmit_head;
    int			xmit_tail;
    int			xmit_cnt;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    struct tq_struct	tqueue;
    struct tq_struct	process_flip_tqueue;
#else
    struct work_struct 	tqueue;
    struct work_struct	process_flip_tqueue;
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
    struct termios		normal_termios;
    struct termios		callout_termios;
#else
    struct ktermios		normal_termios;
    struct ktermios		callout_termios;
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(2,4,0))
    struct wait_queue * open_wait;
    struct wait_queue * close_wait;
    struct wait_queue * delta_msr_wait;
#else
    wait_queue_head_t 	open_wait;
    wait_queue_head_t 	close_wait;
    wait_queue_head_t 	delta_msr_wait;
#endif
    struct async_icount	icount;
    struct nd_struct * net_node;


#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
    spinlock_t          tx_lock;
    spinlock_t          nltx_lock;
    struct semaphore    rx_semaphore;
#else
    struct semaphore    tx_lock;
    struct semaphore    nltx_lock;
    struct semaphore	rx_semaphore;
#endif
};

struct nd_struct
{
#if (LINUX_VERSION_CODE < VERSION_CODE(2,4,0))
    struct wait_queue * initialize_wait;
    struct wait_queue * select_in_wait;
    struct wait_queue * select_out_wait;
    struct wait_queue * select_ex_wait;
    struct wait_queue * cmd_rsp_wait;
#else
    wait_queue_head_t	initialize_wait;
    wait_queue_head_t	select_in_wait;
    wait_queue_head_t	select_out_wait;
    wait_queue_head_t	select_ex_wait;
    wait_queue_head_t	cmd_rsp_wait;
#endif
    int 		isxchar;

    int                 tx_ready;
    int 		tx_done;
    int                 rx_ready;
    int                 cmd_ready;
    int 		write_flags;
    int			rx_thread_ready;
		
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
    spinlock_t          write_flags_lock;
#else
    struct semaphore    write_flags_lock;
#endif
    int                 wait_oqueue_responsed;
    int                 oqueue;
    unsigned char       cmd_buffer[CF_CMDRSPLEN];
    unsigned char       rsp_buffer[CF_CMDRSPLEN];
    struct semaphore    cmd_semaphore;
    int                 rsp_length;
    unsigned long       flag;
    struct proc_dir_entry * node_entry;
    struct advtty_struct  * tty_node;
    struct semaphore    semaphore;
    int                 do_session_recovery_len;
};

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
static struct tty_driver	advar_sdriver;
static struct tty_driver	advar_cdriver;
#else
static struct tty_driver *  advar_sdriver;
#endif
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
static int advar_refcount;
#endif

static struct advtty_struct	advar_table[CF_MAXPORTS];
static struct nd_struct 	advar_net_nodes[CF_MAXPORTS];
//static struct tty_struct *	advar_tty[CF_MAXPORTS];
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
static struct tty_struct *  	advar_tty_table[CF_MAXPORTS];
static struct termios * 	advar_termios[CF_MAXPORTS];
static struct termios * 	advar_termios_locked[CF_MAXPORTS];
#else
static struct ktermios * 	advar_termios[CF_MAXPORTS];
static struct ktermios * 	advar_termios_locked[CF_MAXPORTS];
#endif
static int advar_diagflag;
static struct proc_dir_entry *  advar_proc_root;

/*
// Kernel temp buffer for write
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
static unsigned char *		advar_tmp_buf;
static struct semaphore 	advar_tmp_buf_sem;
#else
static unsigned char *		advar_tmp_buf = 0;
static struct semaphore 	advar_tmp_buf_sem = MUTEX;
#endif
*/


//#ifdef MODULE
//int		init_module(void);
//void	cleanup_module(void);
//#endif

int		advtty_init(void);
static int 	advtty_init_tty(void);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
static void	advtty_do_softint(void *);
static void advtty_flush_to_ldisc(void *);
#else
static void	advtty_do_softint(struct work_struct *work);
static void advtty_flush_to_ldisc(struct work_struct *work);
#endif
static int	advtty_open(struct tty_struct *,struct file *);
static void	advtty_close(struct tty_struct *,struct file *);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
static int	advtty_write(struct tty_struct *,int,const unsigned char *,int);
#else
static int	advtty_write(struct tty_struct *,const unsigned char *,int);
#endif
static int	advtty_write_room(struct tty_struct *);
static void	advtty_flush_buffer(struct tty_struct *);
static void	advtty_flush_txbuffer(struct tty_struct *);
static void	advtty_flush_rxbuffer(struct tty_struct *);
static void	advtty_ldisc_flush_buffer(struct tty_struct *);
static int	advtty_chars_in_buffer(struct tty_struct *);
static void	advtty_flush_chars(struct tty_struct *);
static int	advtty_put_char(struct tty_struct *,unsigned char);
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,0,0))
static int	advtty_ioctl(struct tty_struct *,uint,ulong);
#else
static int	advtty_ioctl(struct tty_struct *,struct file *,uint,ulong);
#endif
static void	advtty_throttle(struct tty_struct *);
static void	advtty_unthrottle(struct tty_struct *);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
static void	advtty_set_termios(struct tty_struct *,struct termios *);
static int	advtty_port_init(struct advtty_struct *,struct termios *);
#else
static void	advtty_set_termios(struct tty_struct *,struct ktermios *);
static int	advtty_port_init(struct advtty_struct *,struct ktermios *);
#endif
static void	advtty_stop(struct tty_struct *);
static void	advtty_start(struct tty_struct *);
static void	advtty_hangup(struct tty_struct *);
static inline void advtty_check_modem_status(struct advtty_struct *,int);
static int	advtty_block_til_ready(struct tty_struct *,struct file *, struct advtty_struct *);
static int	advtty_startup(struct advtty_struct *,struct file *, struct tty_struct *);
static void	advtty_shutdown(struct advtty_struct *);
static int 	advtty_port_shutdown(struct advtty_struct *);
static int	advtty_get_serial_info(struct advtty_struct *, struct serial_struct *);
static int	advtty_set_serial_info(struct advtty_struct *, struct serial_struct *);
static int	advtty_get_lsr_info(struct advtty_struct *,unsigned int *);
static void	advtty_send_break(struct advtty_struct *,int);

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static int 	advtty_tiocmget(struct tty_struct *, struct file *);
static int 	advtty_tiocmset(struct tty_struct *, struct file *, unsigned int, unsigned int);
#else
static int	advtty_get_modem_info(struct advtty_struct *,unsigned int *);
static int	advtty_set_modem_info(struct advtty_struct *,unsigned int, unsigned int *);
#endif

static void	advtty_process_notify(struct nd_struct *,char *,int);
static void advtty_do_session_recovery(struct advtty_struct *);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
static void 	advtty_wait_until_sent(struct tty_struct *,int);
#endif
static int 	advtty_wait_and_set_command(struct nd_struct *,char,char);
static int 	advtty_wait_command_completed(struct nd_struct *,char,char, long,char *,int *);
static long advtty_wait_oqueue(struct advtty_struct *,long);
static int 	advtty_linectrl(struct nd_struct *nd,int modem_control);
static int	 advtty_break(struct tty_struct * ttyinfo, int break_state);
static void advtty_start_break(struct nd_struct *nd);
static void advtty_stop_break(struct nd_struct *nd);
static int  advtty_setxon_xoff(struct advtty_struct * info, int cmd);

static struct proc_dir_entry *advtty_create_proc_entry(const char *, mode_t, struct proc_dir_entry *);
static void advtty_remove_proc_entry( struct proc_dir_entry *);



static int send_to_user(struct tty_struct* tty, unsigned short msgtype, unsigned char *buf, int length);

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static struct tty_operations mpvar_ops =
{
	.open = advtty_open,
	.close = advtty_close,
	.write = advtty_write,
	.put_char = advtty_put_char,
	.flush_chars = advtty_flush_chars,
	.write_room = advtty_write_room,
	.chars_in_buffer = advtty_chars_in_buffer,
	.flush_buffer = advtty_ldisc_flush_buffer,
	.ioctl = advtty_ioctl,
	.throttle = advtty_throttle,
	.unthrottle = advtty_unthrottle,
	.set_termios = advtty_set_termios,
	.stop = advtty_stop,
	.start = advtty_start,
	.hangup = advtty_hangup,
	.tiocmget = advtty_tiocmget,
	.tiocmset = advtty_tiocmset,
	#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
    	.wait_until_sent = advtty_wait_until_sent,
	#endif
        .break_ctl = advtty_break,

};
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,16))
static int n_tty_receive_room(struct tty_struct *tty)
{

	unsigned int count, left, rtn;
	count = TTY_FLIPBUF_SIZE - tty->flip.count;
	left = N_TTY_BUF_SIZE - tty->read_cnt - 1;
	rtn = MIN(count, left);
	return rtn;
/*
	unsigned int count;
	count = TTY_FLIPBUF_SIZE - tty->flip.count;
	int left = N_TTY_BUF_SIZE - tty->read_cnt - 1;
	//printk(KERN_INFO "flip.count=%d, count=%d, left=%d\n", tty->flip.count, count, left);
	if(tty->icanon && !tty->canon_data)
		return N_TTY_BUF_SIZE-1;
	if(left > 0)
		return left;
	return 0;
*/
}
/*
static int n_tty_receive_room(struct tty_struct *tty)
{
	int	left = N_TTY_BUF_SIZE - tty->read_cnt - 1;
	if (left <= 0)
		left = tty->icanon && !tty->canon_data;
	return left;
}
static int n_tty_receive_room(struct tty_struct *tty)
{
	int	left = N_TTY_BUF_SIZE - tty->read_cnt - 1;
	return MIN(TTY_FLIPBUF_SIZE-tty->flip.count, left);
}
*/
#endif
//jinxin added begin to support netlink
/* receive */
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,24))
static void kernel_receive(struct sk_buff *skb)
{
	struct nlmsghdr *nlh = NULL;
	unsigned int index, cmd, pid = 0;
	int size, cnt;
	int status;
	int reval = 0;
	unsigned long flags;
	unsigned long flags1;
	struct tty_struct* tty;
	struct advtty_struct *info;
	struct nd_struct * nd;	
	struct msg_info *packet;
	if(skb->len < sizeof(struct nlmsghdr)){
		DBGPRINT(AD_DEBUG_ERROR, "receive wrong packet\n");
		goto out;
	}
	nlh = nlmsg_hdr(skb);//get msg header
	if((nlh->nlmsg_len >= sizeof(struct nlmsghdr))
            && (skb->len >= nlh->nlmsg_len)) 
	{
		unsigned short umsg;
		//high 8 bits: port number
		//low 8 bits: command number
		umsg = nlh->nlmsg_type & 0x00ff;
		index = nlh->nlmsg_type >> 8;//get high 8 bits
		DBGPRINT(AD_DEBUG_INFO, "umsg=%X, index=%d\n", umsg, index);
		if(umsg == NETLINK_USEND_PID)
		{
			write_lock_bh(&user_proc[index].lock);
			user_proc[index].pid = nlh->nlmsg_pid;
			write_unlock_bh(&user_proc[index].lock);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
       			tty_register_devfs(DRV_VAR, 0, index);
       			tty_register_devfs(&advar_cdriver, 0, index);
#else
       			tty_register_device(DRV_VAR, index, NULL);
#endif

			goto out;
		}
		read_lock_bh(&user_proc[index].lock);
		pid = user_proc[index].pid;
		read_unlock_bh(&user_proc[index].lock);
		DBGPRINT(AD_DEBUG_INFO, "pid=%d, devidx=%d, msgnum=%X\n", pid, index, umsg);
		if(0 == pid)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: Advttyp and driver is not initial well, return\n", index);
			goto out;
		}
		TRACE;
		info = advar_table + index;
		if(!info)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: You have not inited tty\n", index);
			goto out;		
		}
		tty = info->tty;
		if(!tty)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: You have not startup the advtty, umsg=%x\n", index, umsg);
			goto out;
		}
		TRACE;
		if(!(nd=info->net_node))
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: nd_struct is not initialed in advtty_init_tty!\n", index);		
			goto out;
		}
		switch (umsg)
		{
			
			case NETLINK_DEAL_COMD:
				DBGPRINT(AD_DEBUG_INFO, "NETLINK_DEAL_COMD\n");
				//msglen = nlh->nlmsg_len;
				//command length must be 2 Bytes
    				packet = NLMSG_DATA(nlh);
				DBGPRINT(AD_DEBUG_INFO, "comd: %x, size: %x\n", packet->comd, packet->len);
				cmd = packet->comd;
				size = packet->len;//It is only used by IOCMD_GET_TTY_STATUS and IOCMD_RESPONSE

				if(down_trylock(&receive_sem)){
					DBGPRINT(AD_DEBUG_ERROR, "Port %d: can not get sem\n", info->port);
					return;
				}

				switch (cmd)
				{
					case IOCMD_RESPONSE://data zone of nlmsghdr is data returned by user 
					{
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_RESPONSE\n");
        					unsigned char rsp_buffer[CF_CMDRSPLEN];
						if(size < 2){
            						DBGPRINT(AD_DEBUG_ERROR, "Port %d: Size of The Command Response is wrong\n", index);
							goto out;
						}
						TRACE;
        					if(size > 84)
            						size = 84;
						/*
        					rtn = access_ok( VERIFY_READ,  (void *)arg, size )?0:-EFAULT;
        					if(rtn) {
            						goto done;
        					}
       		 				if(copy_from_user( (void *)rsp_buffer, (void *)arg,size)) {
            						rtn = -EFAULT;
            						goto done;
        					}*/
						memcpy((void *)rsp_buffer, (void *)packet->data, size);	
						TRACE;
        					if(rsp_buffer[0] == DREQ_DRIVER_CMDSET) {
            						down(&nd->cmd_semaphore);
            						memcpy(nd->rsp_buffer,rsp_buffer,size);
            						nd->rsp_length = size;
            						up(&nd->cmd_semaphore);
            						if ( waitqueue_active(&nd->cmd_rsp_wait))
                						wake_up_interruptible(&nd->cmd_rsp_wait);
           						 break;
        					}
						TRACE;
						down(&nd->cmd_semaphore);
        					if(nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY) {
            						if (rsp_buffer[1] == DREQ_CMD_LINECTRL) {
                						nd->flag &= ~ ADVTTY_NET_DO_SESSION_RECOVERY;
                						up(&nd->cmd_semaphore);
                						break;
            						}
            						else if (rsp_buffer[1] == DREQ_CMD_PORT_INIT) {
                						int state = 0;
                						struct advtty_struct *info;
                						up(&nd->cmd_semaphore);
                						if (!(info=nd->tty_node))
                    							break;
                						if (size != 6)
                    							break;
                						if (rsp_buffer[2] != 3)
                    							break;
                						if (rsp_buffer[3])
                    							state |= UART_MSR_DSR;
                						if (rsp_buffer[4])
                    							state |= UART_MSR_CTS;
                						if (rsp_buffer[5])
                    							state |= UART_MSR_DCD;
                						advtty_check_modem_status(info,state);
            							}
            							else {
                							up(&nd->cmd_semaphore);
                						break;
            							}
        						}
       						 else
            						up(&nd->cmd_semaphore);
        					if(rsp_buffer[1] == DREQ_NOTIFY) {
            						advtty_process_notify(nd, rsp_buffer, size);
        					}
        					else if(rsp_buffer[1] == DREQ_CMD_WAIT_OQUEUE) {
            						if (size == 5) {
                						memcpy(nd->rsp_buffer,rsp_buffer,size);
                						nd->oqueue = rsp_buffer[4] * 16 + rsp_buffer[3];
                						nd->rsp_length = size;
                						nd->wait_oqueue_responsed = 1;
                						if(waitqueue_active(&nd->cmd_rsp_wait))
                    						wake_up_interruptible(&nd->cmd_rsp_wait);
            						}
        					}
        					else {
            						down(&nd->cmd_semaphore);
            						memcpy(nd->rsp_buffer, rsp_buffer, size);
            						nd->rsp_length = size;
            						up(&nd->cmd_semaphore);
            						if(waitqueue_active(&nd->cmd_rsp_wait))
                						wake_up_interruptible(&nd->cmd_rsp_wait);
       						}
						break;
					}/* end IOCMD_RESPONSE*/
					case IOCMD_CONNECTED: {
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_CONNECTED\n");
       			 			struct advtty_struct *info;
						if(!(info=nd->tty_node))
            						break;
        					if(nd->flag & ADVTTY_NET_NODE_DISCONNECTED) {
            						nd->flag &= ~ADVTTY_NET_NODE_DISCONNECTED;
            						nd->flag |= ADVTTY_NET_NODE_CONNECTED;
            						advtty_do_session_recovery(info);
        					}
        					break;
    					}/* end IOCMD_CONNECTED*/
					case IOCMD_DISCONNECTED:
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_DISCONNECTED\n");
        					nd->flag &= ~ADVTTY_NET_NODE_CONNECTED;
        					nd->flag |= ADVTTY_NET_NODE_DISCONNECTED;
        					if(waitqueue_active(&nd->cmd_rsp_wait)){
            						nd->wait_oqueue_responsed = 1;
            						wake_up_interruptible(&nd->cmd_rsp_wait);
        					}
        					break;
					/*end IOCMD_DISCONNECTED*/
					case IOCMD_GET_TTY_STATUS:
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_GET_TTY_STATUS\n");
						struct advtty_struct *info;

        					if(!(info=nd->tty_node))
            						break;
						if(size != sizeof (status))
							break;
        					status = (nd->flag & ADVTTY_NET_TTY_INUSED) ? 1 : 0;
						reval = send_to_user(tty, NETLINK_TTY_STATUS, (unsigned char *)&status, size);
						if(reval <= 0){
							DBGPRINT(AD_DEBUG_ERROR, "Port %d: send status to user space failed! \n", index);
						}
        					break;
					default:
						break;
				}/* end NETLINK_DEAL_COMD*/
				up(&receive_sem);
				break;
			case NETLINK_KSEND_DATA_OK:
				packet = NLMSG_DATA(nlh);
				int count = (int)(*(int*)packet->data);
				//DBGPRINT(AD_DEBUG_ERROR, "port %d: NETLINK_KSEND_DATA_OK\n", info->port);
				DOWN(info->tx_lock, flags);
				if(nd->isxchar){
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        				info->icount.tx++;
#endif
       					UP(info->tx_lock, flags);
					nd->isxchar = 0;
					DOWN(info->nltx_lock, flags1);
					nd->write_flags = 1;
					UP(info->nltx_lock, flags1);
					if(waitqueue_active(&nd->select_in_wait)){
           	 				wake_up_interruptible( &nd->select_in_wait );
					}
					break;
				}
        			info->xmit_cnt -= count;
        			info->xmit_tail += count;
        			info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE - 1);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        			info->icount.tx += count;
#endif
        			UP(info->tx_lock, flags);
				DOWN(info->nltx_lock, flags1);
				nd->write_flags = 1;
				UP(info->nltx_lock, flags1);
				//wake up vcom_process
				if(waitqueue_active(&nd->select_in_wait)){
           				 wake_up_interruptible( &nd->select_in_wait );
				}
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    				if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       					(tty->ldisc.write_wakeup))
        				(*(tty->ldisc.write_wakeup))(tty);
#else
    				if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       					(tty->ldisc.ops->write_wakeup))
        				(*(tty->ldisc.ops->write_wakeup))(tty);
#endif
#else
    				if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       					(tty->ldisc->ops->write_wakeup))
        				(*(tty->ldisc->ops->write_wakeup))(tty);	
#endif
    				wake_up_interruptible(&tty->write_wait);
				break;
			case NETLINK_USEND_DATA:
				DBGPRINT(AD_DEBUG_INFO, "port %d: NETLINK_USEND_DATA\n", info->port);
    				packet = NLMSG_DATA(nlh);
				size = packet->len;
				if(info->flags & ASYNC_CLOSING){
        				DBGPRINT(AD_DEBUG_ERROR, "port [%d] NETLINK_USEND_DATA : port closed\n", info->port);
					goto recfailed;
				}
				down(&info->rx_semaphore);
				if(test_bit(TTY_IO_ERROR, &tty->flags)) {
	        			DBGPRINT(AD_DEBUG_ERROR, "Port [%d] NETLINK_USEND_DATA : IO ERROR\n", info->port);
        				up(&info->rx_semaphore);
					goto recfailed;
				}
				if(!nd->rx_ready) {
	        			DBGPRINT(AD_DEBUG_INFO, "Port [%d] NETLINK_USEND_DATA : rx_ready is 0\n", info->port);
					up(&info->rx_semaphore);
        				goto recfailed;
    				}
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    				if((cnt = MIN(size,TTY_FLIPBUF_SIZE-tty->flip.count)) <= 0) {
#else
    				if((cnt = tty_buffer_request_room(tty, size)) <= 0) 
				{
#endif

        				if(!tty->icanon || tty->canon_data) {
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
            					if(!set_bit(TTY_THROTTLED,&tty->flags)) {
#else
            					if(!test_and_set_bit(TTY_THROTTLED,&tty->flags)) {
#endif
                					advtty_throttle(tty);
           				 	}
        				}
        				up(&info->rx_semaphore);
					DBGPRINT(AD_DEBUG_ERROR, "kernel_receive: tty core requests room failed\n");
        				goto recfailed;
    				}
    				//here, cnt = tty_buffer_request_room, and cnt > 0, so tty_insert_flip_string will be guaranteed right
    				if(!tty->icanon || tty->canon_data) 
				{
        				if((cnt = MIN(cnt,(N_TTY_BUF_SIZE - 1) - tty->read_cnt)) <= 0) 
					{
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
            					if(!set_bit(TTY_THROTTLED,&tty->flags)) {
#else
            					if(!test_and_set_bit(TTY_THROTTLED,&tty->flags)) {
#endif
                					advtty_throttle(tty);
            					}
            					up(&info->rx_semaphore);
						DBGPRINT(AD_DEBUG_ERROR, "kernel_receive: thread may be not enough space\n");
            					goto recfailed;
        				}
    				}

#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    				if(copy_from_user(tty->flip.char_buf_ptr, (unsigned char *)packet->data, cnt)) {
#else
    				if((size = tty_insert_flip_string(tty, (unsigned char *)packet->data, cnt))) {
        				tty_flip_buffer_push(tty);
#endif
        				info->icount.rx += size;
        				up(&info->rx_semaphore);
					//printk(KERN_INFO "rceive %d, true recv %d\n", packet->len, size);
					reval = send_to_user(tty, NETLINK_USEND_DATA_OK, (unsigned char *)&size, sizeof(size));
    					if(reval <= 0){
						DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_OK to advttyp failed\n", info->port);
					}
        				goto out;
    				}
				//printk(KERN_INFO "delay rceive %d, true recv %d\n", packet->len, cnt);
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    				tty->flip.count += cnt;
    				tty->flip.char_buf_ptr += cnt;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
    				info->icount.rx += cnt;
#endif
    				memset(tty->flip.flag_buf_ptr, TTY_NORMAL, cnt);
    				tty->flip.flag_buf_ptr += cnt;
#endif
				up(&info->rx_semaphore);
   				ADQ_TASK(&info->process_flip_tqueue);
				reval = send_to_user(tty, NETLINK_USEND_DATA_OK, (unsigned char *)&cnt, sizeof(cnt));
    				if(reval <= 0){
					DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_OK to advttyp failed 2\n", info->port);
				}
				goto out;
		}
	}
	else{
		DBGPRINT(AD_DEBUG_ERROR, "wrong nlmsg packet\n");
		DBGPRINT(AD_DEBUG_INFO, "msg type is %X, port index is %d\n", (nlh->nlmsg_type & 0x00ff), (nlh->nlmsg_type >> 8));
	}
out:
	return;
recfailed:
	reval = send_to_user(tty, NETLINK_USEND_DATA_NOK, NULL, 0);
    	if(reval <= 0){
		DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_NOK to advttyp failed\n", info->port);
	}
	goto out;
}
#else
static void kernel_receive(struct sock *sk, int klen)
{
    struct sk_buff *skb;
    struct nlmsghdr *nlh = NULL;
    unsigned int index, msglen, cmd, size, pid = 0;
    int rtn = 0, len, status;
    int reval = 0;
    int cnt;
    unsigned long flags;
    unsigned long flags1;
    struct tty_struct* tty;
    struct advtty_struct *info;
    struct nd_struct * nd;	
    struct msg_info *packet;
    DBGPRINT(AD_DEBUG_INFO, "kernel_receive\n");
    do
    {
	//if(down_trylock(&receive_sem)) //i think it is not neccessary
        //        break;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,5,0))
        while((skb = skb_dequeue(&sk->sk_receive_queue)) != NULL)
#else
        while((skb = skb_dequeue(&sk->receive_queue)) != NULL)
#endif
	{
	    if(skb->len < sizeof(struct nlmsghdr)){
		DBGPRINT(AD_DEBUG_ERROR, "receive wrong packet\n");
		break;
	    }
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,22))
	    nlh = nlmsg_hdr(skb);//get msg header
#else
	    nlh = (struct nlmsghdr*)(skb->data);
#endif
	    if((nlh->nlmsg_len >= sizeof(struct nlmsghdr))
                && (skb->len >= nlh->nlmsg_len)) 
	    {
		unsigned short umsg;
		//high 8 bits: port number
		//low 8 bits: command number
		umsg = nlh->nlmsg_type & 0x00ff;
		index = nlh->nlmsg_type >> 8;//get high 8 bits
		DBGPRINT(AD_DEBUG_INFO, "umsg=%X, index=%d\n", umsg, index);
		if(umsg == NETLINK_USEND_PID)
		{
			write_lock_bh(&user_proc[index].lock);
			user_proc[index].pid = nlh->nlmsg_pid;
			write_unlock_bh(&user_proc[index].lock);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
       			tty_register_devfs(DRV_VAR, 0, index);
       			tty_register_devfs(&advar_cdriver, 0, index);
#else
       			tty_register_device(DRV_VAR, index, NULL);
#endif
			continue;
		}
		read_lock_bh(&user_proc[index].lock);
		pid = user_proc[index].pid;
		read_unlock_bh(&user_proc[index].lock);
		DBGPRINT(AD_DEBUG_INFO, "pid=%d, devidx=%d, msgnum=%X\n", pid, index, umsg);
		if(0 == pid)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: Advttyp and driver is not initial well\n", index);
			continue;
		}
		TRACE;
		info = advar_table + index;
		if(!info)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: You have not inited tty\n", index);
			continue;
		}
		tty = info->tty;
		if(!tty)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: You have not startup the advtty, umsg=%x\n", index, umsg);
			continue;
		}
		TRACE;
		if(!(nd=info->net_node))
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: nd_struct is not initialed in advtty_init_tty!\n", index);		
			continue;
		}
		switch (umsg)
		{
			
			case NETLINK_DEAL_COMD:
				DBGPRINT(AD_DEBUG_INFO, "NETLINK_DEAL_COMD\n");
				//msglen = nlh->nlmsg_len;
				//command length must be 2 Bytes
    				packet = NLMSG_DATA(nlh);
				DBGPRINT(AD_DEBUG_INFO, "comd: %x, size: %x\n", packet->comd, packet->len);
				cmd = packet->comd;
				size = packet->len;//It is only used by IOCMD_GET_TTY_STATUS and IOCMD_RESPONSE

				if(down_trylock(&receive_sem)){
					DBGPRINT(AD_DEBUG_ERROR, "Port %d: can not get sem\n", info->port);
					break;
				}

				switch (cmd)
				{
					case IOCMD_RESPONSE://data zone of nlmsghdr is data returned by user 
					{
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_RESPONSE\n");
        					unsigned char rsp_buffer[CF_CMDRSPLEN];
						if(size < 2){
            						DBGPRINT(AD_DEBUG_ERROR, "Port %d: Size of The Command Response is wrong\n", index);
							break;
						}
						TRACE;
        					if(size > 84)
            						size = 84;
						/*
        					rtn = access_ok( VERIFY_READ,  (void *)arg, size )?0:-EFAULT;
        					if(rtn) {
            						goto done;
        					}
       		 				if(copy_from_user( (void *)rsp_buffer, (void *)arg,size)) {
            						rtn = -EFAULT;
            						goto done;
        					}*/
						memcpy((void *)rsp_buffer, (void *)packet->data, size);	
						TRACE;
        					if(rsp_buffer[0] == DREQ_DRIVER_CMDSET) {
            						down(&nd->cmd_semaphore);
            						memcpy(nd->rsp_buffer,rsp_buffer,size);
            						nd->rsp_length = size;
            						up(&nd->cmd_semaphore);
            						if ( waitqueue_active(&nd->cmd_rsp_wait))
                						wake_up_interruptible(&nd->cmd_rsp_wait);
           						 break;
        					}
						TRACE;
						down(&nd->cmd_semaphore);
        					if(nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY) {
            						if (rsp_buffer[1] == DREQ_CMD_LINECTRL) {
                						nd->flag &= ~ ADVTTY_NET_DO_SESSION_RECOVERY;
                						up(&nd->cmd_semaphore);
                						break;
            						}
            						else if (rsp_buffer[1] == DREQ_CMD_PORT_INIT) {
                						int state = 0;
                						struct advtty_struct *info;
                						up(&nd->cmd_semaphore);
                						if (!(info=nd->tty_node))
                    							break;
                						if (size != 6)
                    							break;
                						if (rsp_buffer[2] != 3)
                    							break;
                						if (rsp_buffer[3])
                    							state |= UART_MSR_DSR;
                						if (rsp_buffer[4])
                    							state |= UART_MSR_CTS;
                						if (rsp_buffer[5])
                    							state |= UART_MSR_DCD;
                						advtty_check_modem_status(info,state);
            							}
            							else {
                							up(&nd->cmd_semaphore);
                						break;
            							}
        						}
       						 else
            						up(&nd->cmd_semaphore);
        					if(rsp_buffer[1] == DREQ_NOTIFY) {
            						advtty_process_notify(nd, rsp_buffer, size);
        					}
        					else if(rsp_buffer[1] == DREQ_CMD_WAIT_OQUEUE) {
            						if (size == 5) {
                						memcpy(nd->rsp_buffer,rsp_buffer,size);
                						nd->oqueue = rsp_buffer[4] * 16 + rsp_buffer[3];
                						nd->rsp_length = size;
                						nd->wait_oqueue_responsed = 1;
                						if(waitqueue_active(&nd->cmd_rsp_wait))
                    						wake_up_interruptible(&nd->cmd_rsp_wait);
            						}
        					}
        					else {
            						down(&nd->cmd_semaphore);
            						memcpy(nd->rsp_buffer, rsp_buffer, size);
            						nd->rsp_length = size;
            						up(&nd->cmd_semaphore);
            						if(waitqueue_active(&nd->cmd_rsp_wait))
                						wake_up_interruptible(&nd->cmd_rsp_wait);
       						}
						break;
					}/* end IOCMD_RESPONSE*/
					case IOCMD_CONNECTED: {
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_CONNECTED\n");
       			 			struct advtty_struct *info;
						if(!(info=nd->tty_node))
            						break;
        					if(nd->flag & ADVTTY_NET_NODE_DISCONNECTED) {
            						nd->flag &= ~ADVTTY_NET_NODE_DISCONNECTED;
            						nd->flag |= ADVTTY_NET_NODE_CONNECTED;
            						advtty_do_session_recovery(info);
        					}
        					break;
    					}/* end IOCMD_CONNECTED*/
					case IOCMD_DISCONNECTED:
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_DISCONNECTED\n");
        					nd->flag &= ~ADVTTY_NET_NODE_CONNECTED;
        					nd->flag |= ADVTTY_NET_NODE_DISCONNECTED;
        					if(waitqueue_active(&nd->cmd_rsp_wait)){
            						nd->wait_oqueue_responsed = 1;
            						wake_up_interruptible(&nd->cmd_rsp_wait);
        					}
        					break;
					/*end IOCMD_DISCONNECTED*/
					case IOCMD_GET_TTY_STATUS:
						DBGPRINT(AD_DEBUG_INFO, "IOCMD_GET_TTY_STATUS\n");
						struct advtty_struct *info;

        					if(!(info=nd->tty_node))
            						break;
						if(size != sizeof (status))
							break;
        					status = (nd->flag & ADVTTY_NET_TTY_INUSED) ? 1 : 0;
						reval = send_to_user(tty, NETLINK_TTY_STATUS, (unsigned char *)&status, size);
						if(reval <= 0){
							DBGPRINT(AD_DEBUG_ERROR, "Port %d: send status to user space failed! \n", index);
						}
        					break;
					default:
						break;
				}/* end NETLINK_DEAL_COMD*/
				up(&receive_sem);
				break;
			case NETLINK_KSEND_DATA_OK:
				packet = NLMSG_DATA(nlh);
				int count = (int)(*(int*)packet->data);
				//DBGPRINT(AD_DEBUG_ERROR, "port %d: NETLINK_KSEND_DATA_OK\n", info->port);
				DOWN(info->tx_lock, flags);
				if(nd->isxchar){
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        				info->icount.tx++;
#endif
       					UP(info->tx_lock, flags);
					nd->isxchar = 0;
					DOWN(info->nltx_lock, flags1);
					nd->write_flags = 1;
					UP(info->nltx_lock, flags1);
					if(waitqueue_active(&nd->select_in_wait)){
           	 				wake_up_interruptible( &nd->select_in_wait );
					}
					break;
				}
        			info->xmit_cnt -= count;
        			info->xmit_tail += count;
        			info->xmit_tail = info->xmit_tail & (SERIAL_XMIT_SIZE - 1);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        			info->icount.tx += count;
#endif
        			UP(info->tx_lock, flags);
				DOWN(info->nltx_lock, flags1);
				nd->write_flags = 1;
				UP(info->nltx_lock, flags1);
				//wake up vcom_process
				if(waitqueue_active(&nd->select_in_wait)){
           				 wake_up_interruptible( &nd->select_in_wait );
				}
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    				if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       					(tty->ldisc.write_wakeup))
        				(*(tty->ldisc.write_wakeup))(tty);
#else
    				if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       					(tty->ldisc.ops->write_wakeup))
        				(*(tty->ldisc.ops->write_wakeup))(tty);
#endif
#else
    				if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       					(tty->ldisc->ops->write_wakeup))
        				(*(tty->ldisc->ops->write_wakeup))(tty);	
#endif
    				wake_up_interruptible(&tty->write_wait);
				break;
			case NETLINK_USEND_DATA:
				DBGPRINT(AD_DEBUG_INFO, "port %d: NETLINK_USEND_DATA\n", info->port);
    				packet = NLMSG_DATA(nlh);
				size = packet->len;
				if(info->flags & ASYNC_CLOSING){
        				DBGPRINT(AD_DEBUG_ERROR, "port [%d] NETLINK_USEND_DATA : port closed\n", info->port);
					goto recfailed;
				}
				down(&info->rx_semaphore);
				if(test_bit(TTY_IO_ERROR, &tty->flags)) {
	        			DBGPRINT(AD_DEBUG_ERROR, "Port [%d] NETLINK_USEND_DATA : IO ERROR\n", info->port);
        				up(&info->rx_semaphore);
					goto recfailed;
				}
				if(!nd->rx_ready) {
	        			DBGPRINT(AD_DEBUG_INFO, "Port [%d] NETLINK_USEND_DATA : rx_ready is 0\n", info->port);
					up(&info->rx_semaphore);
        				goto recfailed;
    				}
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    				if((cnt = MIN(size,TTY_FLIPBUF_SIZE-tty->flip.count)) <= 0) {
#else
    				if((cnt = tty_buffer_request_room(tty, size)) <= 0) 
				{
#endif

        				if(!tty->icanon || tty->canon_data) {
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
            					if(!set_bit(TTY_THROTTLED,&tty->flags)) {
#else
            					if(!test_and_set_bit(TTY_THROTTLED,&tty->flags)) {
#endif
                					advtty_throttle(tty);
           				 	}
        				}
        				up(&info->rx_semaphore);
					DBGPRINT(AD_DEBUG_ERROR, "kernel_receive: tty core requests room failed\n");
        				goto recfailed;
    				}
    				//here, cnt = tty_buffer_request_room, and cnt > 0, so tty_insert_flip_string will be guaranteed right
    				if(!tty->icanon || tty->canon_data) 
				{
        				if((cnt = MIN(cnt,(N_TTY_BUF_SIZE - 1) - tty->read_cnt)) <= 0) 
					{
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
            					if(!set_bit(TTY_THROTTLED,&tty->flags)) {
#else
            					if(!test_and_set_bit(TTY_THROTTLED,&tty->flags)) {
#endif
                					advtty_throttle(tty);
            					}
            					up(&info->rx_semaphore);
						DBGPRINT(AD_DEBUG_ERROR, "kernel_receive: thread may be not enough space\n");
            					goto recfailed;
        				}
    				}

#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    				//if(copy_from_user(tty->flip.char_buf_ptr, (unsigned char *)packet->data, cnt)) {
    				if(!memcpy(tty->flip.char_buf_ptr, (unsigned char *)packet->data, cnt)) {
					printk(KERN_INFO "copy data to char_buf_ptr failed\n");
        				up(&info->rx_semaphore);
					goto recfailed;
				}
#else
    				if((size = tty_insert_flip_string(tty, (unsigned char *)packet->data, cnt))) {
        				tty_flip_buffer_push(tty);
        				info->icount.rx += size;
        				up(&info->rx_semaphore);
					//printk(KERN_INFO "rceive %d, true recv %d\n", packet->len, size);
					reval = send_to_user(tty, NETLINK_USEND_DATA_OK, (unsigned char *)&size, sizeof(size));
    					if(reval <= 0){
						DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_OK to advttyp failed\n", info->port);
					}
					break;
    				}
#endif
				//printk(KERN_INFO "delay rceive %d, true recv %d\n", packet->len, cnt);
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    				tty->flip.count += cnt;
    				tty->flip.char_buf_ptr += cnt;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
    				info->icount.rx += cnt;
#endif
    				memset(tty->flip.flag_buf_ptr, TTY_NORMAL, cnt);
    				tty->flip.flag_buf_ptr += cnt;
#endif
				up(&info->rx_semaphore);
   				ADQ_TASK(&info->process_flip_tqueue);
				break;
recfailed:
				reval = send_to_user(tty, NETLINK_USEND_DATA_NOK, NULL, 0);
    				if(reval <= 0){
					DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_NOK to advttyp failed\n", info->port);
				}
				break;
			default:
				break;
		}// switch end
	    }// if  end
            kfree_skb(skb);
	}//end while(skb_dequeue)
	//up(&receive_sem);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,5,0))
    }while(nlfd && nlfd->sk_receive_queue.qlen);
#else
    }while(nlfd && nlfd->receive_queue.qlen);
#endif
    return;
}
#endif
/**********************************************************************************
 * the real send function
 * msgtype: message type, it assigns data int buf is command or pure data or else.
 * buf: 
 * length: length of  Valid data in buf
 *********************************************************************************/
static int send_to_user(struct tty_struct* tty, unsigned short msgtype, unsigned char *buf, int length)
{
    int ret, index;
    int size, pid;
    unsigned char *old_tail;
    struct sk_buff *skb;
    struct nlmsghdr *nlh;
    unsigned char *packet;
    index = PORTNO(tty);
    /* size = sizeof(*info) + sizeof(nlmsghdr) */
    read_lock(&user_proc[index].lock);
    pid = user_proc[index].pid;
    read_unlock(&user_proc[index].lock);
    if(!pid){
	return -EINVAL;    
    }
    size = NLMSG_LENGTH(length);

    skb = alloc_skb(size, GFP_ATOMIC);
    old_tail = skb->tail;
    /* initial a netlink message head */
    //nlh = NLMSG_PUT(skb, 0, 0, msgtype, NLMSG_LENGTH(length));
    nlh = NLMSG_PUT(skb, 0, 0, msgtype, length);
    /* data zone */
    packet = NLMSG_DATA(nlh);
    if(length > 0)
    	memcpy(packet, buf, length);
    //nlh->nlmsg_len = length;//skb->tail - old_tail;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,14))
    NETLINK_CB(skb).dst_group = 0;
#else
    NETLINK_CB(skb).dst_groups = 0;
#endif
    ret = netlink_unicast(nlfd, skb, pid, MSG_DONTWAIT);
    if(ret <= 0){
	DBGPRINT(AD_DEBUG_ERROR, "can not unicast skb(%d)\n", ret);
	return ret;
    }
    return ret;

nlmsg_failure:
    if(skb)
        kfree_skb(skb);
    return -1;
}

int vcom_read(char *buffer,
        char **buffer_location,
        off_t offset, int buffer_length, int *eof, void *data)
{
	if(buffer_length != sizeof(nl_pseudo)){
		buffer_length = sizeof(nl_pseudo);
	}
        int ret;
        if (offset > 0) {
		printk(KERN_INFO "offset > 0\n");
                ret = 0;
        } else {
		memcpy(buffer, &nl_pseudo, buffer_length);
        }
        return buffer_length;
}

#ifdef MODULE
INIT_FUNC_RET INIT_FUNC(void)
{
    int	ret, i, j;
  
    for(i=0; i<CF_MAXPORTS; i++){
	    rwlock_init(&user_proc[i].lock);
			user_proc[i].pid = 0;
	}
    printk(KERN_ERR "Loading advtty module major(%d), coutmajor(%d)...\n",
             _tty_major, _cout_major);

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,0,0))
	sema_init(&receive_sem, 1);
#else
    init_MUTEX(&receive_sem);
#endif
#else
    receive_sem = MUTEX;
#endif
    for(j=MAX_LINKS-1; j>0; j--){
    //jinxin added begin  to support netlink
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,24))
    	nlfd = netlink_kernel_create(&init_net, j, 0, kernel_receive, NULL, THIS_MODULE);
#else  // < VERSION_CODE(2,6,24)
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,22))
    	nlfd = netlink_kernel_create(j, 0, kernel_receive, NULL, THIS_MODULE);
#else  // < VERSION_CODE(2,6,22)
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,14))
    	nlfd = netlink_kernel_create(j, 0, kernel_receive, THIS_MODULE);
#else  // < VERSION_CODE(2,6,14)
    	nlfd = netlink_kernel_create(j, kernel_receive);
#endif // end VERSION_CODE(2,6,14)
#endif // end VERSION_CODE(2,6,22)
#endif// end VERSION_CODE(2,6,24)
	if(nlfd){
		nl_pseudo = j;
		printk(KERN_ERR "Register nelink ok, id is %d\n", nl_pseudo);
		break;
	}
    }
    if(!nlfd) {
        DBGPRINT(AD_DEBUG_ERROR, "can not create a netlink socket\n");
        return -1;
    }

    advar_proc_root = advtty_create_proc_entry( "vcom", 0644, NULL );
    if (advar_proc_root == NULL) {
	ret = -ENOMEM;
      	printk(KERN_INFO "fortune: Couldn't create proc entry\n");
    } 
    else{
      	advar_proc_root->read_proc = vcom_read;
      	//advar_proc_root->write_proc = vcom_write;
    }

    ret = advtty_init();
    DBGPRINT(AD_DEBUG_INFO, "Done.\n");
    TRACE;
    return(ret);
}

CLEAR_FUNC_RET CLEAR_FUNC(void)
{
    int i,err = 0;
    struct advtty_struct *info;
    TRACE;
    //release socket file descriptor
    if(nlfd) {
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,5,71))
        sock_release(nlfd->sk_socket);
#else
        sock_release(nlfd->socket);
#endif
    }
    TRACE;
    info = &advar_table[0];
    //release dev file
    for(i = 0; i < CF_MAXPORTS; i++, info++) {
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
    	tty_unregister_device(DRV_VAR, i);
#else 
    	tty_unregister_devfs(DRV_VAR, i);
    	tty_unregister_devfs(&advar_cdriver, i);
#endif
    }
    if(advar_proc_root) {
        advtty_remove_proc_entry(advar_proc_root);
        advar_proc_root = NULL;
    }

    /*
    for(i = 0; i < CF_MAXPORTS; i++, info++) {
        if(info->net_node) {
            if((de=((struct nd_struct *)(info->net_node))->node_entry))
                advtty_remove_proc_entry(de);
            ((struct nd_struct *)(info->net_node))->node_entry = NULL;
        }
    }
    TRACE;
    if(advar_proc_root) {
        advtty_remove_proc_entry(advar_proc_root);
        advar_proc_root = NULL;
    }*/
    DBGPRINT(AD_DEBUG_INFO, "Unloading Advtty module\n");
    TRACE;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    if((err |= tty_unregister_driver(&advar_cdriver))) {
        DBGPRINT(AD_DEBUG_ERROR, "Couldn't unregister Advtty callout driver\n");
    }
#endif
    TRACE;
    if((err |= tty_unregister_driver(DRV_VAR))) {
        DBGPRINT(AD_DEBUG_ERROR, "Couldn't unregister Advtty TTY driver\n");
    }
    TRACE;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
    put_tty_driver(DRV_VAR);
#endif
    TRACE;
    DBGPRINT(AD_DEBUG_INFO, "Done.\n");

}
#endif

static int advtty_init_tty(void)
{
    struct advtty_struct * tty_node;
    int	i;
    struct nd_struct *net_node;
    char	buf[4];
    tty_node = &advar_table[0];
    net_node = &advar_net_nodes[0];

    for(i = 0; i < CF_MAXPORTS; i++, tty_node++, net_node++) {
        sprintf(buf,"%d",i);
        net_node->tty_node = tty_node;
        net_node->flag = 0;
        
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,0,0))
		sema_init(&net_node->semaphore, 1);
        sema_init(&net_node->cmd_semaphore, 1);
#else
        init_MUTEX(&net_node->semaphore);
        init_MUTEX(&net_node->cmd_semaphore);
#endif
        init_waitqueue_head(&net_node->initialize_wait);
        init_waitqueue_head(&net_node->select_in_wait);
        init_waitqueue_head(&net_node->select_out_wait);
        init_waitqueue_head(&net_node->select_ex_wait);
        init_waitqueue_head(&net_node->cmd_rsp_wait);
#else
        net_node->semaphore = MUTEX;
        net_node->cmd_semaphore = MUTEX;
#endif

        tty_node->net_node = net_node;
        tty_node->port = i;
        tty_node->type = PORT_16550A;
        tty_node->flags = 0;
        tty_node->xmit_fifo_size = 16;
        tty_node->close_delay = 5 * (HZ / 10);
        tty_node->closing_wait = 30 * HZ;
        
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))        
        INIT_WORK(&tty_node->tqueue, advtty_do_softint, tty_node);
        INIT_WORK(&tty_node->process_flip_tqueue, advtty_flush_to_ldisc, tty_node);
#else
		INIT_WORK(&tty_node->tqueue, advtty_do_softint);
        INIT_WORK(&tty_node->process_flip_tqueue, advtty_flush_to_ldisc);
#endif

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
        tty_node->callout_termios = advar_cdriver.init_termios;
#endif

        tty_node->normal_termios = DRV_VAR_P(init_termios);
        
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,0,0))
	tty_node->tx_lock = __SPIN_LOCK_UNLOCKED(tty_node->tx_lock);
	tty_node->nltx_lock = __SPIN_LOCK_UNLOCKED(tty_node->nltx_lock);
#else
        tty_node->tx_lock = SPIN_LOCK_UNLOCKED;
        tty_node->nltx_lock = SPIN_LOCK_UNLOCKED;
#endif
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,0,0))
		sema_init(&tty_node->rx_semaphore, 1);
#else
        init_MUTEX(&tty_node->rx_semaphore);
#endif
        init_waitqueue_head(&tty_node->open_wait);
        init_waitqueue_head(&tty_node->close_wait);
        init_waitqueue_head(&tty_node->delta_msr_wait);
#else
        tty_node->tx_lock = MUTEX;
        tty_node->nltx_lock = MUTEX;
        net_node->write_flags_lock = MUTEX;
        tty_node->rx_semaphore = MUTEX;
        tty_node->modem_control = 0;
#endif

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        tty_node->icount.rx = tty_node->icount.tx = 0;
#endif

        tty_node->icount.cts = tty_node->icount.dsr
                             = tty_node->icount.dsr
                             = tty_node->icount.dcd = 0;
                             
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        tty_node->icount.frame = tty_node->icount.overrun
                               = tty_node->icount.brk
                               = tty_node->icount.parity = 0;
#endif
    }
    return 0;
}



int advtty_init(void)
{
    int	ret1, ret2;

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
    advar_sdriver = alloc_tty_driver(CF_MAXPORTS);
    if(!advar_sdriver)
        return -ENOMEM;
#endif

    DBGPRINT(AD_DEBUG_INFO, "Advtty driver ttymajor %d coutmajor %d verbose %d (%s)\n",
           _tty_major, _cout_major, _verbose, CF_VERSION);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    memset(&advar_sdriver, 0, sizeof(struct tty_driver));
#endif
    // Initialize the tty_driver structure
    DRV_VAR_P(magic) = TTY_DRIVER_MAGIC;
#if (LINUX_VERSION_CODE > 0x2032D && defined (CONFIG_DEVFS_FS))
    DRV_VAR_P(name) = "vttyap/%d";
#else
    DRV_VAR_P(name) = "vttyAP";
#endif

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,18))
    DRV_VAR_P(devfs_name) = "tts/v";
#endif
#endif
    DRV_VAR_P(major) = _tty_major;
    DRV_VAR_P(minor_start) = 0;
    DRV_VAR_P(num) = CF_MAXPORTS;
    DRV_VAR_P(type) = TTY_DRIVER_TYPE_SERIAL;
    DRV_VAR_P(subtype) = SERIAL_TYPE_NORMAL;
    DRV_VAR_P(init_termios) = tty_std_termios;
    DRV_VAR_P(init_termios.c_cflag) = B9600 | CS8| CREAD | HUPCL | CLOCAL;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,18))
    DRV_VAR_P(flags) = TTY_DRIVER_REAL_RAW | TTY_DRIVER_NO_DEVFS;
#else
    DRV_VAR_P(flags) = TTY_DRIVER_REAL_RAW | TTY_DRIVER_DYNAMIC_DEV;
#endif

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
    tty_set_operations(DRV_VAR, &mpvar_ops);
    //DRV_VAR_P(ttys) = advar_tty;
#else
    DRV_VAR_P(refcount) = &advar_refcount;
    DRV_VAR_P(table) = advar_tty_table;
    DRV_VAR_P(open) = advtty_open;
    DRV_VAR_P(close) = advtty_close;
    DRV_VAR_P(write) = advtty_write;
    DRV_VAR_P(put_char) = advtty_put_char;
    DRV_VAR_P(flush_chars) = advtty_flush_chars;
    DRV_VAR_P(write_room) = advtty_write_room;
    DRV_VAR_P(chars_in_buffer) = advtty_chars_in_buffer;
    DRV_VAR_P(flush_buffer) = advtty_flush_buffer;
    DRV_VAR_P(ioctl) = advtty_ioctl;
    DRV_VAR_P(throttle) = advtty_throttle;
    DRV_VAR_P(unthrottle) = advtty_unthrottle;
    DRV_VAR_P(set_termios) = advtty_set_termios;
    DRV_VAR_P(stop) = advtty_stop;
    DRV_VAR_P(start) = advtty_start;
    DRV_VAR_P(hangup) = advtty_hangup;

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
    DRV_VAR_P(wait_until_sent) = advtty_wait_until_sent;    
#endif
    DRV_VAR_P(break_ctl) = advtty_break;
#endif

    DRV_VAR_P(termios) = advar_termios;
    DRV_VAR_P(termios_locked) = advar_termios_locked;

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    /*
     * The callout device is just like normal device except for
     * major number and the subtype code.
     */
    advar_cdriver = advar_sdriver;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,3,26) && defined (CONFIG_DEVFS_FS))
    advar_cdriver.name = "cuv/%d";
#else
    advar_cdriver.name = "cuv";
#endif 
    advar_cdriver.major = _cout_major;
    advar_cdriver.subtype = SERIAL_TYPE_CALLOUT;
#endif
    DBGPRINT(AD_DEBUG_INFO, "TTY devices MAJOR = %d, callout devices MAJOR = %d\n",
             _tty_major, _cout_major);

    advar_diagflag = 0;
    memset(advar_table, 0, CF_MAXPORTS * sizeof(struct advtty_struct));

    ret1 = 0;
    ret2 = 0;
    if(!(ret1 = tty_register_driver(DRV_VAR))) {
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
        if ((ret2=tty_register_driver(&advar_cdriver))) {
            tty_unregister_driver(DRV_VAR);
            DBGPRINT(AD_DEBUG_ERROR, "Couldn't install Advtty callout driver !\n");
        }
#endif
    }
    else {
        DBGPRINT(AD_DEBUG_ERROR, "Couldn't install Advtty driver !\n");
    }

    if(ret1 || ret2) {
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
        put_tty_driver(DRV_VAR);
#endif
        return -1;
    }
    if(advtty_init_tty() != 0) {
        tty_unregister_driver(DRV_VAR);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
        tty_unregister_driver(&advar_cdriver);
#endif
        DBGPRINT(AD_DEBUG_ERROR, "Couldn't install ADVTTY driver !\n");
    }
    return(0);
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
static void advtty_do_softint(void *private_)
{
    struct advtty_struct *	info = (struct advtty_struct *)private_;
#else
static void advtty_do_softint(struct work_struct *work)
{
    struct advtty_struct *	info = 
    	container_of(work, struct advtty_struct, tqueue);
#endif

    struct tty_struct *	tty;

    if(!info)
        goto done;

    tty = info->tty;
    if(tty) {
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
        if(clear_bit(ADVTTY_EVENT_TXLOW, &info->event)) {
            if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
               (tty->ldisc.ops->write_wakeup))
                (*(tty->ldisc.ops->write_wakeup))(tty);
            wake_up_interruptible(&tty->write_wait);
        }
        if(clear_bit(ADVTTY_EVENT_HANGUP, &info->event)) {
            tty_hangup(tty);
        }
#else
        if(test_and_clear_bit(ADVTTY_EVENT_TXLOW, &info->event)) {
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
            if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
		tty->ldisc.write_wakeup )
                (*(tty->ldisc.write_wakeup))(tty);

#else
            if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
                    tty->ldisc.ops->write_wakeup )
                (*(tty->ldisc.ops->write_wakeup))(tty);
#endif
#else
	    if ( (tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
                    tty->ldisc->ops->write_wakeup )
                (*(tty->ldisc->ops->write_wakeup))(tty);
#endif
            wake_up_interruptible(&tty->write_wait);
        }
        if(test_and_clear_bit(ADVTTY_EVENT_HANGUP, &info->event)) {
            tty_hangup(tty);
        }
#endif
    }

done:
    ;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    AD_MOD_DEC;
#endif
}


// This routine is called whenever a serial port is opened.
static int advtty_open(struct tty_struct * tty, struct file * filp)
{
    struct advtty_struct *info;
    int line, pid;
    int rtn = 0, send = 0;
    struct nd_struct * nd;
    AD_MOD_INC;
    line = PORTNO(tty);
    DBGPRINT(AD_DEBUG_INFO, "line=%d\n", line);
    //jinxin added begin
    /* initial the advttyp */
    read_lock(&user_proc[line].lock);
    pid = user_proc[line].pid;
    read_unlock(&user_proc[line].lock);
    DBGPRINT(AD_DEBUG_INFO, "pid=%d\n", pid);
    if(!pid)
    {
	DBGPRINT(AD_DEBUG_ERROR, "ERROR: Can not get pid!\n");
        rtn = -ENODEV;
	goto done;
    }
    info = advar_table + line;
    TRACE;
    nd = info->net_node;
    TRACE;
    /* tell advttyp to initial */
    send = send_to_user(tty, NETLINK_KINIT_LINE, NULL, 0);
    if(send <= 0)
    {
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_open> Send  msg NETLINK_KINIT_LINE to use space failed !\n", info->port);
	rtn = -EPERM;
	goto done;
    }
    TRACE;
    //jinxin added end
#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
    if(!suser()) {
#else
    if(!capable(CAP_SYS_ADMIN)) {
#endif
        rtn = -EPERM;
        goto done;
    }
    TRACE;
    down(&nd->semaphore);
    if(nd->flag & ADVTTY_NET_NODE_OPENED) {
        rtn = -EBUSY;
        up( &nd->semaphore );
	goto done;
    }
    TRACE;
    nd->flag |= ADVTTY_NET_NODE_OPENED;
    nd->tx_ready = 0;
    nd->tx_done = 0;
    nd->rx_ready = 1;
    nd->cmd_ready = 0;
    nd->write_flags = 1;
    nd->isxchar = 0;
    nd->rx_thread_ready = 0;
    up( &nd->semaphore );
    TRACE;
    tty->driver_data = info;
    info->count++;
    if((line < 0) || (line >= CF_MAXPORTS)) {
        DBGPRINT(KERN_DEBUG,"invalid line (%d)\n", line);
        rtn = -ENODEV;
	goto done;
    }
    TRACE;
    if(!nd  || !(nd->flag & ADVTTY_NET_NODE_OPENED)) {
        DBGPRINT(KERN_DEBUG,"net device not ready\n");
        rtn = -ENODEV;
	goto done;
    }
    TRACE;
    //jinxin added end
    if(advtty_startup(info,filp,tty)) {
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_startup failed\n", info->port);
        return(-EIO);
    }
    TRACE;
    if(advtty_block_til_ready(tty, filp, info)) {
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_block_til_ready failed\n", info->port);
        return(-EIO);
    }
    TRACE;
    if((info->count == 1) && (info->flags & ASYNC_SPLIT_TERMIOS)) {
        if(AD_TTY_DRV(subtype) == SERIAL_TYPE_NORMAL)
            *tty->termios = info->normal_termios;
        else
            *tty->termios = info->callout_termios;
        if(advtty_port_init(info, 0)) {
            DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_port_init failed\n", info->port);
            return(-EIO);
        }
    TRACE;
    }
    TRACE;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,6))
    info->session = current->session;
#else

#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
    info->session = pid_nr(AD_SESSION());
#else
    info->session =current->signal->session ;
#endif

#endif
    info->pgrp = AD_CGRP();

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
    tty->low_latency = 1;
#endif
done:
    TRACE;
    if(rtn) {
        AD_MOD_DEC;
    }
    return rtn;
}

static void advtty_close(struct tty_struct * tty, struct file * filp)
{
    struct advtty_struct * info;
    struct nd_struct *nd;
    int line, pid;
    long  timeout,et,ret;
    int	cnt;
    //jinxin added begin
    TRACE;
    line = PORTNO(tty);
    read_lock(&user_proc[line].lock);
    pid = user_proc[line].pid;
    read_unlock(&user_proc[line].lock);
    TRACE;
    if(!pid)
   	return;
    //jinxin added end
    info = (struct advtty_struct *)tty->driver_data;
    if(!info){
        return;
    }
    DBGPRINT(AD_DEBUG_ERROR, "port %d advtty_shutdown\n", info->port);
    nd = info->net_node;
    if(!nd){
	return;
    }
    if(tty_hung_up_p(filp)) {
        info->count--;
        AD_MOD_DEC;
        return;
    }
#ifndef SP1
    if((tty->count == 1) && (info->count != 1)) {
#else
#if (LINUX_VERSION_CODE < VERSION_CODE(2,4,21))
    if((tty->count == 1) && (info->count != 1)) {
#else
    if((atomic_read(&tty->count) == 1) && (info->count != 1)) {
#endif
#endif
	info->count = 1;
    }
    if(--info->count < 0) {
        info->count = 0;
    }
    if(info->count) {
        AD_MOD_DEC;
        return;
    }

    if(info->flags & ASYNC_CLOSING)
	return;

    info->flags |= ASYNC_CLOSING;
    tty->closing = 1;
    if(info->flags & ASYNC_NORMAL_ACTIVE)
    {
	info->normal_termios = *tty->termios;
	}
    if(info->flags & ASYNC_CALLOUT_ACTIVE)
    {
        info->callout_termios = *tty->termios;
     }   
    TRACE;
    // If the open mode is nonblocking mode, don't block on close.
    if(!(filp->f_flags & O_NONBLOCK) && 
       (info->closing_wait != ASYNC_CLOSING_WAIT_NONE)) {
    	TRACE;
        et = jiffies + info->closing_wait;
        tty_wait_until_sent(tty, info->closing_wait);
        cnt = 0;
        while((timeout = et - jiffies) > 0) {
            if((ret = advtty_wait_oqueue(info,timeout)) == 0) {
                if(++cnt >= 3) {
                    break;
                }
            }
            else if (ret < 0)
	    {
		break;
		}
            else
	    {
		cnt = 0;
		}
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(HZ / 100);
        }
    }
    TRACE;
    advtty_flush_buffer(tty);
    TRACE;
    advtty_ldisc_flush_buffer(tty);
    TRACE;
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    if(tty->ldisc.flush_buffer)
    {	
	tty->ldisc.flush_buffer(tty);
    }
#else
    if(tty->ldisc.ops->flush_buffer)
    {
	tty->ldisc.ops->flush_buffer(tty);
    }
#endif
#else
    if(tty->ldisc->ops->flush_buffer)
    {
	tty->ldisc->ops->flush_buffer(tty);
    }
#endif
    TRACE;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
    if (kvcom_task[line])
    	kthread_stop(kvcom_task[line]);
    kvcom_task[line] = NULL;
#endif

    advtty_shutdown(info);
    tty->closing = 0;
    down(&nd->semaphore);
    nd->flag  &= ~ADVTTY_NET_NODE_OPENED;

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    /*if(kvcom_pid[line]){
    	kill_proc(kvcom_pid[line], SIGTERM, 1);
	kvcom_pid[line] = 0;
    }*/
    kvcom_pid[line] = 0;
#endif

    up(&nd->semaphore);
    AD_MOD_DEC;
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
//jinxin added to support linux-2.4.x
#ifndef wait_event_interruptible_timeout
#define __wait_event_interruptible_timeout(wq, condition, ret)		\
do {									\
	wait_queue_t __wait;						\
	init_waitqueue_entry(&__wait, current);				\
									\
	add_wait_queue(&wq, &__wait);					\
	for (;;) {							\
		set_current_state(TASK_INTERRUPTIBLE);			\
		if (condition)						\
			break;						\
		if (!signal_pending(current)) {				\
			ret = schedule_timeout(ret);			\
			if (!ret)					\
				break;					\
			continue;					\
		}							\
		ret = -ERESTARTSYS;					\
		break;							\
	}								\
	current->state = TASK_RUNNING;					\
	remove_wait_queue(&wq, &__wait);				\
} while (0)

#define wait_event_interruptible_timeout(wq, condition, timeout)	\
({									\
	long __ret = timeout;						\
	if (!(condition))						\
		__wait_event_interruptible_timeout(wq, condition, __ret); \
	__ret;								\
})
#endif
#endif
//jinxin added end



static int advtty_sendto_advttyp_thread(void *__tty)
{
    struct tty_struct* tty = (struct tty_struct*)__tty; 
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    struct nd_struct  	*nd;
    unsigned long        flags;
    int cnt, reval;
    int waitsec;
    if(!info){
	DBGPRINT(AD_DEBUG_ERROR, "port is not initialed\n");
        return -ENXIO;
    }

    if(!tty || !info->xmit_buf /*|| !advar_tmp_buf*/){
	DBGPRINT(AD_DEBUG_ERROR, "port is not initialed or xmit_buf is NULL\n");
        return -ENXIO;
    }
    if(!(nd = info->net_node)){
	DBGPRINT(AD_DEBUG_ERROR, "port is not initialed\n");
        return -ENXIO;
    }

    while(nd->flag & ADVTTY_NET_NODE_OPENED)
    {
again:
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
	if(kthread_should_stop()){
		DBGPRINT(AD_DEBUG_ERROR, "Port [%d] :  port closed, exit\n", info->port);
		break;	
	}
#endif
	if(nd->write_flags)
	{
    		if(info->x_char){//send xon/xoff char
			DOWN(info->nltx_lock, flags);
			reval = send_to_user(tty, NETLINK_KSEND_DATA, (unsigned char *)&info->x_char, 1);
			if(reval <= 0){
        			UP(info->nltx_lock, flags);
				DBGPRINT(AD_DEBUG_ERROR, "port [%d] : send x_char failed, oops\n", info->port);
				goto again;
			}
			nd->write_flags = 0;
        		UP(info->nltx_lock, flags);
			info->x_char = 0;
			nd->isxchar = 1;
			goto beginwait;

		}
		cnt = MIN(1024,  MIN(info->xmit_cnt,
                        SERIAL_XMIT_SIZE - info->xmit_tail));//be careful
		if(cnt <= 0){//there is no data in xmit buffer
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
			if(kthread_should_stop()){
				DBGPRINT(AD_DEBUG_ERROR, "Port [%d] : port closed, exit\n", info->port);
				break;	
			}
#else
			if(!(nd->flag & ADVTTY_NET_NODE_OPENED)){
				DBGPRINT(AD_DEBUG_ERROR, "Port [%d] : port closed, exit\n", info->port);
				break;
			}
#endif
			//DBGPRINT(AD_DEBUG_ERROR, "Port [%d] : no data in xmit buffer\n", info->port);
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    			if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       				(tty->ldisc.write_wakeup))
        			(*(tty->ldisc.write_wakeup))(tty);
#else
    			if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       				(tty->ldisc.ops->write_wakeup))
        			(*(tty->ldisc.ops->write_wakeup))(tty);
#endif
#else
    			if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       				(tty->ldisc->ops->write_wakeup))
        			(*(tty->ldisc->ops->write_wakeup))(tty);	
#endif
    			wake_up_interruptible(&tty->write_wait);
			set_current_state(TASK_INTERRUPTIBLE);
			schedule_timeout (HZ/20);
			goto again;	
		}
		DOWN(info->nltx_lock, flags);
		reval = send_to_user(tty, NETLINK_KSEND_DATA, info->xmit_buf + info->xmit_tail, cnt);
    		if(reval <= 0){
        		UP(info->nltx_lock, flags);
			DBGPRINT(AD_DEBUG_ERROR, "port [%d] : write data failed, oops\n", info->port);
			goto again;
		}
		nd->write_flags = 0;
        	UP(info->nltx_lock, flags);
		goto beginwait;
	}
beginwait:
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
	if(kthread_should_stop()){
		DBGPRINT(AD_DEBUG_ERROR, "Port [%d] :  port closed, exit\n", info->port);
		break;	
	}
#else
	if(!(nd->flag & ADVTTY_NET_NODE_OPENED)){
		DBGPRINT(AD_DEBUG_ERROR, "Port [%d] :  port closed, exit\n", info->port);
		break;		
	}
#endif
	wait_event_interruptible_timeout(nd->select_in_wait, nd->write_flags == 1, HZ/2);
	if(nd->write_flags)
		continue;
	else{
		goto beginwait;	
	}
    }
    return 0;
}



#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,10))
static int advtty_write(struct tty_struct * tty, int from_user,
                        const unsigned char * buf, int count)
#else
static int advtty_write(struct tty_struct * tty,
                        const unsigned char * buf, int count)
#endif
{
    int c, total = 0, line;//ret;
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    struct nd_struct  	*nd;
    unsigned long        flags;
    unsigned char name[16];
    line = PORTNO(tty);
    sprintf(name, "kvcom%d", line);
    pid_t pid;
    if(!info){
	DBGPRINT(AD_DEBUG_ERROR, "port is not initialed\n");
        return -ENXIO;
    }

    if(!tty || !info->xmit_buf /*|| !advar_tmp_buf*/){
	DBGPRINT(AD_DEBUG_ERROR, "port is not initialed or xmit_buf is NULL\n");
        return 0;
    }
    if(!(nd = info->net_node)){
	DBGPRINT(AD_DEBUG_ERROR, "port is not initialed\n");
        return -ENXIO;
    }
    while(1) {
        c = MIN(count, MIN(SERIAL_XMIT_SIZE - info->xmit_cnt - 1,
                           SERIAL_XMIT_SIZE - info->xmit_head));//be careful
        if (c <= 0)
            break;
        memcpy(info->xmit_buf + info->xmit_head, buf, c);
        DOWN(info->tx_lock, flags);
        info->xmit_head = (info->xmit_head + c) & (SERIAL_XMIT_SIZE - 1);
        info->xmit_cnt += c;
        UP(info->tx_lock, flags);

        buf += c;
        count -= c;
	total += c;
    }
    //printk(KERN_INFO "debug 1: total is %d, xmit_cnt is %d\n", total, info->xmit_cnt);
    if(info->xmit_cnt) {
    	nd->tx_ready = 1;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
	if((!kvcom_task[line]) && (nd->flag & ADVTTY_NET_NODE_OPENED)){
		kvcom_task[line] = kthread_run(advtty_sendto_advttyp_thread, tty, name);
		if (!IS_ERR(kvcom_task[line])){
    			//printk(KERN_INFO "debug 2: creat thread succ %d\n", total);
			goto done;
		}
		else{
			DBGPRINT(AD_DEBUG_ERROR, "port [%d] advtty_write : can not create kthread\n", line);
			return -1;
		}
	}
#else
	if((!kvcom_pid[line]) && (nd->flag & ADVTTY_NET_NODE_OPENED)){
		pid = kernel_thread(advtty_sendto_advttyp_thread, tty, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
		if (pid > 0) {
			kvcom_pid[line] = pid;
			goto done;
		}
		else{
			DBGPRINT(AD_DEBUG_ERROR, "port [%d] advtty_write : can not create kthread\n", line);
			return -1;
	    	}
	}

#endif

    }
done:
    return total;
}

//put_char(tty, ch) or write(tty, &ch, 1)? I have no idea!
static int advtty_put_char(struct tty_struct * tty, unsigned char ch)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    struct nd_struct * nd;
    unsigned long flags;
    int line;
    unsigned char name[16];
    line = PORTNO(tty);
    sprintf(name, "kvcom%d", line);
    pid_t pid;

    if(!info)
        return 0;

    if(!tty || !info->xmit_buf)
        return 0;

    DOWN(info->tx_lock, flags);
    if(!info->xmit_buf) {
        UP(info->tx_lock, flags);
        return 0;
    }

    if(info->xmit_cnt >= SERIAL_XMIT_SIZE - 1) {
        UP(info->tx_lock, flags);
        return 0;
    }

    if (!(nd = info->net_node)) {
        UP(info->tx_lock, flags);
        return 0;
    }

    info->xmit_buf[info->xmit_head++] = ch;
    info->xmit_head &= SERIAL_XMIT_SIZE - 1;
    info->xmit_cnt++;

    UP(info->tx_lock, flags);

    if(info->xmit_cnt) {
    	nd->tx_ready = 1;
#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,0))
	if((!kvcom_task[line]) && (nd->flag & ADVTTY_NET_NODE_OPENED)){
		kvcom_task[line] = kthread_run(advtty_sendto_advttyp_thread, tty, name);
		if (!IS_ERR(kvcom_task[line])){
    			//printk(KERN_INFO "debug 2: creat thread succ %d\n", total);
			return 0;
		}
		else{
			DBGPRINT(AD_DEBUG_ERROR, "port [%d] advtty_write : can not create kthread\n", line);
			return -1;
		}
	}
#else
	if((!kvcom_pid[line]) && (nd->flag & ADVTTY_NET_NODE_OPENED)){
		pid = kernel_thread(advtty_sendto_advttyp_thread, tty, CLONE_FS | CLONE_FILES | CLONE_SIGHAND);
		if (pid > 0) {
			kvcom_pid[line] = pid;
			return 0;
		}
		else{
			DBGPRINT(AD_DEBUG_ERROR, "port [%d] advtty_write : can not create kthread\n", line);
			return -1;
	    	}
	}
#endif

    }
    return 0;
}

static void advtty_flush_chars(struct tty_struct * tty)
{
	return;
}

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
static void advtty_wait_until_sent(struct tty_struct * tty,int timeout)
{

    struct advtty_struct *info;

    if ((info = (struct advtty_struct *)tty->driver_data))
        advtty_wait_oqueue(info, timeout);
    return;
}
#endif

static int advtty_write_room(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    int	ret;

    if(!info)
        return 0;

    ret = SERIAL_XMIT_SIZE - info->xmit_cnt - 1;
    if(ret < 0)
        ret = 0;
    return(ret);
}

static int advtty_chars_in_buffer(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;

    if (!info)
        return (-EIO);
    return(info->xmit_cnt);

}

static void advtty_flush_buffer(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_flush_buffer \n", info->port);
    struct nd_struct * nd;
    char rsp_buffer[CF_CMDRSPLEN];
    int  rsp_length = sizeof(rsp_buffer);
    unsigned long flags;
    int send, len;
    if(!info)
    {
	return;
    }

    DOWN(info->tx_lock, flags);
    info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
    UP(info->tx_lock, flags);
    //jinxin commented out it
    if(waitqueue_active(&tty->write_wait))
    	wake_up_interruptible(&tty->write_wait);
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc.write_wakeup))
	{
        (*(tty->ldisc.write_wakeup))(tty);
	}
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc.ops->write_wakeup))
	{
        (*(tty->ldisc.ops->write_wakeup))(tty);
	}
#endif
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc->ops->write_wakeup))
	{
        (*(tty->ldisc->ops->write_wakeup))(tty);
	}
#endif

    if(!(nd=info->net_node))
    {
	return;
	}
    nd->tx_ready = 0;
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, DREQ_CMD_FLUSH) < 0)
    {
	return;
	}
    nd->cmd_buffer[2] = 1;
    nd->cmd_buffer[3] = DREQ_FLUSH_ALL_BUFFER;
    nd->cmd_ready = 1;


    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_flush_buffer> send command DREQ_CMD_FLUSH failed! \n", info->port);
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

    advtty_wait_command_completed(nd,
                                  DREQ_DEVICE_CMDSET,
                                  DREQ_CMD_FLUSH,
                                  ADVTTY_CMD_TIMEOUT,
                                  rsp_buffer,
                                  &rsp_length);
}

static void advtty_flush_txbuffer(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_flush_txbuffer \n", info->port);
    struct nd_struct * nd;
    char rsp_buffer[CF_CMDRSPLEN];
    int  rsp_length = sizeof(rsp_buffer);
    unsigned long flags;
    int send, len;
    if(!info)
    {
	return;
    }

    DOWN(info->tx_lock, flags);
    info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
    UP(info->tx_lock, flags);

    if(waitqueue_active(&tty->write_wait))
    	wake_up_interruptible(&tty->write_wait);
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc.write_wakeup))
	{
        (*(tty->ldisc.write_wakeup))(tty);
	}
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc.ops->write_wakeup))
	{
        (*(tty->ldisc.ops->write_wakeup))(tty);
	}
#endif
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc->ops->write_wakeup))
	{
        (*(tty->ldisc->ops->write_wakeup))(tty);
	}
#endif

    if(!(nd=info->net_node))
    {
	return;
	}
    nd->tx_ready = 0;
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, DREQ_CMD_FLUSH) < 0)
    {
	return;
	}
    nd->cmd_buffer[2] = 1;
    nd->cmd_buffer[3] = DREQ_FLUSH_TX_BUFFER;
    nd->cmd_ready = 1;

   if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_flush_txbuffer> send command DREQ_CMD_FLUSH failed! \n", info->port);
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;


    advtty_wait_command_completed(nd,
                                  DREQ_DEVICE_CMDSET,
                                  DREQ_CMD_FLUSH,
                                  ADVTTY_CMD_TIMEOUT,
                                  rsp_buffer,
                                  &rsp_length);
}

static void advtty_flush_rxbuffer(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_flush_rxbuffer \n", info->port);
    struct nd_struct * nd;
    char rsp_buffer[CF_CMDRSPLEN];
    int  rsp_length = sizeof(rsp_buffer);
    unsigned long flags;
    int send, len;
    if(!info)
    {
	return;
    }

    DOWN(info->tx_lock, flags);
    info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
    UP(info->tx_lock, flags);

    if(waitqueue_active(&tty->write_wait))
    	wake_up_interruptible(&tty->write_wait);
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc.write_wakeup))
	{
        (*(tty->ldisc.write_wakeup))(tty);
	}
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc.ops->write_wakeup))
	{
        (*(tty->ldisc.ops->write_wakeup))(tty);
	}
#endif
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
        (tty->ldisc->ops->write_wakeup))
	{
        (*(tty->ldisc->ops->write_wakeup))(tty);
	}
#endif

    if(!(nd=info->net_node))
    {
	return;
	}
    nd->tx_ready = 0;
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, DREQ_CMD_FLUSH) < 0)
    {
	return;
	}
    nd->cmd_buffer[2] = 1;
    nd->cmd_buffer[3] = DREQ_FLUSH_RX_BUFFER;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_flush_rxbuffer> send command DREQ_CMD_FLUSH failed! \n", info->port);
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;


    advtty_wait_command_completed(nd,
                                  DREQ_DEVICE_CMDSET,
                                  DREQ_CMD_FLUSH,
                                  ADVTTY_CMD_TIMEOUT,
                                  rsp_buffer,
                                  &rsp_length);
}

static void advtty_ldisc_flush_buffer(struct tty_struct * tty)
{
    struct advtty_struct * info = (struct advtty_struct *)tty->driver_data;
    DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_ldisc_flush_buffer \n", info->port);
    int send, len;
    struct nd_struct * nd;
    unsigned long flags;

    if(!info)
    {
	return;
	}

    DOWN(info->tx_lock, flags);
    info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
    UP(info->tx_lock, flags);
    //jinxin commented out it
    //wake_up_interruptible(&tty->write_wait);
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       (tty->ldisc.write_wakeup))
       {
	(*(tty->ldisc.write_wakeup))(tty);
	}
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       (tty->ldisc.ops->write_wakeup))
       {
	(*(tty->ldisc.ops->write_wakeup))(tty);
	}
#endif
#else
    if((tty->flags & (1 << TTY_DO_WRITE_WAKEUP)) &&
       (tty->ldisc->ops->write_wakeup))
       {
	(*(tty->ldisc->ops->write_wakeup))(tty);
	}
#endif
    if(!(nd = info->net_node))
    {
	return;
	}
    nd->tx_ready = 0;
    if(nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
    {
	return;
	}
    nd->cmd_buffer[0] = DREQ_DEVICE_CMDSET;
    nd->cmd_buffer[1] = DREQ_CMD_FLUSH;
    nd->cmd_buffer[2] = 1;
    nd->cmd_buffer[3] = DREQ_FLUSH_ALL_BUFFER;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
        len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_flush_buffer> send command DREQ_CMD_FLUSH failed! \n", info->port);
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ / 100);
#endif
    return;
}
#if (LINUX_VERSION_CODE >= VERSION_CODE(3,0,0))
static int advtty_ioctl(struct tty_struct * tty,
                        unsigned int cmd, unsigned long arg)
#else
static int advtty_ioctl(struct tty_struct * tty, struct file * file,
                        unsigned int cmd, unsigned long arg)
#endif
{
    int			error;
    struct advtty_struct *	info = (struct advtty_struct *)tty->driver_data;
    int			retval;
    struct async_icount	cprev, cnow;
    struct serial_icounter_struct *p_cuser;
    unsigned long 		templ;

    if(!info)
        return -ENODEV;

    if((cmd != TIOCGSERIAL) && (cmd != TIOCMIWAIT) && (cmd != TIOCGICOUNT)) {
        if(tty->flags & (1 << TTY_IO_ERROR)) {
            return(-EIO);
        }
    }
    
    switch(cmd) {
    case TCFLSH:
        if((retval = tty_check_change(tty)))
            return retval;

        switch(arg) {
        case TCIFLUSH:
	     //flushes data received but not read
#if 0
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
            if (tty->ldisc.flush_buffer)
                tty->ldisc.flush_buffer(tty);
#else
            if (tty->ldisc.ops->flush_buffer)
                tty->ldisc.ops->flush_buffer(tty);
#endif
#else
            if (tty->ldisc->ops->flush_buffer)
                tty->ldisc->ops->flush_buffer(tty);	
#endif

#endif
	     //flushes data received but not read
	    advtty_flush_rxbuffer(tty);
            break;
        case TCIOFLUSH:
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,31))
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,6,27))
           if (tty->ldisc.flush_buffer)
                tty->ldisc.flush_buffer(tty);
#else
           if (tty->ldisc.ops->flush_buffer)
                tty->ldisc.ops->flush_buffer(tty);
#endif
#else
            if (tty->ldisc->ops->flush_buffer)
                tty->ldisc->ops->flush_buffer(tty);
#endif
        case TCOFLUSH:
	    //flushes data written but not transmitted
            advtty_flush_txbuffer(tty);
            break;
        default:
            return -EINVAL;
        }
        return 0;
        
    case TCSBRK:
        retval = tty_check_change(tty);
        if(retval)
            return(retval);
        tty_wait_until_sent(tty, 0);
        if(!arg)
            advtty_send_break(info, HZ / 4);
        return(0);
        
    case TCSBRKP:	// support for POSIX tcsendbreak()
        retval = tty_check_change(tty);
        if(retval)
            return(retval);
        tty_wait_until_sent(tty, 0);
        advtty_send_break(info, arg ? arg*(HZ / 10) : HZ / 4);
        return(0);
        
    case TIOCGSOFTCAR:
        error = access_ok(VERIFY_WRITE, (void *)arg, 
                          sizeof(long)) ? 0 : -EFAULT;
        if(error)
            return(error);
        put_to_user(C_CLOCAL(tty) ? 1 : 0, (unsigned int *)arg);
        return 0;
        
    case TIOCSSOFTCAR:
        error = access_ok(VERIFY_READ, (void *)arg,
                          sizeof(long)) ? 0 : -EFAULT;
        if(error)
            return(error);
        get_from_user(templ,(unsigned int *)arg);
        arg = templ;
        tty->termios->c_cflag = ((tty->termios->c_cflag & ~CLOCAL) |
                                 (arg ? CLOCAL : 0));
        return(0);
        
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    case TIOCMGET:
        error = access_ok(VERIFY_WRITE, (void *)arg,
                          sizeof(unsigned int)) ? 0 : -EFAULT;
        if ( error )
            return(error);
        return(advtty_get_modem_info(info, (unsigned int *)arg));
        
    case TIOCMBIS:
    case TIOCMBIC:
    case TIOCMSET:
        return(advtty_set_modem_info(info, cmd, (unsigned int *)arg));
#endif

    case TIOCGSERIAL:
        error = access_ok(VERIFY_WRITE, (void *)arg,
                          sizeof(struct serial_struct)) ? 0 : -EFAULT;
        if(error)
            return(error);
        return(advtty_get_serial_info(info, (struct serial_struct *)arg));
        
    case TIOCSSERIAL:
        error = access_ok(VERIFY_READ, (void *)arg,
                          sizeof(struct serial_struct)) ? 0 : -EFAULT;
        if(error)
            return(error);
        return(advtty_set_serial_info(info, (struct serial_struct *)arg));
        
    case TIOCSERGETLSR:
        error = access_ok(VERIFY_WRITE, (void *)arg,
                          sizeof(unsigned int)) ? 0 : -EFAULT;
        if(error)
            return(error);
        else
            return(advtty_get_lsr_info(info, (unsigned int *)arg));

    case TIOCMIWAIT: {
        DECLARE_WAITQUEUE(wait, current);
        int ret;
        cprev = info->icount;
        add_wait_queue(&info->delta_msr_wait, &wait);
        while(1) {
            cnow = info->icount;
            set_current_state(TASK_INTERRUPTIBLE);
            if(((arg & TIOCM_RNG) && (cnow.rng != cprev.rng)) ||
               ((arg & TIOCM_DSR) && (cnow.dsr != cprev.dsr)) ||
               ((arg & TIOCM_CD) && (cnow.dcd != cprev.dcd)) ||
               ((arg & TIOCM_CTS) && (cnow.cts != cprev.cts))) {
                ret = 0;
                break;
            }
            if (signal_pending(current)) {
                ret = -ERESTARTSYS;
                break;
            }
            cprev = cnow;
            schedule();
        }
        current->state = TASK_RUNNING;
        remove_wait_queue(&info->delta_msr_wait, &wait);
        return ret;}
        
    case TIOCGICOUNT:
        error = access_ok(VERIFY_WRITE, (void *)arg,
                          sizeof(struct serial_icounter_struct)) ? 0: -EFAULT;
        if(error)
            return(error);
        cnow = info->icount;
        p_cuser = (struct serial_icounter_struct *)arg;
        
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        if(put_user(cnow.frame, &p_cuser->frame))
            return -EFAULT;
        if(put_user(cnow.brk, &p_cuser->brk))
            return -EFAULT;
        if(put_user(cnow.overrun, &p_cuser->overrun))
            return -EFAULT;
        if(put_user(cnow.buf_overrun, &p_cuser->buf_overrun))
            return -EFAULT;
        if(put_user(cnow.parity, &p_cuser->parity))
            return -EFAULT;
        if(put_user(cnow.rx, &p_cuser->rx))
            return -EFAULT;
        if(put_user(cnow.tx, &p_cuser->tx))
            return -EFAULT;
#endif

        put_to_user(cnow.cts, &p_cuser->cts);
        put_to_user(cnow.dsr, &p_cuser->dsr);
        put_to_user(cnow.rng, &p_cuser->rng);
        put_to_user(cnow.dcd, &p_cuser->dcd);
        return(0);

    case TCXONC:
        retval = tty_check_change(tty);
        if (retval)
            return(retval);
        switch(arg) {
        case TCOOFF:
            return advtty_setxon_xoff(info, DREQ_CMD_SETXOFF);
        case TCOON:
            return advtty_setxon_xoff(info, DREQ_CMD_SETXON);
        default:
            return -EINVAL;
        }
        return 0;
        
    default:
        return(-ENOIOCTLCMD);
    }
    return(0);
}

static void advtty_throttle(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    struct nd_struct *nd;

    if(!info)
        return;
    if(!(nd = info->net_node))
        return;
    nd->rx_ready = 0;
    return;
}

static void advtty_unthrottle(struct tty_struct * tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    struct nd_struct *nd;

    if(!info)
        return;
    if(!(nd = info->net_node))
        return;
    nd->rx_ready = 1;
    return;
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
static void advtty_set_termios(struct tty_struct * tty,
                               struct termios * old_termios)
#else
static void advtty_set_termios(struct tty_struct * tty,
                               struct ktermios * old_termios)
#endif
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    advtty_port_init(info, old_termios);
    return;
}

static void advtty_stop(struct tty_struct * tty)
{
	return;
}

static void advtty_start(struct tty_struct * tty)
{
	return;
}

static void advtty_hangup(struct tty_struct *tty)
{
    struct advtty_struct *info = (struct advtty_struct *)tty->driver_data;
    printk(KERN_INFO "port %d advtty_hangup\n", info->port);
    if(!info)
        return;

    if (info->flags & ASYNC_CLOSING)
        return;

    info->flags |=  ASYNC_CLOSING;
    advtty_shutdown(info);
    return;
}

static inline void advtty_check_modem_status(struct advtty_struct *info,
                                             int status)
{
    int	is_dcd_changed = 0;

    if((info->modem_status & UART_MSR_DSR) != (status & UART_MSR_DSR))
        info->icount.dsr++;
    if((info->modem_status & UART_MSR_DCD) != (status & UART_MSR_DCD)) {
        info->icount.dcd++;
        is_dcd_changed = 1;
    }
    if((info->modem_status & UART_MSR_CTS) != (status & UART_MSR_CTS))
        info->icount.cts++;
    info->modem_status = status;
    wake_up_interruptible(&info->delta_msr_wait);

    if((info->flags & ASYNC_CHECK_CD) && (is_dcd_changed)) {
        if(status & UART_MSR_DCD) {
            wake_up_interruptible(&info->open_wait);
        }
        else {
            set_bit(ADVTTY_EVENT_HANGUP,&info->event);
            ADQ_TASK(&info->tqueue);
        }
    }
    return;
}

static int advtty_block_til_ready(struct tty_struct *tty, struct file * filp,
                                  struct advtty_struct *info)
{
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
    DECLARE_WAITQUEUE(wait, current);
#else
    struct wait_queue wait = {current, NULL};
#endif
    TRACE;
    int			retval;
    int			do_clocal = 0;
    struct nd_struct	*nd;

    if(!(nd = info->net_node))
        return (-EIO);

    if(AD_TTY_DRV(subtype) == SERIAL_TYPE_CALLOUT) {
        if(info->flags & ASYNC_NORMAL_ACTIVE)
            return(-EBUSY);
        if((info->flags & ASYNC_CALLOUT_ACTIVE) &&
           (info->flags & ASYNC_SESSION_LOCKOUT) &&
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,6))
           (info->session != current->session))
#else

#if (LINUX_VERSION_CODE > VERSION_CODE(2,6,19))
           (info->session != pid_nr(AD_SESSION())))
#else
           (info->session != current->signal->session))	
#endif

#endif
            return(-EBUSY);
        if((info->flags & ASYNC_CALLOUT_ACTIVE) &&
           (info->flags & ASYNC_PGRP_LOCKOUT) &&
           (info->pgrp != AD_CGRP()))
            return(-EBUSY);
        info->flags |= ASYNC_CALLOUT_ACTIVE;
        return(0);
    }
    TRACE;
    if((filp->f_flags & O_NONBLOCK) ||
       (tty->flags & (1 << TTY_IO_ERROR))) {
        if(info->flags & ASYNC_CALLOUT_ACTIVE) {
            return(-EBUSY);
        }
        info->flags |= ASYNC_NORMAL_ACTIVE;
        return(0);
    }
    TRACE;
    if(info->flags & ASYNC_CALLOUT_ACTIVE) {
        if(info->normal_termios.c_cflag & CLOCAL)
            do_clocal = 1;
    }
    else {
        if(tty->termios->c_cflag & CLOCAL)
            do_clocal = 1;
    }
    TRACE;
    retval = 0;
    add_wait_queue(&info->open_wait, &wait);
    while(1) {
        set_current_state(TASK_INTERRUPTIBLE);
        if(tty_hung_up_p(filp) || (info->flags & ASYNC_CLOSING)) {
            if(!tty_hung_up_p(filp)) {
#ifdef SERIAL_DO_RESTART
                if(info->flags & ASYNC_HUP_NOTIFY) {
                    retval = -EAGAIN;
                }
                else
                    retval = -ERESTARTSYS;
#else
                retval = -EAGAIN;
#endif
            }
            break;
        }
        
        if(!(info->flags & ASYNC_CALLOUT_ACTIVE) &&
           !(info->flags & ASYNC_CLOSING) &&
            (do_clocal || (info->modem_status & UART_MSR_DCD)))
            break;
            
        if(signal_pending(current)) {
            retval = -EIO;
            break;
        }
        schedule();
    }
    current->state = TASK_RUNNING;
    remove_wait_queue(&info->open_wait, &wait);
    if(retval)
        return(retval);
    info->flags |= ASYNC_NORMAL_ACTIVE;
    return(0);
}

static int advtty_startup(struct advtty_struct * info,struct file *filp,struct tty_struct *tty)
{
	unsigned long page;
	struct	nd_struct * nd;
	char rsp_buffer[CF_CMDRSPLEN];
	int  rsp_length = sizeof(rsp_buffer);
	int	cnt = 0, send, len;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
	DECLARE_WAITQUEUE(wait, current);
#else
	struct wait_queue wait = {current, NULL };
#endif

	if(!(nd = info->net_node)) 
	{
		DBGPRINT(AD_DEBUG_ERROR, "info->net_node is null\n");
		return -EIO;
	}
	TRACE;
	add_wait_queue(&nd->initialize_wait, &wait);
#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
	while(set_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag)) {
#else
	while (test_and_set_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag)) {
#endif
		if(signal_pending(current)) 
		{
			DBGPRINT(AD_DEBUG_ERROR, "signal_pending break\n");
			break;
		}
		schedule();
	}
	current->state = TASK_RUNNING;
	remove_wait_queue(&nd->initialize_wait, &wait);
	info->tty = tty;
	if(signal_pending(current)) {
		DBGPRINT(AD_DEBUG_ERROR, "signal_pending occurred\n");
		if(waitqueue_active(&nd->initialize_wait))
			wake_up_interruptible(&nd->initialize_wait);
        	return -EIO;
	}
	if(info->flags & ASYNC_INITIALIZED) {
		clear_bit(ADVTTY_NET_DO_INITIALIZE,&nd->flag);
		if(waitqueue_active(&nd->initialize_wait))
		wake_up_interruptible(&nd->initialize_wait);
		return(0);
	}
	page = GET_FPAGE(GFP_KERNEL);
	if(!page) {
		clear_bit(ADVTTY_NET_DO_INITIALIZE,&nd->flag);
		if(waitqueue_active(&nd->initialize_wait))
			wake_up_interruptible(&nd->initialize_wait);
		DBGPRINT(AD_DEBUG_ERROR, "allocate page failed\n");
		return(-ENOMEM);
	}
	if(!(nd->flag & ADVTTY_NET_TTY_INUSED)) 
	{
		nd->cmd_buffer[0] = DREQ_FREE_CMDSET;
		advtty_wait_and_set_command(nd, DREQ_DRIVER_CMDSET, CMD_DRV_TTY_OPEN);
		nd->cmd_buffer[2] = 0;
		nd->cmd_ready = 1;
		TRACE;

		if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
			len = nd->do_session_recovery_len;
		else
			len = (int)nd->cmd_buffer[2] + 3;
		send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
		if(send <= 0)
		{
			DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_startup> send command CMD_DRV_TTY_OPEN failed! \n", info->port);
			goto startup_err;
		}
		nd->cmd_buffer[0] = 0;
		nd->cmd_ready = 0;

		
		TRACE;
      		if(advtty_wait_command_completed(nd,
                                         DREQ_DRIVER_CMDSET,
                                         CMD_DRV_TTY_OPEN,
                                         ADVTTY_CMD_TIMEOUT,
                                         rsp_buffer,
                                         &rsp_length) != 0) 
		{
            		DBGPRINT(AD_DEBUG_ERROR, "Port %d: wait for CMD_DRV_TTY_OPEN response failed\n", info->port);
            		advtty_wait_and_set_command(nd, DREQ_DRIVER_CMDSET, CMD_DRV_TTY_CLOSE);
            		nd->cmd_buffer[2] = 0;
            		nd->cmd_ready = 1;
			TRACE; 

			if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
				len = nd->do_session_recovery_len;
			else
				len = (int)nd->cmd_buffer[2] + 3;
			send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
	      		if(send <= 0)
	        	{
				DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_startup>  send command CMD_DRV_TTY_CLOSE failed! \n", info->port);
		 		goto startup_err;
	    		}
			nd->cmd_buffer[0] = 0;
			nd->cmd_ready = 0;


            		advtty_wait_command_completed(nd,
                                          DREQ_DRIVER_CMDSET,
                                          CMD_DRV_TTY_CLOSE,
                                          ADVTTY_CMD_TIMEOUT,
                                          rsp_buffer,
                                          &rsp_length);
            		goto startup_err;
        	}
		TRACE;
		nd->flag |= ADVTTY_NET_TTY_INUSED;
	}
	else 
	{
		while((nd->cmd_ready == 1) && (cnt++ < 10)) 
		{
			current->state = TASK_INTERRUPTIBLE;
			schedule_timeout(HZ / 100);
		}
	}
	TRACE;
	nd->flag &= ~ADVTTY_NET_DO_SESSION_RECOVERY;
	info->modem_status = 0;
	info->modem_control = 0;
	if(info->tty->termios->c_cflag & CBAUD)
		info->modem_control = UART_MCR_DTR | UART_MCR_RTS;
	if(advtty_port_init(info, 0) != 0) 
	{
		DBGPRINT(AD_DEBUG_ERROR, "Port %d: advtty_port_init() failed\n", info->port);
		goto startup_err;
	}
	if(info->xmit_buf)
		free_page(page);
	else
		info->xmit_buf = (unsigned char *)page;

#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
	if(info->tty)
		clear_bit(TTY_IO_ERROR, &info->tty->flags);
#else
	if(info->tty)
		test_and_clear_bit(TTY_IO_ERROR, &info->tty->flags);
#endif

	info->xmit_cnt = info->xmit_head = info->xmit_tail = 0;
	info->flags |= ASYNC_INITIALIZED;
	clear_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag);
	if(waitqueue_active(&nd->initialize_wait))
		wake_up_interruptible(&nd->initialize_wait);
	return(0);
    
startup_err:
	free_page(page);
	clear_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag);
	if(waitqueue_active(&nd->initialize_wait))
		wake_up_interruptible( &nd->initialize_wait );
	return -EIO;
}

static void advtty_shutdown(struct advtty_struct * info)
{
    struct nd_struct * nd;
    unsigned long flags;
    if (!(nd = info->net_node)) {
        DBGPRINT(AD_DEBUG_ERROR, "nd is null\n");
        return;
    }

#if (LINUX_VERSION_CODE <  VERSION_CODE(2,1,0))
    while(set_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag)) {
#else
    while(test_and_set_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag)) {
#endif
	if(signal_pending(current))
            break;
        current->state = TASK_INTERRUPTIBLE;
        schedule_timeout(HZ/100);
    }
    
    if(!(nd->flag & ADVTTY_NET_TTY_INUSED)) {
	goto shutdown_ok;
    }

    DOWN(info->tx_lock, flags);
    if(info->xmit_buf) {
        free_page((unsigned long)info->xmit_buf);
        info->xmit_buf = 0;
    }
    UP(info->tx_lock, flags);

    if(info->tty) {
        set_bit(TTY_IO_ERROR, &info->tty->flags);
        advtty_unthrottle(info->tty);
    }

    if(!info->tty || (info->tty->termios->c_cflag & HUPCL)) {
        info->modem_control &= ~(UART_MCR_DTR | UART_MCR_RTS);
    }
    advtty_port_shutdown(info);
shutdown_ok:
    info->flags &= ~(ASYNC_NORMAL_ACTIVE | ASYNC_CALLOUT_ACTIVE | 
                     ASYNC_INITIALIZED | ASYNC_CLOSING);
    down(&info->rx_semaphore);
    info->tty = 0;
    up (&info->rx_semaphore);
    clear_bit(ADVTTY_NET_DO_INITIALIZE, &nd->flag);
    wake_up_interruptible(&info->open_wait);
    wake_up_interruptible(&info->close_wait);
    if(waitqueue_active(&nd->initialize_wait))
        wake_up_interruptible(&nd->initialize_wait);
    return;
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
static int advtty_port_init(struct advtty_struct *info,
                            struct termios *old_termios)
#else
static int advtty_port_init(struct advtty_struct *info,
                            struct ktermios *old_termios)
#endif
{ 
    TRACE;
    struct tty_struct* tty;
    tty = info->tty;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
    struct 	termios	*termio;
#else
	struct 	ktermios	*termio;
#endif
    int32_t     baud,mode;
    int		baudIndex,modem_status,send, len;
    struct 	nd_struct	*nd;
    char rsp_buffer[CF_CMDRSPLEN];
    int  rsp_length = sizeof(rsp_buffer);

    nd = info->net_node;
    if(!info->tty || !nd) {
        DBGPRINT(AD_DEBUG_ERROR, "info->tty or nd is null\n");
        return -EIO;
    }
    if(!(termio = info->tty->termios)) {
        DBGPRINT(AD_DEBUG_ERROR, "info->tty->termios is null\n");
        return -EIO;
    }

    mode = termio->c_cflag & CSIZE;
    if(mode == CS5)
        mode = DREQ_IOCTL_BITS5;
    else if(mode == CS6)
        mode = DREQ_IOCTL_BITS6;
    else if(mode == CS7)
        mode = DREQ_IOCTL_BITS7;
    else if(mode == CS8)
        mode = DREQ_IOCTL_BITS8;

    if(termio->c_cflag & CSTOPB)
        mode |= DREQ_IOCTL_STOP2;
    else
        mode |= DREQ_IOCTL_STOP1;

    if(termio->c_cflag & PARENB) {
#ifdef CMSPAR
        if(termio->c_cflag & CMSPAR)
            if(termio->c_cflag & PARODD)
                mode |= DREQ_IOCTL_MARK;
            else
                mode |= DREQ_IOCTL_SPACE;
        else
#endif
            if(termio->c_cflag & PARODD)
                mode |= DREQ_IOCTL_ODD;
            else
                mode |= DREQ_IOCTL_EVEN;
    }
    else
        mode |= DREQ_IOCTL_NONE;


    switch(termio->c_cflag & (CBAUD|CBAUDEX)) {
    case B921600:
        baud = 921600L;
        baudIndex = DREQ_IOCTL_B921600;
        break;
    case B460800:
        baud = 460800;
        baudIndex = DREQ_IOCTL_B460800;
        break;
    case B230400:
        baud = 230400L;
        baudIndex = DREQ_IOCTL_B230400;
        break;
    case B115200:
        baud = 115200L;
        baudIndex = DREQ_IOCTL_B115200;
        break;
    case B57600:
        baud = 57600L;
        baudIndex = DREQ_IOCTL_B57600;
        break;
    case B38400:
        baud = 38400L;
        baudIndex = DREQ_IOCTL_B38400;
	
	 if(info->flags & ASYNC_SPD_CUST)
    	{		
		if (  info->custom_divisor > 0 
		   && ADV_BAUDBASE%(info->custom_divisor) == 0)
		{
			switch (ADV_BAUDBASE / info->custom_divisor)
			{
			case 300:
				baud = 300L;
            			baudIndex = DREQ_IOCTL_B300;
				break;
			case 600:
				baud = 600L;
            			baudIndex = DREQ_IOCTL_B600;
				break;
			case 1200:
				baud = 1200L;
            			baudIndex = DREQ_IOCTL_B1200;
				break;
			case 2400:
				baud = 2400L;
            			baudIndex = DREQ_IOCTL_B2400;
				break;
			case 4800:
				baud = 4800L;
            			baudIndex = DREQ_IOCTL_B4800;
				break;
			case 7200:
				baud = 7200L;
            			baudIndex = DREQ_IOCTL_B7200;
				break;
			case 9600:
				baud = 9600L;
            			baudIndex = DREQ_IOCTL_B9600;
				break;
			case 19200:
				baud = 19200L;
            			baudIndex = DREQ_IOCTL_B19200;
				break;
			case 57600:
				baud = 57600L;
            			baudIndex = DREQ_IOCTL_B57600;

				break;
			case 115200:
				baud = 115200L;
            			baudIndex = DREQ_IOCTL_B115200;	
				break;
			case 230400:
				 baud = 230400L;
           			 baudIndex = DREQ_IOCTL_B230400;
				break;
			case 460800:
				baud = 460800L;
            			baudIndex = DREQ_IOCTL_B460800;
				break;
			case 921600:
				baud = 921600L;
            			baudIndex = DREQ_IOCTL_B921600;
				break;
			case 150:
				baud = 150L;
            			baudIndex = DREQ_IOCTL_B150;
				break;
			case 134:
				baud = 134L;
				baudIndex = DREQ_IOCTL_B134;
				break;
			case 110:
				baud = 110L;
            			baudIndex = DREQ_IOCTL_B110;
				break;
			case 75:
				baud = 75L;
            			baudIndex = DREQ_IOCTL_B75;
				break;	
			case 50:
				baud = 50L;
            			baudIndex = DREQ_IOCTL_B50;
				break;
			};
		}
	 	else
   		{
		//jinxin added to support 110
			switch (ADV_BAUDBASE / info->custom_divisor)
			{
				case 110:
					baud = 110L;
            				baudIndex = DREQ_IOCTL_B110;
					break;
			}
		//wrong input value... do nothing.
   		}

    	}

        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI) {
            baud = 57600L;
            baudIndex = DREQ_IOCTL_B57600;
        }
        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI) {
            baud = 115200L;
            baudIndex = DREQ_IOCTL_B115200;
        }

#ifdef ASYNC_SPD_SHI
        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI) {
            baud = 230400L;
            baudIndex = DREQ_IOCTL_B230400;
        }
#endif

#ifdef ASYNC_SPD_WARP
        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP) {
            baud = 460800L;
            baudIndex = DREQ_IOCTL_B460800;
        }
#endif
	        break;
    case B19200:
        baud = 19200L;
        baudIndex = DREQ_IOCTL_B19200;
        break;
    case B9600:
        baud = 9600L;
        baudIndex = DREQ_IOCTL_B9600;
        break;
    case B4800:
        baud = 4800L;
        baudIndex = DREQ_IOCTL_B4800;
        break;
    case B2400:
        baud = 2400L;
        baudIndex = DREQ_IOCTL_B2400;
        break;
    case B1800:
        baud = 1800L;
        baudIndex = 0xff;
        break;
    case B1200:
        baud = 1200L;
        baudIndex = DREQ_IOCTL_B1200;
        break;
    case B600:
        baud = 600L;
        baudIndex = DREQ_IOCTL_B600;
        break;
    case B300:
        baud = 300L;
        baudIndex = DREQ_IOCTL_B300;
        break;
    case B200:
        baud = 200L;
        baudIndex = 0xff;
        break;
    case B150:
        baud = 150L;
        baudIndex = DREQ_IOCTL_B150;
        break;
    case B134:
        baud = 134L;
        baudIndex = DREQ_IOCTL_B134;
        break;
    case B110:
        baud = 110L;
        baudIndex = DREQ_IOCTL_B110;
        break;
    case B75:
        baud = 75L;
        baudIndex = DREQ_IOCTL_B75;
        break;
    case B50:
        baud = 50L;
        baudIndex = DREQ_IOCTL_B50;
        break;
    default:
        baud = 0;
        baudIndex = 0xff;
    }
    if(baud > 921600L) {
        termio->c_cflag &= ~(CBAUD|CBAUDEX);
        termio->c_cflag |= old_termios->c_cflag & (CBAUD|CBAUDEX);
    }
    nd->cmd_buffer[2] = 8;

    // baud rate
    nd->cmd_buffer[3] = baudIndex;
    
    // mode
    nd->cmd_buffer[4] = mode;
    
    // line control
    if(info->modem_control & UART_MCR_DTR)
        nd->cmd_buffer[5] = 1;
    else
        nd->cmd_buffer[5] = 0;
    if(info->modem_control & UART_MCR_RTS)
        nd->cmd_buffer[6] = 1;
    else
        nd->cmd_buffer[6] = 0;

    // flow control
    if(termio->c_cflag & CRTSCTS) {
        nd->cmd_buffer[7] = 1;
        nd->cmd_buffer[8] = 1;
    }
    else if(termio->c_cflag & CDTRDSR){
        	nd->cmd_buffer[7] = 2;
          nd->cmd_buffer[8] = 2;
    }
    else {
        nd->cmd_buffer[7] = 0;
        nd->cmd_buffer[8] = 0;
    }
    
    if(termio->c_iflag & IXON) {
        nd->cmd_buffer[9] = 1;
    }
    else {
        nd->cmd_buffer[9] = 0;
    }
    
    if(termio->c_iflag & IXOFF) {
        nd->cmd_buffer[10] = 1;
    }
    else {
        nd->cmd_buffer[10] = 0;
    }
    
    nd->cmd_ready = 1;
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, 
                                   DREQ_CMD_PORT_INIT) < 0) {
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: set DREQ_CMD_PORT_INIT failed\n", info->port);
        return (-EIO);
    }

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_port_init>  send command DREQ_CMD_PORT_INIT failed\n", info->port);
	return EAGAIN;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;


    if(advtty_wait_command_completed(nd,
                                     DREQ_DEVICE_CMDSET,
                                     DREQ_CMD_PORT_INIT,
                                     ADVTTY_CMD_TIMEOUT,
                                     rsp_buffer,
                                     &rsp_length)) {
	TRACE;
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: wait DREQ_CMD_PORT_INIT response failed\n", info->port);
        return(-EIO);
    }
    TRACE;
    if(rsp_length != 6) {
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: invalid DREQ_CMD_PORT_INIT response1\n", info->port);
        return(-EIO);
    }
    TRACE;
    if(rsp_buffer[2] != 3) {
        DBGPRINT(AD_DEBUG_ERROR, "Port %d: invalid DREQ_CMD_PORT_INIT response2\n", info->port);
        return(-EIO);
    }
    TRACE;
    modem_status = 0;

    if (((unsigned char)rsp_buffer[3] == 0xff) &&
        ((unsigned char)rsp_buffer[4] == 0xff) &&
        ((unsigned char)rsp_buffer[5] == 0xff)) {
        termio->c_cflag &= ~(CBAUD|CBAUDEX);
        termio->c_cflag |= old_termios->c_cflag & (CBAUD|CBAUDEX);
    }
    else {
        if(rsp_buffer[3])
            modem_status |= UART_MSR_DSR;
        if(rsp_buffer[4])
            modem_status |= UART_MSR_CTS;
        if(rsp_buffer[5]) {
            modem_status |= UART_MSR_DCD;
        }
    }
    TRACE;
    advtty_check_modem_status(info,modem_status);
    TRACE;
    if((baudIndex == 0xff) && (baud != 0)) {
        if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET,
                                       DREQ_CMD_SETBAUD) < 0) {
            DBGPRINT(AD_DEBUG_ERROR, "set DREQ_CMD_SETBAUD failed\n");
            return(-EIO);
        }
        nd->cmd_buffer[2] = 4;
        memcpy(&nd->cmd_buffer[3],&baud,4);
        nd->cmd_ready = 1;

	if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
		len = nd->do_session_recovery_len;
   	else
 		len = (int)nd->cmd_buffer[2] + 3;
   	send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    	if(send <= 0)
    	{
		DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_port_init>  send command DREQ_CMD_SETBAUD failed\n", info->port);
		return EAGAIN;
    	}
    	nd->cmd_buffer[0] = 0;
    	nd->cmd_ready = 0;


        rsp_length = sizeof(rsp_buffer);
        if(advtty_wait_command_completed(nd,
                                         DREQ_DEVICE_CMDSET,
                                         DREQ_CMD_SETBAUD,
                                         ADVTTY_CMD_TIMEOUT,
                                         rsp_buffer,
                                         &rsp_length)) {
    	    TRACE;
            DBGPRINT(AD_DEBUG_ERROR, "wait DREQ_CMD_SETBAUD response failed\n");
            return(-EIO);
        }
   	TRACE;
        if(rsp_length != 4) {
            DBGPRINT(AD_DEBUG_ERROR, "invalid DREQ_CMD_SETBAUD response1\n");
            return(-EIO);
        }
    	TRACE;
        if((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K')) {
            DBGPRINT(AD_DEBUG_ERROR, "invalid DREQ_CMD_SETBAUD response2\n");
            return(-EIO);
        }

    }
    TRACE;
    if(termio->c_iflag & (IXON | IXOFF)) {
        if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, 
                                       DREQ_CMD_XONXOFF)) {
            DBGPRINT(AD_DEBUG_ERROR, "set DREQ_CMD_XONXOFF failed\n");
            return -EIO;
        }
        nd->cmd_buffer[2] = 2;
        nd->cmd_buffer[3] = termio->c_cc[VSTART];
        nd->cmd_buffer[4] = termio->c_cc[VSTOP];
        nd->cmd_ready = 1;

	if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
		len = nd->do_session_recovery_len;
   	else
 		len = (int)nd->cmd_buffer[2] + 3;
   	send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    	if(send <= 0)
    	{
		DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_port_init>  send command DREQ_CMD_XONXOFF failed \n", info->port);
		return EBUSY;
    	}
    	nd->cmd_buffer[0] = 0;
    	nd->cmd_ready = 0;


        rsp_length = sizeof(rsp_buffer);
        if(advtty_wait_command_completed(nd,
                                         DREQ_DEVICE_CMDSET,
                                         DREQ_CMD_XONXOFF,
                                         ADVTTY_CMD_TIMEOUT,
                                         rsp_buffer,
                                         &rsp_length)) {
    	    TRACE;
            DBGPRINT(AD_DEBUG_ERROR, "wait DREQ_CMD_XONXOFF response failed\n");
            return(-EIO);
        }
	TRACE;
        if(rsp_length != 4) {
            DBGPRINT(AD_DEBUG_ERROR, "invalid DREQ_CMD_XONXOFF response1\n");
            return(-EIO);
        }
        TRACE;
        if((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K')) {
            DBGPRINT(AD_DEBUG_ERROR, "invalid DREQ_CMD_XONXOFF response2\n");
            return(-EIO);
        }

    }
    TRACE;
    if(termio->c_cflag & CLOCAL) {
        info->flags &= ~ASYNC_CHECK_CD;
    }
    else
    {
        info->flags |= ASYNC_CHECK_CD;
    }
    if(!info->tty) {
        DBGPRINT(AD_DEBUG_ERROR, "info->tty is null\n");
        return -EIO;
    }
    TRACE;
    return(0);
}
static int advtty_port_shutdown(struct advtty_struct *info)
{
    DBGPRINT(AD_DEBUG_ERROR, "port %d advtty_port_shutdown\n", info->port);
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
    struct 	termios	*termio;
#else
	struct 	ktermios	*termio;
#endif
    struct 	nd_struct	*nd;
    struct tty_struct *tty;
    char rsp_buffer[CF_CMDRSPLEN];
    int  rsp_length = sizeof(rsp_buffer);
    int send, len;

    nd = info->net_node;
    if(!info->tty || !nd) {
        return -EIO;
    }
    tty = info->tty;
    if(!(termio = info->tty->termios)) {
        return -EIO;
    }

    nd->cmd_buffer[0] = DREQ_DRIVER_CMDSET;
    nd->cmd_buffer[1] = CMD_DRV_TTY_CLOSE;
    nd->cmd_buffer[2] = 0;
    nd->cmd_ready = 1;
    send = (int)nd->cmd_buffer[2] + 3;
    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_port_shutdown>  send command CMD_DRV_TTY_CLOSE failed \n", info->port);
	return EAGAIN;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;
    //jinxin moved it here
    nd->flag &= ~ADVTTY_NET_TTY_INUSED;
    advtty_wait_command_completed(nd,
                                  DREQ_DRIVER_CMDSET,
                                  CMD_DRV_TTY_CLOSE,
                                  ADVTTY_CMD_TIMEOUT,
                                  rsp_buffer,
                                  &rsp_length);

    //nd->flag &= ~ADVTTY_NET_TTY_INUSED;
    return(0);
}

static int advtty_get_serial_info(struct advtty_struct * info,
                                  struct serial_struct * retinfo)
{
    struct serial_struct	tmp;

    if(!retinfo)
        return(-EFAULT);
    memset(&tmp, 0, sizeof(tmp));
    tmp.type = info->type;
    tmp.line = info->port;
    tmp.flags = info->flags;
    tmp.close_delay = info->close_delay;
    tmp.closing_wait = info->closing_wait;
    tmp.custom_divisor = info->custom_divisor;
    tmp.baud_base = ADV_BAUDBASE;
    tmp.hub6 = 0;
    if(copy_to_user(retinfo, &tmp, sizeof(*retinfo)))
        return(-EFAULT);
    return(0);
}

static int advtty_set_serial_info(struct advtty_struct * info,
                                  struct serial_struct * new_info)
{
    struct serial_struct	new_serial;
    unsigned int flags;
    int retval = 0;

    if(!new_info)
        return(-EFAULT);
    if(copy_from_user(&new_serial, new_info, sizeof(new_serial)))
        return(-EFAULT);

    flags = info->flags & ASYNC_SPD_MASK;

#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
    if(!suser()) {
#else
    if(!capable(CAP_SYS_ADMIN)) {
#endif
        if((new_serial.close_delay != info->close_delay) ||
           ((new_serial.flags & ~ASYNC_USR_MASK) != 
            (info->flags & ~ASYNC_USR_MASK)))
            return(-EPERM);
        info->flags = ((info->flags & ~ASYNC_USR_MASK) |
                       (new_serial.flags & ASYNC_USR_MASK));
    }
    else {
        info->flags = ((info->flags & ~ASYNC_FLAGS) |
                       (new_serial.flags & ASYNC_FLAGS));
        info->close_delay = new_serial.close_delay * HZ/100;
        if(new_serial.closing_wait == ASYNC_CLOSING_WAIT_NONE)
            info->closing_wait = ASYNC_CLOSING_WAIT_NONE;
        else
            info->closing_wait = new_serial.closing_wait * HZ/100;
    }

    info->type = new_serial.type;
    info->xmit_fifo_size = 16;
    info->custom_divisor = new_serial.custom_divisor;
    if(info->flags & ASYNC_INITIALIZED) {
            retval = advtty_port_init(info, 0);
    }
    return(retval);
}

static int advtty_get_lsr_info(struct advtty_struct * info, unsigned int *value)
{
    unsigned int result = 0;

    if(advtty_wait_oqueue(info,0) == 0)
        result  = TIOCSER_TEMT;
    put_to_user(result, value);
    return(0);
}


static void advtty_start_break(struct nd_struct *nd)
{
    char rsp_buffer[CF_CMDRSPLEN];
    int	rsp_length = sizeof(rsp_buffer);
    int send, len;
    struct advtty_struct *info;
    struct tty_struct *tty;
    if(!nd)
        return;
    if(!nd->tty_node)
	return;
    info = nd->tty_node;
    if(!info->tty)
	return;
    tty = info->tty;

    advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, DREQ_CMD_START_BREAK);
    nd->cmd_buffer[2] = 0;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_start_break>  send command DREQ_CMD_START_BREAK failed\n", info->port);
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;


    if(advtty_wait_command_completed(nd,
                                     DREQ_DEVICE_CMDSET,
                                     DREQ_CMD_START_BREAK,
                                     ADVTTY_CMD_TIMEOUT,
                                     rsp_buffer,
                                     &rsp_length))
        return;
    if(rsp_length != 4)
        return;
    if ((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K'))
        return;
}

static void advtty_stop_break(struct nd_struct *nd)
{
    char rsp_buffer[CF_CMDRSPLEN];
    int	rsp_length = sizeof(rsp_buffer);
    int send, len;
    struct advtty_struct *info;
    struct tty_struct *tty;
    if(!nd)
        return;
    if(!nd->tty_node)
	return;
    info = nd->tty_node;
    if(!info->tty)
	return;
    tty = info->tty;

    advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, DREQ_CMD_STOP_BREAK);
    nd->cmd_buffer[2] = 0;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_stop_break>  send command DREQ_CMD_STOP_BREAK failed\n", info->port);;
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

    rsp_length = sizeof(rsp_buffer);
    if(advtty_wait_command_completed(nd,
                                     DREQ_DEVICE_CMDSET,
                                     DREQ_CMD_STOP_BREAK,
                                     ADVTTY_CMD_TIMEOUT,
                                     rsp_buffer,
                                     &rsp_length))
        return;
    if(rsp_length != 4)
        return;
    if((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K') )
        return;
}

static int advtty_break(struct tty_struct *ttyinfo, int break_state)
{
    struct advtty_struct * info;
    struct nd_struct * nd;

    if(!ttyinfo)
        return 0;

    info = (struct advtty_struct *)ttyinfo->driver_data;

    if(!info)
        return 0;

    if(!(nd = info->net_node))
        return 0;

    if(break_state == -1) {
        advtty_start_break(nd);
    }
    else {
        advtty_stop_break(nd);
    }
    return 0;
}

static void advtty_send_break(struct advtty_struct * info, int duration)
{
    struct nd_struct * nd;

    if(!(nd = info->net_node))
        return;

    advtty_start_break(nd);
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(duration);

    advtty_stop_break(nd);
    return;
}


#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,0))
static int advtty_tiocmget(struct tty_struct *tty, struct file *file)
{
    struct advtty_struct *info = (struct advtty_struct *) tty->driver_data;

    if(!info) {
        DBGPRINT(AD_DEBUG_ERROR, "info is null\n");
        return (-EINVAL);
    }

    if(PORTNO(tty) == CF_MAXPORTS)
        return (-ENOIOCTLCMD);
    if(tty->flags & (1 << TTY_IO_ERROR))
        return (-EIO);

    return ((info->modem_control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
           ((info->modem_control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
           ((info->modem_status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
           ((info->modem_status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
           ((info->modem_status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
           ((info->modem_status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
}

static int advtty_tiocmset(struct tty_struct *tty, struct file *file,
                           unsigned int set, unsigned int clear)
{
    struct advtty_struct *info = (struct advtty_struct *) tty->driver_data;
    struct	nd_struct	*nd;

    if(!info) {
        DBGPRINT(AD_DEBUG_ERROR, "info is null\n");
        return (-EINVAL);
    }

    if(!(nd = info->net_node))
        return(-EINVAL);

    if(PORTNO(tty) == CF_MAXPORTS)
        return (-ENOIOCTLCMD);
    if(tty->flags & (1 << TTY_IO_ERROR))
        return (-EIO);

    if(set & TIOCM_RTS)
        info->modem_control |= UART_MCR_RTS;
    if(set & TIOCM_DTR)
        info->modem_control |= UART_MCR_DTR;

    if(clear & TIOCM_RTS)
        info->modem_control &= ~UART_MCR_RTS;
    if(clear & TIOCM_DTR)
        info->modem_control &= ~UART_MCR_DTR;

    return advtty_linectrl(nd,info->modem_control);
}

#else

static int advtty_get_modem_info(struct advtty_struct * info,
                                 unsigned int *value)
{
    unsigned int	result;

    result = ((info->modem_control & UART_MCR_RTS) ? TIOCM_RTS : 0) |
             ((info->modem_control & UART_MCR_DTR) ? TIOCM_DTR : 0) |
             ((info->modem_status  & UART_MSR_DCD) ? TIOCM_CAR : 0) |
             ((info->modem_status  & UART_MSR_RI)  ? TIOCM_RNG : 0) |
             ((info->modem_status  & UART_MSR_DSR) ? TIOCM_DSR : 0) |
             ((info->modem_status  & UART_MSR_CTS) ? TIOCM_CTS : 0);
    put_to_user(result, value);
    return(0);
}

static int advtty_set_modem_info(struct advtty_struct * info, unsigned int cmd,
                                 unsigned int *value)
{
    int	error;
    unsigned int	arg;
    struct	nd_struct	*nd;

    if(!(nd = info->net_node))
        return(-EINVAL);
    error = access_ok(VERIFY_READ, value, sizeof(int)) ? 0 : -EFAULT;
    if(error)
        return(error);
    get_from_user(arg,value);
    switch (cmd) {
    case TIOCMBIS:
        if(arg & TIOCM_RTS )
            info->modem_control |= UART_MCR_RTS;
        if(arg & TIOCM_DTR )
            info->modem_control |= UART_MCR_DTR;
        break;
    case TIOCMBIC:
        if(arg & TIOCM_RTS )
            info->modem_control &= ~UART_MCR_RTS;
        if(arg & TIOCM_DTR )
            info->modem_control &= ~UART_MCR_DTR;
        break;
    case TIOCMSET:
        info->modem_control =
            ((info->modem_control & ~(UART_MCR_RTS | UART_MCR_DTR)) |
             ((arg & TIOCM_RTS) ? UART_MCR_RTS : 0) |
             ((arg & TIOCM_DTR) ? UART_MCR_DTR : 0));
        break;
    default:
        return(-EINVAL);
    }
    return (advtty_linectrl(nd,info->modem_control));
}
#endif

#if ((LINUX_VERSION_CODE > VERSION_CODE(2,6,15)) || ((LINUX_VERSION_CODE == VERSION_CODE(2,6,15)) && defined(FEDORA)))
static void tty_buffer_free(struct tty_struct *tty, struct tty_buffer *b)
{
    if (b->size >= 512)
        kfree(b);
    else
    {
        b->next = tty->buf.free;
        tty->buf.free = b;
    }
    return;
}
#endif


#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))  
static void advtty_flush_to_ldisc(void *private_)
{
    int reval;
    struct advtty_struct *	info = (struct advtty_struct *)private_;
#else
static void advtty_flush_to_ldisc(struct work_struct *work)
{
    int reval;
    struct advtty_struct *	info = 
    	container_of(work, struct advtty_struct, process_flip_tqueue);
#endif
    struct tty_struct *	tty;
    int		count;

#if ((LINUX_VERSION_CODE > VERSION_CODE(2,6,15)) || ((LINUX_VERSION_CODE == VERSION_CODE(2,6,15)) && defined(FEDORA)))
    struct tty_ldisc *disc;
    struct tty_buffer *tbuf, *head;
    unsigned long flags;
    unsigned char * fp;
    char * cp;
#else
    unsigned char	*cp;
    char		*fp;
#endif

    if(!info)
        goto done;
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    down(&info->rx_semaphore);
#endif
    tty = info->tty;
#if ((LINUX_VERSION_CODE > VERSION_CODE(2,6,15)) || ((LINUX_VERSION_CODE == VERSION_CODE(2,6,15)) && defined(FEDORA)))
    disc = tty_ldisc_ref(tty);
    if(disc == NULL)	/*  !TTY_LDISC */
        return;
#endif
    if(tty && (info->flags & ASYNC_INITIALIZED)) {
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
        if (tty->flip.buf_num) {
            cp = tty->flip.char_buf + TTY_FLIPBUF_SIZE;
            fp = tty->flip.flag_buf + TTY_FLIPBUF_SIZE;
            tty->flip.buf_num = 0;
            tty->flip.char_buf_ptr = tty->flip.char_buf;
            tty->flip.flag_buf_ptr = tty->flip.flag_buf;
        }
        else {
            cp = tty->flip.char_buf;
            fp = tty->flip.flag_buf;
            tty->flip.buf_num = 1;

            tty->flip.char_buf_ptr = tty->flip.char_buf + TTY_FLIPBUF_SIZE;
            tty->flip.flag_buf_ptr = tty->flip.flag_buf + TTY_FLIPBUF_SIZE;
        }
        count = tty->flip.count;
        tty->flip.count = 0;
        tty->ldisc.receive_buf(tty, cp, fp, count);
		reval = send_to_user(tty, NETLINK_USEND_DATA_OK, (unsigned char *)&count,sizeof(count));
        if(reval <= 0){
             DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_OK to advttyp failed 2\n", info->port);
        }
#else
        spin_lock_irqsave(&tty->buf.lock, flags);
        head = tty->buf.head;
        if (head != NULL) {
            tty->buf.head = NULL;
            for(;;) {
                count = head->commit - head->read;
                if(!count) {
                    if (head->next == NULL)
                        break;
                    tbuf = head;
                    head = head->next;
                    tty_buffer_free(tty, tbuf);
                    continue;
                }
                if (!tty->receive_room) {
                    schedule_delayed_work(&tty->buf.work, 1);
                    break;
                }
                if (count > tty->receive_room)
                    count = tty->receive_room;
                cp = head->char_buf_ptr + head->read;
                fp = head->flag_buf_ptr + head->read;
                head->read += count;
                spin_unlock_irqrestore(&tty->buf.lock, flags);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,6,27))
                disc->ops->receive_buf(tty, cp, fp, count);
#else
                disc->receive_buf(tty, cp, fp, count);
#endif
                spin_lock_irqsave(&tty->buf.lock, flags);
            }
            tty->buf.head = head;
        }
        spin_unlock_irqrestore(&tty->buf.lock, flags);
	reval = send_to_user(tty, NETLINK_USEND_DATA_OK, (unsigned char *)&count,sizeof(count));
        if(reval <= 0){
             DBGPRINT(AD_DEBUG_ERROR, "port [%d] send NETLINK_USEND_DATA_OK to advttyp failed 2\n", info->port);
        }
        tty_ldisc_deref(disc);
#endif
    }
#if ((LINUX_VERSION_CODE <= VERSION_CODE(2,6,15)) && !defined(FEDORA))
    up(&info->rx_semaphore);
#endif

done:
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,0))
    AD_MOD_DEC;
#endif

    return;
}

#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
static struct proc_dir_entry *advtty_create_proc_entry(const char * name, mode_t mode,
                                                       struct proc_dir_entry *parent)
{
    struct proc_dir_entry *ent = NULL;
    const char *fn = name;
    int len;

    if(!parent)
        goto out;
    len = strlen(fn);

    ent = kmalloc(sizeof(struct proc_dir_entry) + len + 1, GFP_KERNEL);
    if(!ent)
        goto out;
    memset(ent, 0, sizeof(struct proc_dir_entry));
    memcpy(((char *) ent) + sizeof(*ent), fn, len + 1);
    ent->name = ((char *) ent) + sizeof(*ent);
    ent->namelen = len;

    if(mode & S_IFDIR) {
        mode |= S_IRUGO | S_IXUGO;
        ent->ops = &proc_dir_inode_operations;
        ent->nlink = 2;
    }
    else {
        if(mode == 0)
            mode |= S_IFREG | S_IRUGO;
        ent->nlink = 1;
    }
    ent->mode = mode;

    proc_register_dynamic(parent, ent);

out:
    return ent;
}

static void advtty_remove_proc_entry ( struct proc_dir_entry *pde )
{
    if (!pde) return;

    proc_unregister(pde->parent, pde->low_ino);

    kfree(pde);
    return;
}

#else

static struct proc_dir_entry * 
    advtty_create_proc_entry(const char * name, const mode_t mode,
                             struct proc_dir_entry *parent)
{
    return( create_proc_entry( name, mode, parent ) );
}

static void advtty_remove_proc_entry( struct proc_dir_entry *pde )
{
    if (!pde) return;

    remove_proc_entry(pde->name, pde->parent);
    return;
}

#endif


static int advtty_wait_and_set_command(struct nd_struct * nd, char command_set,
                                       char command)
{
    unsigned long	et;

    if((command_set != DREQ_DRIVER_CMDSET) && 
       ((nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY) || 
        (nd->flag & ADVTTY_NET_NODE_DISCONNECTED))) {
        return (-1);
    }

    et = jiffies + ADVTTY_CMD_TIMEOUT;
    while(1) {
        down (&nd->cmd_semaphore);
        if(nd->cmd_buffer[0] == DREQ_FREE_CMDSET) {
            nd->cmd_buffer[0] = command_set;
            nd->cmd_buffer[1] = command;
            up(&nd->cmd_semaphore);
            return (0);
        }
        else if ((jiffies >= et) || signal_pending(current)) {
            nd->cmd_buffer[0] = command_set;
            nd->cmd_buffer[1] = command;
            up(&nd->cmd_semaphore);
            return (0);
        }
        else {
            up (&nd->cmd_semaphore);
            current->state = TASK_INTERRUPTIBLE;
            schedule_timeout(1);
        }
    }
}
//check if there is expected response data in nd->rsp_buffer until timeout
static int advtty_wait_command_completed(struct nd_struct *nd,
                                         char command_set,
                                         char command,
                                         long timeout,
                                         char * rsp_buf,
                                         int  * rsp_len)
{
    if(command == CMD_DRV_TTY_OPEN)
	DBGPRINT(AD_DEBUG_ERROR, "\n********CMD_DRV_TTY_OPEN*******\n");	
    if(command == DREQ_CMD_FLUSH)   
	DBGPRINT(AD_DEBUG_ERROR, "\n********DREQ_CMD_FLUSH*******\n");
    if(command == CMD_DRV_TTY_CLOSE)   
	DBGPRINT(AD_DEBUG_ERROR, "\n********CMD_DRV_TTY_CLOSE*******\n");
    if(command == DREQ_CMD_PORT_INIT)   
	DBGPRINT(AD_DEBUG_ERROR, "\n********DREQ_CMD_PORT_INIT*******\n");
    if(command == DREQ_CMD_XONXOFF)   
	DBGPRINT(AD_DEBUG_ERROR, "\n********DREQ_CMD_XONXOFF*******\n");	
    long et = 0;
    if((command_set != DREQ_DRIVER_CMDSET) && 
       ((nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY) || 
        (nd->flag & ADVTTY_NET_NODE_DISCONNECTED))) {
    	DBGPRINT(AD_DEBUG_ERROR, "#1#(cmd error)");
	return (-1);
    }
    if(*rsp_len <= 0){
    	DBGPRINT(AD_DEBUG_ERROR, "#2#(cmd len error)");
        return (-1);
    }
    while(1) {
        down(&nd->cmd_semaphore);
        if((nd->rsp_length) && 
           (nd->rsp_buffer[0] == command_set) && 
           (nd->rsp_buffer[1] == command)) {
            if(nd->rsp_length > *rsp_len) {
    		DBGPRINT(AD_DEBUG_ERROR, "#3#(cmd len error)");
            	up(&nd->cmd_semaphore);
                return (-1);
            }
	    TRACE;
	    DBGPRINT(AD_DEBUG_ERROR, "#5#(right)");
            *rsp_len = nd->rsp_length;
            memcpy(rsp_buf, nd->rsp_buffer, *rsp_len);
            nd->rsp_length = 0;
            up(&nd->cmd_semaphore);
            return (0);
        }
        else if (timeout > 0) {
            up(&nd->cmd_semaphore);
            if(signal_pending(current)) {
    		DBGPRINT(AD_DEBUG_ERROR, "#4#(signal interrupt)");
                return(-1);
            }
	    DBGPRINT(AD_DEBUG_ERROR, "#6#(wait)");
            if (timeout != MAX_SCHEDULE_TIMEOUT)
                et = jiffies + timeout;
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
            interruptible_sleep_on_timeout(&nd->cmd_rsp_wait,timeout);
#else
            current->timeout = timeout;
            interruptible_sleep_on(&nd->cmd_rsp_wait);

#endif
            if (timeout != MAX_SCHEDULE_TIMEOUT)
                timeout = et - jiffies;
        }
        else {
	    DBGPRINT(AD_DEBUG_ERROR, "#7#(timeout)");
            up(&nd->cmd_semaphore);
            return (-1);
        }
    }
}

static void advtty_process_notify(struct nd_struct * nd, char * rsp_buffer,
                                  int rsp_length)
{
    int	state;
    struct advtty_struct * info = nd->tty_node;

    if(!info)
        return;
    if(rsp_length != 5)
        return;
    if(rsp_buffer[2] & DREQ_NOTIFY_MSR_CHG) {
        state = 0;
        if(rsp_buffer[3] & 0x10)
            state |= UART_MSR_CTS;
        if(rsp_buffer[3] & 0x20)
            state |= UART_MSR_DSR;
        if(rsp_buffer[3] & 0x80)
            state |= UART_MSR_DCD;
        advtty_check_modem_status(info,state);
    }
    
    if(rsp_buffer[2] & DREQ_NOTIFY_BREAK) {
        struct tty_struct	*tty;

        down (&info->rx_semaphore);
        if(!(tty= info->tty)) {
            up (&info->rx_semaphore);
            return;
        }
        tty_insert_flip_char(tty, 0, TTY_BREAK);
        up (&info->rx_semaphore);
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        info->icount.rx ++;
#endif

#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
        info->icount.brk++;
#endif
        ADQ_TASK(&info->process_flip_tqueue);

        if(info->flags & ASYNC_SAK) {
            do_SAK(info->tty);
        }
    }
    
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
    if(rsp_buffer[2] & DREQ_NOTIFY_PARITY)
        info->icount.parity++;
    if(rsp_buffer[2] & DREQ_NOTIFY_FRAMING)
        info->icount.frame++;
    if((rsp_buffer[2] & DREQ_NOTIFY_SW_OVERRUN) ||
       (rsp_buffer[2] & DREQ_NOTIFY_HW_OVERRUN))
        info->icount.overrun++;
#endif

    return;
}

static void advtty_do_session_recovery(struct advtty_struct * info)
{
    printk(KERN_INFO "port [%d] tx: %d, rx: %d\n", info->port, info->icount.tx, info->icount.rx);
    struct tty_struct *	tty;
    struct nd_struct *	nd;	
    int send, len;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,6,20))
    struct termios * termio;
#else
	struct ktermios * termio;
#endif
    int32_t    		    baud,mode;
    int		baudIndex,index;

    tty = info->tty;
    nd = info->net_node;
    if(!tty || !nd)
        return;
    if(!(nd->flag & ADVTTY_NET_TTY_INUSED))
        return;
    if(info->flags & ASYNC_INITIALIZED) {
        if(!(termio = info->tty->termios))
            return;
    }
    else {
        if(!(termio = &info->normal_termios))
            return;
    }
    //send = send_to_user(tty, NETLINK_DEAL_DELAY_DATA, NULL, 0);
    //if(send < 0){
	//DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_do_session_recovery>  send NETLINK_WRITE_TRY_AGAIN failed\n", info->port);
    //}
    down(&nd->cmd_semaphore);
    mode = termio->c_cflag & CSIZE;
    if(mode == CS5)
        mode = DREQ_IOCTL_BITS5;
    else if(mode == CS6)
        mode = DREQ_IOCTL_BITS6;
    else if(mode == CS7)
        mode = DREQ_IOCTL_BITS7;
    else if(mode == CS8)
        mode = DREQ_IOCTL_BITS8;

    if(termio->c_cflag & CSTOPB)
        mode |= DREQ_IOCTL_STOP2;
    else
        mode |= DREQ_IOCTL_STOP1;

    if(termio->c_cflag & PARENB) {
        if(termio->c_cflag & PARODD)
            mode |= DREQ_IOCTL_ODD;
        else
            mode |= DREQ_IOCTL_EVEN;
    }
    else
        mode |= DREQ_IOCTL_NONE;

    switch(termio->c_cflag & (CBAUD | CBAUDEX)) {
    case B921600:
        baud = 921600L;
        baudIndex = DREQ_IOCTL_B921600;
        break;
    case B460800:
        baud = 460800;
        baudIndex = DREQ_IOCTL_B460800;
        break;
    case B230400:
        baud = 230400L;
        baudIndex = DREQ_IOCTL_B230400;
        break;
    case B115200:
        baud = 115200L;
        baudIndex = DREQ_IOCTL_B115200;
        break;
    case B57600:
        baud = 57600L;
        baudIndex = DREQ_IOCTL_B57600;
        break;
    case B38400:
        baud = 38400L;
        baudIndex = DREQ_IOCTL_B38400;

	if(info->flags & ASYNC_SPD_CUST)
    	{		
		if (  info->custom_divisor > 0 
		   && ADV_BAUDBASE%(info->custom_divisor) == 0)
		{
			switch (ADV_BAUDBASE / info->custom_divisor)
			{
			case 300:
				baud = 300L;
            			baudIndex = DREQ_IOCTL_B300;
				break;
			case 600:
				baud = 600L;
            			baudIndex = DREQ_IOCTL_B600;
				break;
			case 1200:
				baud = 1200L;
            			baudIndex = DREQ_IOCTL_B1200;
				break;
			case 2400:
				baud = 2400L;
            			baudIndex = DREQ_IOCTL_B2400;
				break;
			case 4800:
				baud = 4800L;
            			baudIndex = DREQ_IOCTL_B4800;
				break;
			case 7200:
				baud = 7200L;
            			baudIndex = DREQ_IOCTL_B7200;
				break;
			case 9600:
				baud = 9600L;
            			baudIndex = DREQ_IOCTL_B9600;
				break;
			case 19200:
				baud = 19200L;
            			baudIndex = DREQ_IOCTL_B19200;
				break;
			case 57600:
				baud = 57600L;
            			baudIndex = DREQ_IOCTL_B57600;

				break;
			case 115200:
				baud = 115200L;
            			baudIndex = DREQ_IOCTL_B115200;	
				break;
			case 230400:
				 baud = 230400L;
           			 baudIndex = DREQ_IOCTL_B230400;
				break;
			case 460800:
				baud = 460800L;
            			baudIndex = DREQ_IOCTL_B460800;
				break;
			case 921600:
				baud = 921600L;
            			baudIndex = DREQ_IOCTL_B921600;
				break;
			case 150:
				baud = 150L;
            			baudIndex = DREQ_IOCTL_B150;
				break;
			case 134:
				baud = 134L;
				baudIndex = DREQ_IOCTL_B134;
				break;
			case 110:
				baud = 110L;
            			baudIndex = DREQ_IOCTL_B110;
				break;
			case 75:
				baud = 75L;
            			baudIndex = DREQ_IOCTL_B75;
				break;	
			case 50:
				baud = 50L;
            			baudIndex = DREQ_IOCTL_B50;
				break;
			};
		}
	 	else
   		{
		//wrong input value... do nothing.
   		}

    	}

        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_HI) {
            baud = 57600L;
            baudIndex = DREQ_IOCTL_B57600;
        }
        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_VHI) {
            baud = 115200L;
            baudIndex = DREQ_IOCTL_B115200;
        }

#ifdef ASYNC_SPD_SHI
        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_SHI) {
            baud = 230400L;
            baudIndex = DREQ_IOCTL_B230400;
        }
#endif

#ifdef ASYNC_SPD_WARP
        if((info->flags & ASYNC_SPD_MASK) == ASYNC_SPD_WARP) {
            baud = 460800L;
            baudIndex = DREQ_IOCTL_B460800;
        }
#endif
        break;
    case B19200:
        baud = 19200L;
        baudIndex = DREQ_IOCTL_B19200;
        break;
    case B9600:
        baud = 9600L;
        baudIndex = DREQ_IOCTL_B9600;
        break;
    case B4800:
        baud = 4800L;
        baudIndex = DREQ_IOCTL_B4800;
        break;
    case B2400:
        baud = 2400L;
        baudIndex = DREQ_IOCTL_B2400;
        break;
    case B1800:
        baud = 1800L;
        baudIndex = 0xff;
        break;
    case B1200:
        baud = 1200L;
        baudIndex = DREQ_IOCTL_B1200;
        break;
    case B600:
        baud = 600L;
        baudIndex = DREQ_IOCTL_B600;
        break;
    case B300:
        baud = 300L;
        baudIndex = DREQ_IOCTL_B300;
        break;
    case B200:
        baud = 200L;
        baudIndex = 0xff;
        break;
    case B150:
        baud = 150L;
        baudIndex = DREQ_IOCTL_B150;
        break;
    case B134:
        baud = 134L;
        baudIndex = DREQ_IOCTL_B134;
        break;
    case B110:
        baud = 110L;
        baudIndex = DREQ_IOCTL_B110;
        break;
    case B75:
        baud = 75L;
        baudIndex = DREQ_IOCTL_B75;
        break;
    case B50:
        baud = 50L;
        baudIndex = 0xff;
        break;
        break;
    default:
        baud = 0;
        baudIndex = 0xff;
    }
    nd->cmd_buffer[2] = 8;

    // baud rate
    nd->cmd_buffer[3] = baudIndex;

    // mode
    nd->cmd_buffer[4] = mode;

    // line control
    if(info->modem_control & UART_MCR_DTR)
        nd->cmd_buffer[5] = 1;
    else
        nd->cmd_buffer[5] = 0;
    if(info->modem_control & UART_MCR_RTS)
        nd->cmd_buffer[6] = 1;
    else
        nd->cmd_buffer[6] = 0;

    // flow control
    if(info->flags & ASYNC_INITIALIZED) {
        if(termio->c_cflag & CRTSCTS) {
            nd->cmd_buffer[7] = 1;
            nd->cmd_buffer[8] = 1;
        }
        else if(termio->c_cflag & CDTRDSR){
        	nd->cmd_buffer[7] = 2;
          nd->cmd_buffer[8] = 2;
        }
        else {
            nd->cmd_buffer[7] = 0;
            nd->cmd_buffer[8] = 0;
        }
    }
    else {
        nd->cmd_buffer[7] = 0;
        nd->cmd_buffer[8] = 0;

    }

    if(termio->c_iflag & IXON)
        nd->cmd_buffer[9] = 1;
    else
        nd->cmd_buffer[9] = 0;
    if(termio->c_iflag & IXOFF)
        nd->cmd_buffer[10] = 1;
    else
        nd->cmd_buffer[10] = 0;
    nd->cmd_buffer[0] = DREQ_DEVICE_CMDSET;
    nd->cmd_buffer[1] = DREQ_CMD_PORT_INIT;
    index = 11;

    if((baudIndex == 0xff) && (baud != 0)) {
        nd->cmd_buffer[index + 0] = DREQ_CMD_SETBAUD;
        nd->cmd_buffer[index + 1] = 4;
        memcpy(&nd->cmd_buffer[index + 2], &baud, 4);
        index += 6;
    }

    if(termio->c_iflag & (IXON | IXOFF)) {
        nd->cmd_buffer[index + 0] = DREQ_CMD_XONXOFF;
        nd->cmd_buffer[index + 1] = 2;
        nd->cmd_buffer[index + 2] = termio->c_cc[VSTART];
        nd->cmd_buffer[index + 3] = termio->c_cc[VSTOP];
        index += 4;
    }
    
    nd->cmd_buffer[index + 0] = DREQ_CMD_LINECTRL;
    nd->cmd_buffer[index + 1] = 2;
    if(info->modem_control & UART_MCR_DTR)
        nd->cmd_buffer[index + 2] = 1;
    else
        nd->cmd_buffer[index + 2] = 0;
    if(info->modem_control & UART_MCR_RTS)
        nd->cmd_buffer[index + 3] = 1;
    else
        nd->cmd_buffer[index + 3] = 0;
    index += 4;
    nd->do_session_recovery_len = index;
    nd->flag |= ADVTTY_NET_DO_SESSION_RECOVERY;
    nd->cmd_ready = 1;
    up (&nd->cmd_semaphore);

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_do_session_recovery>  send command failed\n", info->port);
	return;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

/*
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,4,0))
    current->state = TASK_INTERRUPTIBLE;
    schedule_timeout(HZ / 10);
#endif
*/
    return;

}
//send DREQ_CMD_WAIT_OQUEUE command, and wait for the expected response(nd->wait_oqueue_responsed == 0)
static long advtty_wait_oqueue(struct advtty_struct * info, long timeout)
{
    struct	nd_struct	*nd;
    struct 	tty_struct 	*tty;
    int send, len;
    long	et = 0;
#if (LINUX_VERSION_CODE < VERSION_CODE(2,1,0))
    long	st;
#endif
    uint32_t    tout;

    if(!(nd = info->net_node))
        return (-EIO);
    tty = info->tty;
    if(!tty)
	return	(-EIO);
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, 
                                   DREQ_CMD_WAIT_OQUEUE) < 0)
        return (-EIO);
    if(timeout != MAX_SCHEDULE_TIMEOUT) {
        if(timeout < HZ / 10)
            timeout = HZ / 10;
        et = jiffies + timeout;
    }

    if(timeout != MAX_SCHEDULE_TIMEOUT)
        tout = (uint32_t)timeout;
    else
        tout = 0x7FFFFFFF;

    nd->cmd_buffer[2] = 4;
    memcpy(&nd->cmd_buffer[3],(void *)&tout,4);
    nd->wait_oqueue_responsed = 0;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_wait_oqueue>  send command DREQ_CMD_WAIT_OQUEUE failed\n", info->port);;
	return EBUSY;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

    while(nd->cmd_ready == 1) {
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
        interruptible_sleep_on_timeout(&nd->cmd_rsp_wait, 1);
#else
        current->timeout = 1;
        interruptible_sleep_on(&nd->cmd_rsp_wait);

#endif
        if(timeout != MAX_SCHEDULE_TIMEOUT) {
            if (jiffies > et)
                return (-EIO);
        }

    }
    if(timeout != MAX_SCHEDULE_TIMEOUT)
        timeout += 10;
    nd->cmd_buffer[0] = 0;
    //wait for the response
    do {
        if(nd->wait_oqueue_responsed == 0) {
#if (LINUX_VERSION_CODE >= VERSION_CODE(2,1,0))
            timeout = interruptible_sleep_on_timeout(&nd->cmd_rsp_wait,
                                                     timeout);
#else
            st = jiffies;
            current->timeout = timeout;
            interruptible_sleep_on(&nd->cmd_rsp_wait);
            timeout = jiffies - st;

#endif
            if (nd->wait_oqueue_responsed) {
                return (nd->oqueue);
            }
        }
        else {
            return (nd->oqueue);
        }
    } while(timeout > 0);
    return (-EIO);
}

static int advtty_linectrl(struct nd_struct *nd, int modem_control)
{
    char rsp_buffer[CF_CMDRSPLEN];
    int rsp_length = sizeof(rsp_buffer);
    int send, len;
    struct advtty_struct *info;
    struct tty_struct *tty;
    if(!nd)
        return (-EIO);
    if(!nd->tty_node)
	return (-EIO);
    info = nd->tty_node;
    if(!info->tty)
	return (-EIO);
    tty = info->tty;
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET,
                                   DREQ_CMD_LINECTRL) < 0)
        return (-EIO);
    nd->cmd_buffer[2] = 2;
    if(modem_control & UART_MCR_DTR)
        nd->cmd_buffer[3] = 1;
    else
        nd->cmd_buffer[3] = 0;
    if(modem_control & UART_MCR_RTS)
        nd->cmd_buffer[4] = 1;
    else
        nd->cmd_buffer[4] = 0;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_linectrl>  send command DREQ_CMD_LINECTRL failed \n", info->port);;
	return -EAGAIN;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

    if(advtty_wait_command_completed(nd,
                                     DREQ_DEVICE_CMDSET,
                                     DREQ_CMD_LINECTRL,
                                     ADVTTY_CMD_TIMEOUT,
                                     rsp_buffer,
                                     &rsp_length))
        return(-EIO);
    if(rsp_length != 4)
        return(-EIO);
    if((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K'))
        return(-EIO);
    return (0);
}

static int advtty_setxon_xoff(struct advtty_struct * info, int cmd)
{
    char rsp_buffer[CF_CMDRSPLEN];
    int  rsp_length = sizeof(rsp_buffer);
    int send, len;
    struct nd_struct * nd;
    struct tty_struct * tty;
    if(!(nd = info->net_node))
        return (-EIO);
    if(!info->tty)
	return (-EIO);
    tty = info->tty;
    if(advtty_wait_and_set_command(nd, DREQ_DEVICE_CMDSET, cmd) < 0)
        return (-EIO);
    nd->cmd_buffer[2] = 0;
    nd->cmd_ready = 1;

    if (nd->flag & ADVTTY_NET_DO_SESSION_RECOVERY)
	len = nd->do_session_recovery_len;
    else
 	len = (int)nd->cmd_buffer[2] + 3;
    send = send_to_user(tty, NETLINK_KSEND_COMD, nd->cmd_buffer, len);
    if(send <= 0)
    {
	DBGPRINT(AD_DEBUG_ERROR, "Port %d: <advtty_setxon_xoff>  send command failed\n", info->port);;
	return -EAGAIN;
    }
    nd->cmd_buffer[0] = 0;
    nd->cmd_ready = 0;

    if(advtty_wait_command_completed(nd,
                                     DREQ_DEVICE_CMDSET,
                                     cmd,
                                     ADVTTY_CMD_TIMEOUT,
                                     rsp_buffer,
                                     &rsp_length))
        return(-EIO);
    if(rsp_length != 4)
        return(-EIO);
    if((rsp_buffer[2] != 'O') || (rsp_buffer[3] != 'K'))
        return(-EIO);
    return (0);
}

module_init(advtty_module_init);
module_exit(advtty_module_exit);
