/*
 *
 *	Advantech EDG TTY port process
 *
 *	Usage: advttyp
 *
 *	Compilation instructions:
 *		LINUX:	cc -O -o advttyp advttyp.c
 *
 *	Usage: advttyp -l log -t tty -d device_type -a ip_address -p device_port
 *
 *	Version: 1.20	Date: 2007/11/01
 */
#include    <sys/types.h>
#include    <sys/wait.h>
#include    <sys/time.h>
#include    <sys/param.h>
#include    <sys/ioctl.h>
#include    <sys/socket.h>
#include    <sys/sysinfo.h>
#include    <netinet/in.h>
#include    <netinet/tcp.h>
#include    <netdb.h>
#include    <stdlib.h>
#include    <stdio.h>
#include    <errno.h>
#include    <time.h>
#include    <string.h>
#include    <unistd.h>
#include    <fcntl.h>
#include    <signal.h>
#include    <stdarg.h>

#include <linux/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <asm/types.h>
#include <linux/netlink.h>

#ifdef	STREAM
#include    <sys/ptms.h>
#endif

#include    "advttyd.h"
#include    "sock_k.h"

#define STATE_IDLE                  0
#define STATE_NETCONN               1
#define STATE_NETINIT               2
#define STATE_READY                 3
#define STATE_CLOSING               4
#define STATE_WCLOSING              5
#define STATE_CLOSED                6
#define STATE_NETDOWN               10

#define STATUS_ACKOPEN              0x0001
#define STATUS_ACKCLOSE             0x0002
#define STATUS_NWREADY              0x0004
#define STATUS_TXPOLL               0x0008
#define STATUS_WOMSENT              0x0010
#define DELAY_LEN		(4*4096)
#define DELAY_MIN		(1*4096)
#define DELAY_MAX		(2*4096)

typedef struct _TTYCTRL {
    int     mpt_fd;
    int     skt_fd;
    int     curr_state;      
    int     mpt_devidx;
	in_addr_t	dev_ipaddr[2];
	u_short dev_type;
	u_short port_idx;
	long    tty_uptime;
	long    poll_stamp;
	u_long  status;
	u_long  wait_qtime;
	long    wait_qstamp;
	int     wr_datalen;
    int     rd_datalen;
    int     rd_header;
    int     rd_tailer;
    int 	detechtime;
    int 	isstatus;
	
    unsigned char  delay_data[DELAY_LEN];
    int			delay_header;
    int			delay_tailer;
    int			delay_cnt;
    int 		throttle;

    unsigned char usendok;
    unsigned char bufferflags; 	
    unsigned char rd_data[CF_DATABUFLEN];

    int     net_rxlen;
    unsigned char net_rxmsg[CF_NETMSGLEN];
    int     tty_cmdlen;
    char    tty_cmd[CF_CMDRSPLEN];
	char	dev_ipaddr_str[2][CF_MAXSTRLEN];
	int flowCtlOn;
	int previousFreeSize;
} TTYCTRL, * TTYCTRL_P;

int __log_fd = -1;
int _running;
TTYCTRL _tty;
//jinxin added begin
//keep commond
struct msg_com_info
{
      struct nlmsghdr hdr;
      unsigned char data[CF_CMDRSPLEN];
};
//keep prue data
struct msg_data_info
{
      struct nlmsghdr hdr;
      unsigned char data[1024];
};
static int proc_num;
static int devidx;
static int skfd;
static int pid;
static int proc_fd;
static  char nl_pseudo = -1;
static unsigned int cm_flags = 0;
static int user_to_kernel(unsigned short msgtype, unsigned char comd, unsigned char *buf, int length);
//jinxin added end

static int  starup(int argc, char ** argv, TTYCTRL_P tty_p);
static void tty_proc(TTYCTRL_P tty_p);
static void ttycmd_handle(TTYCTRL_P tty_p);
static void poll_dev(TTYCTRL_P tty_p);
static int  do_port_init(TTYCTRL_P tty_p, int cmd_idx);
static void netcmd_handle(TTYCTRL_P tty_p);
static void cleanup(TTYCTRL_P tty_p);
static void shutdown_handle();
static int  log_open(char * log_name);
static void log_close(void);
static void log_msg(const char * msg);
static int  hexstr(char * strp);
static int  get_netmsg(TTYCTRL_P tty_p);
static void proc_netmsg(TTYCTRL_P tty_p);
static void dev_init(TTYCTRL_P tty_p);
static int  net_send(int skt_fd, unsigned char * msg, int len);
static int  net_recv(int skt_fd, unsigned char * msg, int len);

static __inline u_int buad_idx2val(int baudidx);
static __inline void init_netmsg(unsigned char * msg, unsigned int tid,
                                 unsigned short cmd, unsigned short len,
                                 unsigned int p1, unsigned int p2);
static __inline int curr_state(TTYCTRL_P tty_p);
static __inline int next_state(TTYCTRL_P tty_p, int next_state);


main(int argc, char ** argv)
{
    TTYCTRL_P tty_p;
    struct sockaddr_nl local;
    struct sockaddr_nl kpeer;
    struct nlmsghdr info;
    int kpeerlen;
    int  ret;
    int rcvlen = 0;
    int recvbuf=32*1024;//32K
    int sndvbuf=32*1024;//32K
    pid = getpid();
    _running = 1;
    proc_num = 0;
    signal(SIGTERM, ((void (*)())shutdown_handle));
  
    if(starup(argc, argv, tty_p = &_tty) < 0) {
        exit(-1);
    }
    tty_p->usendok=1;//now can send data to kernel
    tty_p->bufferflags= 0;
    tty_p->detechtime = 0;
    tty_p->throttle = 0;
    tty_p->isstatus = 0;
    //jinxin added begin
    proc_fd = open("/proc/vcom", O_RDWR);
    if(proc_fd < 0){
	ADV_LOGMSG("MPT[%d]: open /proc/vcom failed\n", devidx);
	cleanup(tty_p);
    	log_close();
        exit(0);
    }
    lseek(proc_fd, 0, 0);
    read(proc_fd, &nl_pseudo, sizeof(nl_pseudo));
    if(nl_pseudo == -1){
	ADV_LOGMSG("MPT[%d]: get nl_pseudo failed\n", devidx);
	cleanup(tty_p);
    	log_close();
   	close(proc_fd);
        exit(0);
    }
    ADV_LOGMSG("MPT[%d]: get nl_pseudo succ: %d\n", devidx, nl_pseudo);
    skfd = socket(PF_NETLINK, SOCK_RAW, nl_pseudo);
    if(skfd < 0) {
        ADV_LOGMSG("can not create a netlink socket\n");
	cleanup(tty_p);
    	log_close();
        exit(0);
    }
    if(setsockopt(skfd,SOL_SOCKET,SO_RCVBUF,(const char*)&recvbuf,sizeof(int)) != 0){
        ADV_LOGMSG("expand netlink recv buffer failed\n");
    }
    if(setsockopt(skfd,SOL_SOCKET,SO_SNDBUF,(const char*)&sndvbuf,sizeof(int)) != 0){
        ADV_LOGMSG("expand netlink send buffer failed\n");
    }
    memset(&local, 0, sizeof(local));
    local.nl_family = AF_NETLINK;
    local.nl_pid = getpid();
    local.nl_groups = 0;
    if(bind(skfd, (struct sockaddr*)&local, sizeof(local)) != 0) {
        ADV_LOGMSG("bind() error\n");
	goto L_EXIT;
    }

    ret = user_to_kernel((devidx << 8) + NETLINK_USEND_PID, 0, NULL, 0);
    if(ret <= 0){
	ADV_LOGMSG("send NETLINK_RECEV_PID to kernel failed, return\n");
	goto L_EXIT;
    }

    memset(&kpeer, 0, sizeof(kpeer));
    kpeer.nl_family = AF_NETLINK;
    kpeer.nl_pid = 0;
    kpeer.nl_groups = 0;
    kpeerlen = sizeof(struct sockaddr_nl);  
    while(1)	
    {
	rcvlen = recvfrom(skfd, &info, sizeof(struct nlmsghdr),
                MSG_DONTWAIT, (struct sockaddr*)&kpeer, &kpeerlen);
	//MSG_DONTWAIT enables non-blocking operation
	if(rcvlen > 0)
		break;
	usleep(100000);
	if(getppid() == 1)
		goto L_EXIT;
    }
    if(info.nlmsg_type != NETLINK_KINIT_LINE)
    {
	ADV_LOGMSG("MPT[%d] Not receive a NETLINK_KINIT_LINE msg\n", devidx);
	goto L_EXIT;
    }
    while(_running) {
        if(getppid() == 1)
            break;
	proc_num ++;
        tty_proc(tty_p);
    }
L_EXIT:
    cleanup(tty_p);
    log_close();
    close(skfd);
    close(proc_fd);
    exit(0);
}
static int user_to_kernel(unsigned short msgtype, unsigned char comd, unsigned char *buf, int length)
{
    struct sockaddr_nl kpeer;
    struct msg_info* pmessage;
    int txret, txidx;
    int retval;

    memset(&kpeer, 0, sizeof(kpeer));
    kpeer.nl_family = AF_NETLINK;
    kpeer.nl_pid = 0;
    kpeer.nl_groups = 0;

    pmessage = malloc(sizeof(struct msg_info) + length);
    memset(pmessage, 0, sizeof(struct msg_info) + length);
    if(length > 0 && buf){
    	memcpy(pmessage->data, buf, length);
    }
    pmessage->comd = comd;
    pmessage->len = length;
    pmessage->hdr.nlmsg_len = NLMSG_LENGTH(2*4 + length);
    pmessage->hdr.nlmsg_flags |= NLM_F_ACK;//try reliable communication
    pmessage->hdr.nlmsg_type = msgtype;
    pmessage->hdr.nlmsg_pid = pid;
    //make sure to send all data
    for(txidx = 0; txidx < pmessage->hdr.nlmsg_len;){
         if((txret = sendto(skfd, pmessage + txidx, pmessage->hdr.nlmsg_len - txidx, 0,
            (struct sockaddr*)&kpeer, sizeof(kpeer))) <= 0)
		break;
	 txidx += txret;
    }
    if(txidx != pmessage->hdr.nlmsg_len){
	ADV_LOGMSG("MPT[%d] user_to_kernel: not send all, actually send %d, plan to send %d\n", devidx, txidx, pmessage->hdr.nlmsg_len);
    	free(pmessage);
	return -1;
    }
    free(pmessage);
    return txidx;
}

static int starup(int argc, char ** argv, TTYCTRL_P tty_p)
{
    int i;
    
    if(argc != 13 && argc != 11) {
        return -1;
    }
    
    memset(tty_p, 0, sizeof(TTYCTRL));
    for(i = 1; i < argc; ++i) {
        if(strcmp(argv[i], "-l") == 0) {
        		printf("log_open log file = %s\n", argv[i+1]);
            log_open(argv[++i]);
            printf("log_open log file = %s\n", argv[i]);
        }
        if(strcmp(argv[i], "-t") == 0) {
            tty_p->mpt_devidx = atoi(argv[++i]);
	    devidx = tty_p->mpt_devidx;
        }
        else if(strcmp(argv[i], "-d") == 0) {
            tty_p->dev_type = (u_short)hexstr(argv[++i]);
        }
        else if(strcmp(argv[i], "-a") == 0) {
        		ADV_LOGMSG("argv[%d] = %s\n", i + 1, argv[i + 1]);
            strcpy((tty_p->dev_ipaddr_str)[0], argv[++i]);
            (tty_p->dev_ipaddr)[0] = inet_addr((tty_p->dev_ipaddr_str)[0]);
            ADV_LOGMSG("argv[%d] = %s\n", i, argv[i]);
        }
        else if(strcmp(argv[i], "-p") == 0) {
            tty_p->port_idx = (u_short)atoi(argv[++i]);
        } 
        else if(strcmp(argv[i], "-r") == 0) {
        		ADV_LOGMSG("argv[%d] = %s\n", i + 1, argv[i + 1]);
            strcpy((tty_p->dev_ipaddr_str)[1], argv[++i]);
            (tty_p->dev_ipaddr)[1] = inet_addr((tty_p->dev_ipaddr_str)[1]); 
            ADV_LOGMSG("argv[%d] = %s\n", i, argv[i]);
        }
    }

    if((tty_p->dev_type == 0) || (tty_p->dev_ipaddr[0] == 0) ||
       (tty_p->port_idx == 0)) {
        ADV_LOGMSG("Invalid argument mpt[%d] type[%x] ip[%s] port[%d]\n",
                   tty_p->mpt_devidx, tty_p->dev_type,
                   tty_p->dev_ipaddr[0], tty_p->port_idx);
        log_close();
        return -1;
    }
    skfd = -1;
    tty_p->skt_fd = -1;
    ADV_LOGMSG("return 0\n", i + 1, argv[i + 1]);
    return 0;
}

int static QueueUsedSize(TTYCTRL_P tty_p)
{
	return 	(tty_p->rd_tailer >= tty_p->rd_header)?
        	(tty_p->rd_tailer-tty_p->rd_header):
          (CF_DATABUFLEN +(tty_p->rd_tailer-tty_p->rd_header));
}
int static QueueKernelSize(TTYCTRL_P tty_p)
{
	return 	(tty_p->delay_tailer >= tty_p->delay_header)?
        	(tty_p->delay_tailer-tty_p->delay_header):
          (DELAY_LEN +(tty_p->delay_tailer-tty_p->delay_header));
}

void static sendQueueFreeSize(TTYCTRL_P tty_p, int unlockOnly)
{
		//static int previousFreeSize = 0;
		unsigned long queueFreeSize= 0;
		char msg[20];
		
		if(unlockOnly && (tty_p->flowCtlOn == 0))
		{
			return;
		}
		
		queueFreeSize = (unsigned long)(CF_DATABUFLEN -QueueUsedSize(tty_p) - 1);

		if ( queueFreeSize <=  14*1024)
		//if ( queueFreeSize <= 500 )
			queueFreeSize = 0;
			
		if(queueFreeSize != tty_p->previousFreeSize)
		{
		
			if ((QueueUsedSize(tty_p) > CF_DATABUFFCH) && !unlockOnly)
			{
				init_netmsg(msg, 0xffffffff,0x0900, 0, 0, queueFreeSize/2);
				tty_p->flowCtlOn = 1;
			}
			else if(QueueUsedSize(tty_p) < CF_DATABUFFCL && 
						tty_p->previousFreeSize < queueFreeSize)
			{
				init_netmsg(msg, 0xffffffff,0x0900, 0, 0, queueFreeSize);
				tty_p->flowCtlOn = 0;
			}
			else
				return;
		
		tty_p->previousFreeSize = queueFreeSize;
		net_send(tty_p->skt_fd, msg, 20);
	}
			
	return;
}

static void tty_proc(TTYCTRL_P tty_p)
{
    int maxfd;
    int sel_ret;
    unsigned short datalen = 0;
    int tmp_skt[2]= {-1};
    int i;
    fd_set rd_fd, wr_fd, ex_fd;
    struct timeval tm;
    struct sysinfo sys_info;
    char tty_rsp[CF_CMDRSPLEN];
    unsigned char msg[128];
    struct timeval connectTimeout;
		fd_set connectFd;
		int selectResult;
		int maxFd = 0;
		int arg;
    struct sockaddr_nl kpeer;
    int rcvlen, sendlen, kpeerlen, status;
    static minlen = 0;
    struct nlmsghdr msginfo;
    struct msg_com_info cmd_info;
    struct msg_data_info data_info;//1024+16
    
    maxfd = -1;
    if(skfd > maxfd)
        maxfd = skfd;
    if(tty_p->skt_fd > maxfd)
        maxfd = tty_p->skt_fd;

    memset(&kpeer, 0, sizeof(kpeer));
    kpeer.nl_family = AF_NETLINK;
    kpeer.nl_pid = 0;
    kpeer.nl_groups = 0;
    kpeerlen = sizeof(struct sockaddr_nl);
    switch(curr_state(tty_p)) {
    case STATE_IDLE: {
        //ADV_LOGMSG("MPT[%d] STATE_IDLE\n", devidx);
	tm.tv_sec = 3;
        tm.tv_usec = 0;
        FD_SET(skfd, &rd_fd);
        if((sel_ret= select(skfd + 1, &rd_fd, NULL, NULL, &tm)) <= 0)
            break;
        if(!FD_ISSET(skfd, &rd_fd)) {
            break;
        }
        if(tty_p->tty_cmd[0] != DREQ_FREE_CMDSET) {
            break;
        }
	memset(&data_info, 0, sizeof(data_info));//16 + data[1024]
        rcvlen = recvfrom(skfd, &data_info, sizeof(struct msg_data_info),
                MSG_DONTWAIT, (struct sockaddr*)&kpeer, &kpeerlen);
	if(rcvlen <= 0)
		break;
	switch(data_info.hdr.nlmsg_type){
		case NETLINK_KSEND_COMD://deal with command
		{
        		if(tty_p->tty_cmd[0] != 0){
				ADV_LOGMSG("STATE_READY: MPT[%d] missing command!\n", devidx);
				break;
			}
			tty_p->tty_cmdlen = data_info.hdr.nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
			if(tty_p->tty_cmdlen > 0){
				memcpy(tty_p->tty_cmd, data_info.data, tty_p->tty_cmdlen);
				if(tty_p->tty_cmdlen > 0)
                			ttycmd_handle(tty_p);
			}
			break;
		}
		case NETLINK_KINIT_LINE:
		{
			ADV_LOGMSG("Port %d opened \n", devidx);			
			break;
		}
		default:
		{
			ADV_LOGMSG(" MPT[%d] receive other msg(%x) on STATE_IDLE\n", devidx, data_info.hdr.nlmsg_type);
			break;
		}
	}
        break;}
        
    case STATE_NETCONN: {
        //ADV_LOGMSG("MPT[%d] STATE_NETCONN\n", devidx);
        int setarg;
        struct sockaddr_in sin;
        
        if(tty_p->skt_fd >= 0) {
            close(tty_p->skt_fd);
            tty_p->skt_fd = -1;
        }

        connectTimeout.tv_sec = 5;
	connectTimeout.tv_usec = 0;
	FD_ZERO(&connectFd);
        
        for(i = 0; i< 2; i++)
        {
        	if(tty_p->dev_ipaddr[i] <= 0)
        	{
        		tmp_skt[i] = -1;
        		continue;
        	}
        		
        	if((tmp_skt[i] = socket(AF_INET, SOCK_STREAM, 0)) < 0)
        	{
        		ADV_LOGMSG("MPT[%d] create socket fail [%d]\n",tty_p->mpt_devidx, tmp_skt[i]);     
            		goto L_CHKCONN;
        	}        	
        	setarg = 1;
		setsockopt(tmp_skt[i], IPPROTO_TCP, TCP_NODELAY, 
				(char *)&setarg, sizeof(setarg));
		setarg = 1;
        	if(setsockopt(tmp_skt[i], SOL_SOCKET, SO_KEEPALIVE,
                      (char *)&setarg, sizeof(setarg)) < 0) 
		{
           		 ADV_LOGMSG("MPT[%d] Set TCP keep alive fail\n",
                       	tty_p->mpt_devidx);            
            		close(tmp_skt[i]);
            		tmp_skt[i] = -1;
           	 	goto L_CHKCONN;
        	}
        
        	sin.sin_family = AF_INET;
        	sin.sin_addr.s_addr = tty_p->dev_ipaddr[i];
        	sin.sin_port = htons(5202);
        	
        	arg |= O_NONBLOCK;
 		if( fcntl(tmp_skt[i], F_SETFL, arg) < 0) {
    			ADV_LOGMSG( "Error fcntl(..., F_SETFL) (%s)\n", strerror(errno));
    			goto L_CHKCONN;
 		}			
        	connect(tmp_skt[i], (struct sockaddr *)&sin, sizeof(sin));
        	FD_SET(tmp_skt[i], &connectFd);
        }
        
        maxFd = (tmp_skt[1] >= tmp_skt[0])?tmp_skt[1]:tmp_skt[0];
     		
     	if(maxFd < 0)
        	goto L_CHKCONN;
        	
        selectResult = select(maxFd + 1 , 0, &connectFd, 0, &connectTimeout);
        
        if(selectResult > 0)
        {
        	for(i = 0; i < 2; i++)
        	{
        		if(FD_ISSET(tmp_skt[i], &connectFd))
        		{
        			tty_p->skt_fd = tmp_skt[i];
        			break;
        		}
        	}
        	
        	for(i = 0; i < 2; i++)
        	{
        		if(tty_p->skt_fd != tmp_skt[i])
        		{
        			close(tmp_skt[i]);
        			tmp_skt[i] = -1;
        		}
        	}
        	
        	arg &= (~O_NONBLOCK);
 		if( fcntl(tty_p->skt_fd, F_SETFL, arg) < 0) {
    			ADV_LOGMSG( "Error setting socket to blocking mode(..., F_SETFL) (%s)\n", strerror(errno));
    			goto L_CHKCONN;
 		}
 		// set timeout params for SO_KEEPALIVE
		int keepTime = 5; //the interval between the last data packet sent (simple ACKs are not considered data) and the first keepalive probe
		int keepProbe = 1; //number of retry
		int keepInvl = 1; //the interval between subsequential keepalive probes, regardless of what the connection has exchanged in the meantime
		if( setsockopt(tty_p->skt_fd, SOL_TCP, TCP_KEEPIDLE, &keepTime, sizeof(keepTime)) == -1
			||setsockopt(tty_p->skt_fd, SOL_TCP, TCP_KEEPCNT, &keepProbe, sizeof(keepProbe)) == -1
			||setsockopt(tty_p->skt_fd, SOL_TCP, TCP_KEEPINTVL, &keepInvl, sizeof(keepInvl)) == -1)
		{
 			ADV_LOGMSG("Socket %d setKeepAlive timeout failed.\n", tty_p->skt_fd);
 			close(tty_p->skt_fd);
 			goto L_CHKCONN;
		}
        }
        else
        {
	        ADV_LOGMSG("MPT[%d] select failed\n", devidx);
        	goto L_CHKCONN;
        }
        
        init_netmsg(msg, 0xffffffff, 0, 0,
                    (((u_int)tty_p->dev_type)<<16) | (u_int)0x100,
                    (u_int)tty_p->port_idx);
        if(net_send(tty_p->skt_fd, msg, 20) != 20) {
		close(tty_p->skt_fd);
		tty_p->skt_fd = -1;
		ADV_LOGMSG("MPT[%d] STATE_NETCONN: net_send failed \n", devidx);
		goto L_CHKCONN;
        }
        next_state(tty_p, STATE_NETINIT);
        break;
        
L_CHKCONN:
        	
        for(i = 0; i < 2; i++)
        {
        	if(tmp_skt[i] > 0)
        	{
        		close(tmp_skt[i]);
        		tmp_skt[i] = -1;
        	}
        }
        if(tty_p->status & STATUS_ACKOPEN) {
            sysinfo(&sys_info);
            if(sys_info.uptime - tty_p->tty_uptime > 60) {
                //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_DISCONNECTED, 0), 0);
	    	user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_DISCONNECTED, NULL, 0);
                tty_p->status &= ~STATUS_ACKOPEN;
                next_state(tty_p, STATE_IDLE);
            }
        }
        else {
            next_state(tty_p, STATE_NETDOWN);
        }
        
        break;
	
	}
        
    case STATE_NETINIT: {
        //ADV_LOGMSG("MPT[%d] STATE_NETINIT\n", devidx);
        uint32_t ret;
        tm.tv_sec = 1;
        tm.tv_usec = 0;
        FD_ZERO(&rd_fd);
        FD_SET(tty_p->skt_fd, &rd_fd);
        if((sel_ret= select(maxfd + 1, &rd_fd, NULL, NULL, &tm)) <= 0) {
            sysinfo(&sys_info);
            if(sys_info.uptime - tty_p->tty_uptime > 60) {
                next_state(tty_p, STATE_NETDOWN);
            }
            return;
        }
        if(!FD_ISSET(tty_p->skt_fd, &rd_fd)) {
            break;
        }
        if(net_recv(tty_p->skt_fd, msg, 20) != 20) {
            close(tty_p->skt_fd);
            tty_p->skt_fd = -1;
            next_state(tty_p, STATE_NETDOWN);
            break;
        }
        ret = ntohl(*((unsigned int *)&msg[12]));
        if(ret != 0) {
            close(tty_p->skt_fd);
            tty_p->skt_fd = -1;
            sleep(1);
            next_state(tty_p, STATE_NETDOWN);
            break;
        }
        
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_CONNECTED, 0), 0);
	user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_CONNECTED, NULL, 0);
	
        tty_p->status |= STATUS_NWREADY;
        if(tty_p->status & STATUS_ACKOPEN) {
            tty_rsp[0] = DREQ_DRIVER_CMDSET;
            tty_rsp[1] = CMD_DRV_TTY_OPEN;
            //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 2), tty_rsp);
            user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 2);
	    tty_p->status &= ~STATUS_ACKOPEN;
        }
             
        sysinfo(&sys_info);
        tty_p->poll_stamp = sys_info.uptime;
        next_state(tty_p, STATE_READY);
        break;}
        
    case STATE_READY: { 
	//ADV_LOGMSG("# MPT[%d] STATE_READY\n", devidx);
	int sendsize;
    	int count, c;
        tm.tv_sec = 1;
        tm.tv_usec = 0;
        FD_ZERO(&rd_fd);
        FD_ZERO(&wr_fd);
        if(tty_p->detechtime)
		tty_p->detechtime--;
        if(tty_p->tty_cmd[0] == 0)
            FD_SET(skfd, &rd_fd);
        if(!tty_p->net_rxlen)
            FD_SET(tty_p->skt_fd, &rd_fd);
        if(!tty_p->wr_datalen)
            FD_SET(skfd, &rd_fd);
        if(	QueueUsedSize(tty_p) > 0 && !tty_p->detechtime)
            FD_SET(skfd, &wr_fd);
            
        if((sel_ret= select(maxfd + 1, &rd_fd, &wr_fd, NULL, &tm)) <= 0) {
            sysinfo(&sys_info);
            if((sys_info.uptime - tty_p->poll_stamp > CF_EKI_HEARTBEAT) &&
               (tty_p->status & STATUS_TXPOLL)) {
                tty_p->status &= ~STATUS_TXPOLL;
                next_state(tty_p, STATE_NETDOWN);
		ADV_LOGMSG("# MPT[%d] netdown at STATE_READY\n", devidx);
                return;      
            }
            poll_dev(tty_p);
            return;
        }

	if(FD_ISSET(tty_p->skt_fd, &rd_fd) &&
           (get_netmsg(tty_p) > 0)) {
	    //ADV_LOGMSG("# MPT[%d]  skf_rd rd_fd\n", devidx);
            proc_netmsg(tty_p);
            sysinfo(&sys_info);
            tty_p->poll_stamp = sys_info.uptime;
            tty_p->status &= (~STATUS_TXPOLL);
        }  

        if(FD_ISSET(skfd, &rd_fd)) {
		memset(&data_info, 0, sizeof(data_info));//16 + data[1024]
        	rcvlen = recvfrom(skfd, &data_info, sizeof(struct msg_data_info),
                	MSG_DONTWAIT, (struct sockaddr*)&kpeer, &kpeerlen);
		if(rcvlen <= 0)
			break;
		switch(data_info.hdr.nlmsg_type){
		    case NETLINK_KSEND_COMD://deal with command
		    {
			//ADV_LOGMSG("# MPT[%d] skfd rd_fd NETLINK_KSEND_COMD\n", devidx);
        		if(tty_p->tty_cmd[0] != 0){
				ADV_LOGMSG("STATE_READY: MPT[%d] cmd handle is busy!\n", devidx);
				break;
			}
			tty_p->tty_cmdlen = data_info.hdr.nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
			if(tty_p->tty_cmdlen > 0){
				memcpy(tty_p->tty_cmd, data_info.data, tty_p->tty_cmdlen);
				if(tty_p->tty_cmdlen > 0)
                			ttycmd_handle(tty_p);
			}
			break;
		    }        
		    case NETLINK_KSEND_DATA://deal with data
	    	    {
			//buffer the kernel data to delay_data[];
			int movedata_len;
			int tmp1, tmp2;
			count = data_info.hdr.nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
			if((movedata_len = count) + QueueKernelSize(tty_p) >= DELAY_LEN)
        		{
            			movedata_len = DELAY_LEN - QueueKernelSize(tty_p) - 1;
            			char msg1[500];
          			sprintf(msg1, "data trunked from kernel, movedata_len = %d\n", movedata_len);
           			log_msg(msg1);
         		}   	
            
        		if(tty_p->delay_tailer + movedata_len < DELAY_LEN) {
            			memcpy(&tty_p->delay_data[tty_p->delay_tailer],
                   		data_info.data, movedata_len);
            			tty_p->delay_tailer += movedata_len;
            
        		}
        		else if(tty_p->delay_tailer + movedata_len == DELAY_LEN) {
            			memcpy(&tty_p->delay_data[tty_p->delay_tailer],
                   			data_info.data, movedata_len);
            			tty_p->delay_tailer = 0;
            
        		}
        		else{
            			tmp1 = DELAY_LEN - tty_p->delay_tailer;
            			memcpy(&tty_p->delay_data[tty_p->delay_tailer],
                   			data_info.data, tmp1);
            			tmp2 = movedata_len - tmp1;
            			memcpy(&tty_p->delay_data[0],
                   			&data_info.data[tmp1], tmp2);
            			tty_p->delay_tailer = tmp2;
        		}	
			//now throttle the kernel data, so don't send NETLINK_KSEND_DATA_OK this time
			if(QueueKernelSize(tty_p) >= DELAY_MAX){
				tty_p->throttle = 1;		
				break;
			}
			//ADV_LOGMSG("# MPT[%d] in QueueKernelSize(tty_p) is %d\n", devidx, QueueKernelSize(tty_p));
			sendlen = user_to_kernel((devidx << 8)+NETLINK_KSEND_DATA_OK, 0, &count, sizeof(count));
			if(sendlen <= 0){
				ADV_LOGMSG("MPT[%d] send delay data : send NETLINK_KSEND_DATA_OK failed\n", devidx);
			}	
			else
				;//ADV_LOGMSG("# MPT[%d] in QueueKernelSize(tty_p) is %d, send DATA_OK  1\n", devidx, QueueKernelSize(tty_p));				
			break;
	    	    }	
		    case NETLINK_USEND_DATA_OK:
		    {
			tty_p->usendok = 1;
			sendsize = (int)(*(int*)data_info.data);
			if(tty_p->bufferflags == 1){
				tty_p->rd_header += sendsize;
				tty_p->bufferflags = 0;
			}
			else if(tty_p->bufferflags == 2){
				tty_p->rd_header = (tty_p->rd_header + sendsize) % 
                                   CF_DATABUFLEN;
				tty_p->bufferflags = 0;
			}
            		if(QueueUsedSize(tty_p) == 0){
                		tty_p->rd_datalen = tty_p->rd_header = tty_p->rd_tailer = 0;
			}
            		sendQueueFreeSize(tty_p, 1);
			break;
		   }
		   case NETLINK_USEND_DATA_NOK:
		   {
			tty_p->usendok = 1;
			tty_p->detechtime = 10;
			break;
		   }
		   default:
			ADV_LOGMSG("STATE_READY: MPT[%d] receive other msg(%x) is a bug\n", devidx, data_info.hdr.nlmsg_type);
			break;
		}
        }

   	       
        
        if(FD_ISSET(skfd, &wr_fd) && tty_p->usendok && QueueUsedSize(tty_p)>0) 
	{
	   	//ADV_LOGMSG("# MPT[%d]  skfd wr_fd\n", devidx);
		int rd_len;
		/*
 		if(QueueUsedSize(tty_p) <= 0)
            	{
            		char msg1[500];
            		sprintf(msg1, "queue size is empty\n");
            		log_msg(msg1);
            		goto L_COMPLETE;
            	}*/
		if(tty_p->rd_header < tty_p->rd_tailer) 
		{
			minlen = MIN(2048, tty_p->rd_tailer - tty_p->rd_header);
                	rd_len = user_to_kernel((devidx << 8)+NETLINK_USEND_DATA, 0,
                               &tty_p->rd_data[tty_p->rd_header],
                               minlen);
                	if(rd_len <= 0){
				ADV_LOGMSG("STATE_READY: MPT[%d] send data to kernel failed\n", devidx);
				break;
			}
			tty_p->usendok = 0;
			tty_p->bufferflags = 1;

                 }
            	else 
		{
			minlen = MIN(2048, CF_DATABUFLEN - tty_p->rd_header);
                	rd_len = user_to_kernel((devidx << 8)+NETLINK_USEND_DATA, 0,
                               &tty_p->rd_data[tty_p->rd_header],
                               minlen);
                	if(rd_len <= 0){
				ADV_LOGMSG("STATE_READY: MPT[%d] send data to kernel failed\n", devidx);
				break;
			}
			tty_p->usendok = 0;
			tty_p->bufferflags = 2;

            	}
        }
	if(tty_p->throttle){
		if(QueueKernelSize(tty_p) <= DELAY_MIN){
			//now unthrottle the kernel data, so  send NETLINK_KSEND_DATA_OK to continue
			sendlen = user_to_kernel((devidx << 8)+NETLINK_KSEND_DATA_OK, 0, &count, sizeof(count));
			if(sendlen <= 0){
				ADV_LOGMSG("MPT[%d] send delay data : send NETLINK_KSEND_DATA_OK failed\n", devidx);
			}
			else{
				tty_p->throttle = 0;		
				//ADV_LOGMSG("# MPT[%d] in QueueKernelSize(tty_p) is %d, send DATA_OK 2\n", devidx, QueueKernelSize(tty_p));							
			}	
		}
	}
			
	//ADV_LOGMSG("STATE_READY: MPT[%d] wr_datalen is %d, QueueKernelSize(tty_p) is %d\n", devidx, tty_p->wr_datalen, QueueKernelSize(tty_p));
	if(!tty_p->wr_datalen){
		unsigned char msg[CF_NETMSGLEN];
		int len;
		if(QueueKernelSize(tty_p) <= 0){
            		//ADV_LOGMSG("MPT[%d] kernel queue size is empty, tty_p->throttle is %d\n", devidx, tty_p->throttle);
            		goto COMPLETE;
            	}

		//ADV_LOGMSG("# MPT[%d] out 1 QueueKernelSize(tty_p) is %d\n", devidx, QueueKernelSize(tty_p));					
		if(tty_p->delay_header < tty_p->delay_tailer) {
			len = MIN(1024, tty_p->delay_tailer - tty_p->delay_header);
			if(len <= 0)
				goto COMPLETE;
			init_netmsg(msg, 0xfffffffe, 0x04, (unsigned short)len, 0, 0);
			memcpy(msg+20, &tty_p->delay_data[tty_p->delay_header], len);
	
                	if(net_send(tty_p->skt_fd, msg, 20 + len) == 20 + len)
			{
                 	  	tty_p->wr_datalen = len;
			}
			else
			{
				ADV_LOGMSG("MPT[%d] send delay data : net_send not equ receive\n", devidx);
				break;
			}
                	tty_p->delay_header += len;
                 }
            	else {
                	len = MIN(1024, DELAY_LEN - tty_p->delay_header);
			if(len <= 0)
				goto COMPLETE;
			init_netmsg(msg, 0xfffffffe, 0x04, (unsigned short)len, 0, 0);
			memcpy(msg+20, &tty_p->delay_data[tty_p->delay_header], len);
	
                	if(net_send(tty_p->skt_fd, msg, 20 + len) == 20 + len)
			{
                 	  	tty_p->wr_datalen = len;
			}
			else
			{
				ADV_LOGMSG("MPT[%d] send delay data : net_send not equ receive\n", devidx);
				break;
			}
			tty_p->delay_header = (tty_p->delay_header + len) % DELAY_LEN;  
            	}
		//ADV_LOGMSG("# MPT[%d] out 2 QueueKernelSize(tty_p) is %d\n", devidx, QueueKernelSize(tty_p));					
		COMPLETE:
		if(QueueKernelSize(tty_p) == 0)
                	tty_p->delay_cnt = tty_p->delay_header = tty_p->delay_tailer = 0;


	}
        
        if(!FD_ISSET(tty_p->skt_fd, &rd_fd))/*if no packet has been recieved for a perioud send heartbeat*/
        {
        	if(!(tty_p->status & STATUS_TXPOLL))
        	{
        		poll_dev(tty_p);/*send heartbeat; need to dismiss STATUS_TXPOLL if data is recieved from socket*/
        	}
        	else if(tty_p->status & STATUS_TXPOLL)
        	{       		
        		sysinfo(&sys_info);
          	if((sys_info.uptime - tty_p->poll_stamp) > CF_EKI_HEARTBEAT)/*heartbeat timeout break connection*/
          	{
               	tty_p->status &= ~STATUS_TXPOLL;
               	next_state(tty_p, STATE_NETDOWN);
             		return;      
           	}
          }
        }
        
        
         
        break;}
        
    case STATE_CLOSING: {
        //ADV_LOGMSG("MPT[%d] STATE_CLOSING\n", devidx);
        if((tty_p->skt_fd == -1) || !(tty_p->status & STATUS_NWREADY)){
            next_state(tty_p, STATE_CLOSED);
            break;
        }
        init_netmsg(msg, 0xffffffff, 0x12, 0, 0, 0);
        if(net_send(tty_p->skt_fd, msg, 20) != 20) {
            next_state(tty_p, STATE_CLOSED);
            break;
        }
        init_netmsg(msg, 0xffffffff, 0x02, 0, 0, 0);
        if(net_send(tty_p->skt_fd, msg, 20) != 20) {
            next_state(tty_p, STATE_CLOSED);
            break;
        }
        sysinfo(&sys_info);
        tty_p->poll_stamp = sys_info.uptime;
        next_state(tty_p, STATE_WCLOSING);
        break;}
        
    case STATE_WCLOSING: {
        //ADV_LOGMSG("MPT[%d] STATE_WCLOSING\n", devidx);
        tm.tv_sec = 1;
        tm.tv_usec = 0;
        FD_ZERO(&rd_fd);
        FD_SET(tty_p->skt_fd, &rd_fd);
        if(select(maxfd + 1, &rd_fd, NULL, NULL, &tm) <= 0) {
            sysinfo(&sys_info);
            if(sys_info.uptime - tty_p->poll_stamp > 60) {
                next_state(tty_p, STATE_CLOSED);
            }
            break;
        }
        
        if(FD_ISSET(tty_p->skt_fd, &rd_fd) && (get_netmsg(tty_p) > 0)) {
            if(ntohs(*((unsigned short *)&tty_p->net_rxmsg[8])) == 0x02) {
                next_state(tty_p, STATE_CLOSED);
            }
            tty_p->net_rxlen = 0;
        }
        break;}
        
    case STATE_CLOSED: {
        //ADV_LOGMSG("MPT[%d] STATE_CLOSED\n", devidx);
        if(tty_p->status & STATUS_ACKCLOSE) {
            tty_rsp[0] = DREQ_DRIVER_CMDSET;
            tty_rsp[1] = CMD_DRV_TTY_CLOSE;
            //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 2), tty_rsp);
            user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 2);        

	    tty_p->status &= ~STATUS_ACKCLOSE;
        }
        dev_init(tty_p);
        next_state(tty_p, STATE_IDLE);
        break;}
        
    case STATE_NETDOWN: {
        //ADV_LOGMSG("MPT[%d] STATE_NETDOWN\n", devidx);
        int tty_status = 0;
        int new_state;
        
        tty_p->status &= ~(STATUS_TXPOLL | STATUS_WOMSENT);
	//notify kernel to disconnect begin
        if(tty_p->status & STATUS_NWREADY) {
	    user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_DISCONNECTED, NULL, 0);
            tty_p->status &= ~STATUS_NWREADY;
        }
        //notify kernel to disconnect end
        if(tty_p->skt_fd >= 0) {
            close(tty_p->skt_fd);
            tty_p->skt_fd = -1;
        }

        tty_p->wr_datalen = 0;
        tty_p->net_rxlen = 0;
        
        new_state = STATE_IDLE; 
	if(!tty_p->isstatus){	
		user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_GET_TTY_STATUS, (unsigned char*)&status, sizeof(status));
		tty_p->isstatus = 1;	
	}
	if(tty_p->status & STATUS_ACKCLOSE) {
            new_state = STATE_CLOSING;
        }
	if((tty_p->status & STATUS_ACKOPEN) || tty_status) {
            new_state = STATE_NETCONN;
        }

	tm.tv_sec = 0;
        tm.tv_usec = 200 * 1000;
        FD_ZERO(&rd_fd);
	FD_SET(skfd, &rd_fd);
        if((sel_ret= select(skfd + 1, &rd_fd, NULL, NULL, &tm)) <= 0){
		goto end;
		//break;
	}
        if(!FD_ISSET(skfd, &rd_fd)) {
		goto end;
           	//break;
        }
	memset(&data_info, 0, sizeof(data_info));
	rcvlen = recvfrom(skfd, &data_info, sizeof(struct msg_data_info),
        MSG_DONTWAIT, (struct sockaddr*)&kpeer, &kpeerlen);
	if(rcvlen <= 0){
		ADV_LOGMSG("STATE_NETDOWN: MPT[%d] recvfrom nothing\n", devidx);
		break;
	}
	switch(data_info.hdr.nlmsg_type)
	{
		case NETLINK_TTY_STATUS://receive tty status 
			tty_p->tty_cmdlen = data_info.hdr.nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
			memcpy(&tty_status, data_info.data, tty_p->tty_cmdlen);
			if(tty_p->status & STATUS_ACKCLOSE) {
           			new_state = STATE_CLOSING;
        		}
			if((tty_p->status & STATUS_ACKOPEN) || (tty_status)) {
            			new_state = STATE_NETCONN;
        		}
			tty_p->isstatus = 0;//now you can send next IOCMD_GET_TTY_STATUS msg	
			break;
		case NETLINK_KSEND_COMD:
			if(tty_p->tty_cmd[0] != 0){
				ADV_LOGMSG("STATE_READY: MPT[%d] missing command!\n", devidx);
				break;
			}
			tty_p->tty_cmdlen = data_info.hdr.nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
			if(tty_p->tty_cmdlen > 0){
				memcpy(tty_p->tty_cmd, data_info.data, tty_p->tty_cmdlen);
				if(tty_p->tty_cmdlen > 0)
                			ttycmd_handle(tty_p);
			}
			new_state = STATE_NETDOWN; 
			break;
		//copy from STATE_READY begin
		case NETLINK_KSEND_DATA:
	    	{
			ADV_LOGMSG("STATE_NETDOWN: MPT[%d] get NETLINK_KSEND_DATA\n", devidx);
			int movedata_len;
			int tmp1, tmp2, count;
			count = data_info.hdr.nlmsg_len - NLMSG_ALIGN(sizeof(struct nlmsghdr));
			if((movedata_len = count) + QueueKernelSize(tty_p) >= DELAY_LEN)
        		{
            			movedata_len = DELAY_LEN - QueueKernelSize(tty_p) - 1;
            			char msg1[500];
          			sprintf(msg1, "data trunked from kernel, movedata_len = %d\n", movedata_len);
           			log_msg(msg1);
         		}   	
            
        		if(tty_p->delay_tailer + movedata_len < DELAY_LEN) {
            			memcpy(&tty_p->delay_data[tty_p->delay_tailer],
                   		data_info.data, movedata_len);
            			tty_p->delay_tailer += movedata_len;
            
        		}
        		else if(tty_p->delay_tailer + movedata_len == DELAY_LEN) {
            			memcpy(&tty_p->delay_data[tty_p->delay_tailer],
                   			data_info.data, movedata_len);
            			tty_p->delay_tailer = 0;
            
        		}
        		else{
            			tmp1 = DELAY_LEN - tty_p->delay_tailer;
            			memcpy(&tty_p->delay_data[tty_p->delay_tailer],
                   			data_info.data, tmp1);
            			tmp2 = movedata_len - tmp1;
            			memcpy(&tty_p->delay_data[0],
                   			&data_info.data[tmp1], tmp2);
            			tty_p->delay_tailer = tmp2;
        		}	
			/* 
			   now, kernel should not send, so throttle kernel data
			   so don't send NETLINK_KSEND_DATA_OK this time too.
			   if we are luck to recover to STATE READY, we can continue
			   when QueueKernelSize(tty_p) <= DELAY_MIN
			*/
			tty_p->throttle = 1;	
			new_state = STATE_NETDOWN; 				
			break;
	    	}	
		case NETLINK_USEND_DATA_OK:
		{
			ADV_LOGMSG("STATE_NETDOWN: MPT[%d] get NETLINK_USEND_DATA_OK\n", devidx);
			int sendsize;
			tty_p->usendok = 1;
			sendsize = (int)(*(int*)data_info.data);
			if(tty_p->bufferflags == 1){
				tty_p->rd_header += sendsize;
				tty_p->bufferflags = 0;
			}
			else if(tty_p->bufferflags == 2){
				tty_p->rd_header = (tty_p->rd_header + sendsize) % 
                                   CF_DATABUFLEN;
				tty_p->bufferflags = 0;
			}
            		if(QueueUsedSize(tty_p) == 0){
                		tty_p->rd_datalen = tty_p->rd_header = tty_p->rd_tailer = 0;
			}
			new_state = STATE_NETDOWN; 
			break;
		}
		case NETLINK_USEND_DATA_NOK:
		{
			ADV_LOGMSG("STATE_NETDOWN: MPT[%d] get NETLINK_USEND_DATA_NOK\n", devidx);
			tty_p->usendok = 1;
			tty_p->detechtime = 10;
			new_state = STATE_NETDOWN; 
			break;
		}
		//copy from STATE_READY end

		default:
			ADV_LOGMSG("STATE_NETDOWN: MPT[%d] nlmsg %x is not processed\n", devidx, data_info.hdr.nlmsg_type);//can ignore msg 0x14
			new_state = STATE_NETDOWN; 
			break;
        
	}
end:
        if(new_state == STATE_IDLE) {
            dev_init(tty_p);
        }
        else if(new_state == STATE_NETCONN) {
            sysinfo(&sys_info);
            tty_p->tty_uptime = sys_info.uptime;
        }
        next_state(tty_p, new_state);
        break;}
    }
    return;
}

static void ttycmd_handle(TTYCTRL_P tty_p)
{
    struct sysinfo sys_info;
 
    switch(tty_p->tty_cmd[0]){
    case DREQ_DEVICE_CMDSET:
        netcmd_handle(tty_p);
        return;
    case DREQ_DRIVER_CMDSET:
        switch(tty_p->tty_cmd[1]) {
        case CMD_DRV_TTY_OPEN:
	    ADV_LOGMSG("# \tCMD_DRV_TTY_OPEN[%d]\n", devidx);
            if(curr_state(tty_p) != STATE_IDLE) {
                ADV_LOGMSG("MPT[%d] (ttycmd_handle) busy reopen\n",
                           tty_p->mpt_devidx);
                break;
            }
            tty_p->status |= STATUS_ACKOPEN;
            sysinfo(&sys_info);
            tty_p->tty_uptime = sys_info.uptime;
            next_state(tty_p, STATE_NETCONN);
            break;
        case CMD_DRV_TTY_CLOSE:
	    ADV_LOGMSG("# \tCMD_DRV_TTY_CLOSE[%d]\n", devidx);
            tty_p->status |= STATUS_ACKCLOSE;
            next_state(tty_p, STATE_CLOSING);
            break;
        default:
            ADV_LOGMSG("MPT[%d] (ttycmd_handle) unknown local cmd[%x]\n",
                       tty_p->mpt_devidx, tty_p->tty_cmd[1]);
            break;
        }
        break;
    default:
        ADV_LOGMSG("MPT[%d] (ttycmd_handle) unknown tty cmd[%x]\n",
                   tty_p->mpt_devidx, tty_p->tty_cmd[0]);
        break;
    }
    tty_p->tty_cmd[0] = DREQ_FREE_CMDSET;
    return;
}

static void netcmd_handle(TTYCTRL_P tty_p)
{
    int cmd_idx;
    int txmsg_len;
    int fail_rsp;
    int retval;
    unsigned char rsp_cmd;
    unsigned char msg[256];
    char tty_rsp[CF_CMDRSPLEN];
    struct sysinfo sys_info;
    
    if(tty_p->tty_cmd[0] != DREQ_DEVICE_CMDSET)
        return;
    cmd_idx = 1;
        
    L_MORECMD:
    fail_rsp = txmsg_len = 0;
    switch(rsp_cmd = tty_p->tty_cmd[cmd_idx]) {
    case DREQ_CMD_LINECTRL:
	ADV_LOGMSG("# \tDREQ_CMD_LINECTRL MPT[%d]\n", devidx);
        if(tty_p->tty_cmd[cmd_idx + 1] != 2) {
            goto L_WRONGCMD;
        }
        fail_rsp = 2;
        init_netmsg(msg, 0xffffffff, 0x0e, 0,
                    tty_p->tty_cmd[cmd_idx + 2] ? 9 : 10, 0);
        if(net_send(tty_p->skt_fd, msg, 20) != 20)
            break;     
        init_netmsg(msg, DREQ_CMD_LINECTRL, 0x0e, 0,
                    tty_p->tty_cmd[cmd_idx + 3] ? 12 : 13, 0);
        txmsg_len = 20;
        break;
    case DREQ_CMD_FLUSH:
	ADV_LOGMSG("# \tDREQ_CMD_FLUSH MPT[%d]\n", devidx);
        if(tty_p->tty_cmd[cmd_idx + 1] != 1) {
            goto L_WRONGCMD;
        }
        fail_rsp = 2;
        init_netmsg(msg, DREQ_CMD_FLUSH, 0x0e, 4, 19, 0);
        switch(tty_p->tty_cmd[cmd_idx + 2]) {
        case 0:
            *((unsigned int *)&msg[20]) = 0x08;
            tty_p->rd_datalen = tty_p->rd_header = tty_p->rd_tailer = 0;
            break;
        case 1:
            *((unsigned int *)&msg[20]) = 0x04;
            break;
        case 2:
            *((unsigned int *)&msg[20]) = 0x0c;
            tty_p->rd_datalen = tty_p->rd_header = tty_p->rd_tailer = 0;
            break;
        }
        txmsg_len = 24;
        break;
    case DREQ_CMD_SETBAUD:
	ADV_LOGMSG("# \tDREQ_CMD_SETBAUD\n");
        if(tty_p->tty_cmd[cmd_idx + 1] != 4) {
            goto L_WRONGCMD;
        }
        fail_rsp = 2;
        init_netmsg(msg, DREQ_CMD_SETBAUD, 0x0e, 4, 1, 0);
        memcpy(&msg[20], &tty_p->tty_cmd[cmd_idx + 2], 4);
        txmsg_len =24;
        break;
    case DREQ_CMD_START_BREAK:
	ADV_LOGMSG("# \tDREQ_CMD_START_BREAK\n");
        if(tty_p->tty_cmd[cmd_idx + 1] != 0) {
            goto L_WRONGCMD;
        }
        fail_rsp = 2;
        init_netmsg(msg, DREQ_CMD_START_BREAK, 0x0e, 0, 4, 0);
        txmsg_len = 20;
        break;
    case DREQ_CMD_STOP_BREAK:
	ADV_LOGMSG("# \tDREQ_CMD_STOP_BREAK\n");
        if(tty_p->tty_cmd[cmd_idx + 1] != 0) {
            goto L_WRONGCMD;
        }
        fail_rsp = 2;
        init_netmsg(msg, DREQ_CMD_STOP_BREAK, 0x0e, 0, 5, 0);
        txmsg_len = 20;
        break;
        
    case DREQ_CMD_PORT_INIT:
	ADV_LOGMSG("# \tDREQ_CMD_PORT_INIT MPT[%d]\n", devidx);
        if(tty_p->tty_cmd[cmd_idx + 1] != 8) {
            goto L_WRONGCMD;
        }
        if(do_port_init(tty_p, cmd_idx) < 0)
            fail_rsp = 3;
        break;       
    case DREQ_CMD_WAIT_OQUEUE:
	ADV_LOGMSG("# \tDREQ_CMD_WAIT_OQUEUE\n");
        if(tty_p->tty_cmd[cmd_idx + 1] != 4) {
            goto L_WRONGCMD;
        }
        fail_rsp = 4;
        sysinfo(&sys_info);
        tty_p->wait_qstamp = sys_info.uptime;
        memcpy((char *)&tty_p->wait_qtime,
               (char *)&tty_p->tty_cmd[cmd_idx + 2], 4);
        init_netmsg(msg, DREQ_CMD_WAIT_OQUEUE, 0x0e, 0, 27, 0);
        txmsg_len = 20;
        break;

    case DREQ_CMD_XONXOFF:
	ADV_LOGMSG("# \tDREQ_CMD_XONXOFF MPT[%d]\n", devidx);
        fail_rsp = 2;
        if(tty_p->tty_cmd[cmd_idx + 1] != 2) {
            goto L_WRONGCMD;
        }
        init_netmsg(msg, DREQ_CMD_XONXOFF, 0x0e, 6, 23, 0);
        msg[20] = msg[21] = msg[22] = msg[23] = 0;
        msg[24] = tty_p->tty_cmd[3];
        msg[25] = tty_p->tty_cmd[4];
        txmsg_len = 26;
        break;

    case DREQ_CMD_SETXON:
	ADV_LOGMSG("# \tDREQ_CMD_SETXON\n");
        fail_rsp = 2;
        init_netmsg(msg, DREQ_CMD_SETXON, 0x0e, 0, 15, 0);
        txmsg_len = 20;
        break;
        
    case DREQ_CMD_SETXOFF:
	ADV_LOGMSG("# \tDREQ_CMD_SETXOFF\n");
        fail_rsp = 2;
        init_netmsg(msg, DREQ_CMD_SETXON, 0x0e, 0, 14, 0);
        txmsg_len = 20;
        break;
        
    default:
        ADV_LOGMSG("MPT[%d] (netcmd_handle) unknown tty dreq[%x]\n",
                   tty_p->mpt_devidx, tty_p->tty_cmd[1]);
        goto L_WRONGCMD;
    }
    
    if((txmsg_len > 0) &&
       (net_send(tty_p->skt_fd, msg, txmsg_len) == txmsg_len)) {
        fail_rsp = 0;
    }
    switch(fail_rsp) {
    case 1:
        tty_rsp[0] = DREQ_DEVICE_CMDSET;
        tty_rsp[1] = rsp_cmd;
        tty_rsp[2] = 'O';
        tty_rsp[3] = 'K';
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 4), tty_rsp);
        user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 4);  
	break;
    case 2:
        tty_rsp[0] = DREQ_DEVICE_CMDSET;
        tty_rsp[1] = rsp_cmd;
        tty_rsp[2] = 'N';
        tty_rsp[3] = 'G';
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 4), tty_rsp);
        user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 4);        
        break;
    case 3:
        tty_rsp[0] = DREQ_DEVICE_CMDSET;
        tty_rsp[1] = rsp_cmd;
        tty_rsp[2] = 3;
        tty_rsp[3] = 0xff;
        tty_rsp[4] = 0xff;
        tty_rsp[5] = 0xff;
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 6), tty_rsp);
        user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 6);                
	break;
    case 4:
        tty_rsp[0] = DREQ_DEVICE_CMDSET;
        tty_rsp[1] = rsp_cmd;
        tty_rsp[2] = 2;
        tty_rsp[3] = 0;
        tty_rsp[4] = 0;
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 5), tty_rsp);
	user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 5);                
	break;
    }
    if((cmd_idx += 2 + tty_p->tty_cmd[cmd_idx + 1]) < tty_p->tty_cmdlen)
        goto L_MORECMD;
    tty_p->tty_cmd[0] = DREQ_FREE_CMDSET;
    return;
    
    L_WRONGCMD:
    tty_p->tty_cmd[0] = DREQ_FREE_CMDSET;
    return;
}

static void poll_dev(TTYCTRL_P tty_p)
{
    unsigned char msg[256];
    
    if(tty_p->status & STATUS_TXPOLL)
        return;
    
    init_netmsg(msg, 0xfffffffe, 0x0e, 0, 16, 0);
    if(net_send(tty_p->skt_fd, msg, 20) != 20) {
        next_state(tty_p, STATE_NETDOWN);
        return;
    }
    tty_p->status |= STATUS_TXPOLL;
    return;
}

static int do_port_init(TTYCTRL_P tty_p, int cmd_idx)
{
    int i;
    u_int baudrate;
    unsigned char msg[256];
    char tty_rsp[CF_CMDRSPLEN];
 
    if((baudrate = buad_idx2val(tty_p->tty_cmd[cmd_idx + 2]))) {
        init_netmsg(msg, 0xffffffff, 0x0e, 4, 1, 0);
        *((u_int *)&msg[20]) = baudrate;
        if(net_send(tty_p->skt_fd, msg, 24) != 24)
            return -1;
    }
    
    init_netmsg(msg, 0xffffffff, 0x0e, 3, 3, 0);
    msg[20] = tty_p->tty_cmd[cmd_idx + 3] & 0x4 ? 2 : 0;
    switch(tty_p->tty_cmd[cmd_idx + 3] & 0x38) {
    default:
    case 0:
        msg[21] = 0;;
        break;
    case 8:
        msg[21] = 2;
        break;
    case 16:
        msg[21] = 1;
        break;
    case 24:
        msg[21] = 3;
        break;
    case 32:
        msg[21] = 4;
        break;
    }
    msg[22] = (tty_p->tty_cmd[cmd_idx + 3] & 0x3) + 5;
    if(net_send(tty_p->skt_fd, msg, 23) != 23)
        return -1;
        
    init_netmsg(msg, 0xffffffff, 0x0e, 0,
                tty_p->tty_cmd[cmd_idx + 4] ? 9 : 10, 0);
    if(net_send(tty_p->skt_fd, msg, 20) != 20)
        return -1;
        
    init_netmsg(msg, 0xffffffff, 0x0e, 0,
                tty_p->tty_cmd[cmd_idx + 5] ? 12 : 13, 0);
    if(net_send(tty_p->skt_fd, msg, 20) != 20)
        return -1;
       
    init_netmsg(msg, 0xffffffff, 0x0e, 16, 25, 0);
    if((tty_p->tty_cmd[cmd_idx + 6] == 1) || (tty_p->tty_cmd[cmd_idx + 7] == 1)) {
        *((u_int *)&msg[20]) = 0x80000009;
        *((u_int *)&msg[24]) = 0x80000080;
    }else if((tty_p->tty_cmd[cmd_idx + 6] == 2) || (tty_p->tty_cmd[cmd_idx + 7] == 2)){
    	*((u_int *)&msg[20]) = 0x80000012;
        *((u_int *)&msg[24]) = 0x80000040;
    }else {
        *((u_int *)&msg[20]) = 0x80000000;//=0x80000000
        *((u_int *)&msg[24]) = 0x00000000;//=0x00000000
    }
    if(tty_p->tty_cmd[cmd_idx + 8]) {
        *((u_int *)&msg[24]) = 0x01;//=0x01
    }
    if(tty_p->tty_cmd[cmd_idx + 9]) {
        *((u_int *)&msg[24]) = 0x43;//=0x43
    }
    *((u_int *)&msg[28]) = 2048;
    *((u_int *)&msg[32]) = 512;
    if(net_send(tty_p->skt_fd, msg, 36) != 36)
        return -1;
        
    init_netmsg(msg, 0xffffffff, 0x0e, 4, 17, 0);
    *((u_int *)&msg[20]) = 0x01f8;
    if(net_send(tty_p->skt_fd, msg, 24) != 24)
        return -1;
        
    if(!(tty_p->status & STATUS_WOMSENT)) {
        init_netmsg(msg, 0xfffffffe, 0x0e, 0, 18, 0);
        if(net_send(tty_p->skt_fd, msg, 20) != 20)
            return -1;
        tty_p->status |= STATUS_WOMSENT;
    }
        
    init_netmsg(msg, DREQ_CMD_PORT_INIT, 0x0e, 0, 26, 0);
    if(net_send(tty_p->skt_fd, msg, 20) != 20)
        return -1;  
    return 0;
}

static int get_netmsg(TTYCTRL_P tty_p)
{
    unsigned short datalen;
    
    if(net_recv(tty_p->skt_fd, tty_p->net_rxmsg, 20) != 20) {
        next_state(tty_p, STATE_NETDOWN);
	ADV_LOGMSG("# MPT[%d] netdown at get_netmsg \n", devidx); 
        return 0;
    }
    
    datalen = ntohs(*((unsigned short *)&tty_p->net_rxmsg[10]));
    if (datalen > (CF_NETMSGLEN - 20))
    	datalen = CF_NETMSGLEN - 20;
    if((datalen != 0) &&
       (net_recv(tty_p->skt_fd, &tty_p->net_rxmsg[20], datalen) != datalen)) {
        next_state(tty_p, STATE_NETDOWN);
        return 0;
    }
    return tty_p->net_rxlen = 20 + datalen;
}

static void proc_netmsg(TTYCTRL_P tty_p)
{
    int i;
    int rsp_type;
    unsigned short cmd;
    unsigned short datalen;
    unsigned int id;
    unsigned int p1;
    unsigned int p2;
    unsigned int val;
    char tty_rsp[CF_CMDRSPLEN];
    unsigned char rsp_cmd;
    unsigned char msg[256];
    int sendlen;
    unsigned short len;
    unsigned char data[CF_NETMSGLEN];
    
    if(!tty_p->net_rxlen)
    {
	return;
	}
        
    id = *((u_int *)&tty_p->net_rxmsg[4]);
    cmd = ntohs(*((unsigned short *)&tty_p->net_rxmsg[8])) & 0x7fff;
    datalen = ntohs(*((unsigned short *)&tty_p->net_rxmsg[10]));
    p1 = ntohl(*((unsigned int *)&tty_p->net_rxmsg[12]));
    p2 = ntohl(*((unsigned int *)&tty_p->net_rxmsg[16]));
    
                
    if(id == 0xffffffff) {
        goto L_EXIT;
    }
    
    rsp_type = 0;
    switch(cmd) {
    case 0x0e:
        switch(p2) {
        case 26:
            if(id == DREQ_CMD_PORT_INIT) {        
                val = *((unsigned int *)&tty_p->net_rxmsg[20]);
                tty_rsp[0] = DREQ_DEVICE_CMDSET;
                tty_rsp[1] = DREQ_CMD_PORT_INIT;
                tty_rsp[2] = 3;
                tty_rsp[3] = val & 0x20 ? 1 : 0;
                tty_rsp[4] = val & 0x10 ? 1 : 0;
                tty_rsp[5] = val & 0x80 ? 1 : 0;
                //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 6), tty_rsp);
				user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 6);
		
            }
            break;
        case 16:
            tty_p->status &= ~STATUS_TXPOLL;
            break;
        case 18:
            tty_p->status &= ~STATUS_WOMSENT;
            if((curr_state(tty_p) != STATE_READY) || (p1 != 0))
	    {
                break;
		}
            init_netmsg(msg, 0xfffffffe, 0x0e, 0, 18, 0);
            if(net_send(tty_p->skt_fd, msg, 20) == 20)
                tty_p->status |= STATUS_WOMSENT;
            val = *((unsigned int *)&tty_p->net_rxmsg[20]);
            tty_rsp[0] = DREQ_DEVICE_CMDSET;
            tty_rsp[1] = DREQ_NOTIFY;
            if(val & 0x38) {
                tty_rsp[2] = DREQ_NOTIFY_MSR_CHG;
                tty_rsp[3] = 0;
                tty_rsp[3] |= (val & 0x08 ? 0x10 : 0x00);
                tty_rsp[3] |= (val & 0x10 ? 0x20 : 0x00);
                tty_rsp[3] |= (val & 0x20 ? 0x80 : 0x00);
                tty_rsp[4] = 0;
            }
            else if(val & 0x40) {
                tty_rsp[2] = DREQ_NOTIFY_BREAK;
                tty_rsp[3] = tty_rsp[4] = 0;
            }
            else if(val & 0x80) {
                init_netmsg(msg, DREQ_NOTIFY, 0x0e, 0, 27, 0);
                net_send(tty_p->skt_fd, msg, 20);
                break;
            }
            else
	    {
		break;
		}
		user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 5);
            break;
        case 1:
        case 4:
        case 5:
        case 12:
        case 13:
        case 14:
        case 15:
        case 19:
        case 23:
            rsp_cmd = (unsigned char)id;
            rsp_type = p1 ? 2 : 1;
            break;
        case 27:
            switch(id) {
            case DREQ_NOTIFY:
                val = *((unsigned int *)&tty_p->net_rxmsg[20]);
                val &= 0x1e;
                tty_rsp[0] = DREQ_DEVICE_CMDSET;
                tty_rsp[1] = id;
                tty_rsp[3] = tty_rsp[4] = 0;
                for(i = 1; i < 5; i++) {
                    if(val & (1<<i))
                        continue;
                    switch(i) {
                    case 1:
                        tty_rsp[2] = DREQ_NOTIFY_FRAMING;
                        break;
                    case 2:
                        tty_rsp[2] = DREQ_NOTIFY_HW_OVERRUN;
                        break;
                    case 3:
                        tty_rsp[2] = DREQ_NOTIFY_SW_OVERRUN;
                        break;
                    case 4:
                        tty_rsp[2] = DREQ_NOTIFY_SW_OVERRUN;
                        break;
                    }
                    //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 5), tty_rsp);
            	    user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 5);
		    
                }
                break;
            case DREQ_CMD_WAIT_OQUEUE:
                val = *((unsigned int *)&tty_p->net_rxmsg[32]);
                tty_rsp[0] = DREQ_DEVICE_CMDSET;
                tty_rsp[1] = id;
                tty_rsp[2] = 2;
                tty_rsp[3] = val & 0xff;
                tty_rsp[4] = (val>>8) & 0xff;
                //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 5), tty_rsp);
            	user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 5);

                break;
            }
            break;
        
        }
        break;
    case 0x04:
	//ADV_LOGMSG("# MPT[%d] wr_datalen=%x, datalen=%x, p1=%x, p2=%x\n", devidx, tty_p->wr_datalen, datalen, p1, p2); 
        //'p1 = 1' means that edgserver successfully write data to serial device
	if(p1 == 0){
		tty_p->wr_datalen = 0;
		/*
		if(QueueKernelSize(tty_p) <= 0){
			ADV_LOGMSG("MPT[%d] no data in delay buffer\n", devidx);
			sendlen = user_to_kernel((devidx << 8)+NETLINK_KSEND_DATA_UNTHROTTLE, 0, NULL, 0);
			if(!sendlen){
				ADV_LOGMSG("MPT[%d] send delay data : send NETLINK_KSEND_DATA_UNTHROTTLE failed\n", devidx);
			}	
			break;
		}				
		else{
			if(tty_p->delay_header < tty_p->delay_tailer) {
				len = MIN(1024, tty_p->delay_tailer - tty_p->delay_header);
				init_netmsg(data, 0xfffffffe, 0x04, (unsigned short)len, 0, 0);
				memcpy(data+20, &tty_p->delay_data[tty_p->delay_header], len);
	
                		if(net_send(tty_p->skt_fd, data, 20 + len) == 20 + len)
				{
                 	   		tty_p->wr_datalen = len;
				}
				else
				{
					ADV_LOGMSG("MPT[%d] send delay data : net_send not equ receive\n", devidx);
					break;
				}
                		tty_p->delay_header += len;
                	}
            		else {
                		len = MIN(1024, DELAY_LEN - tty_p->delay_header);
				init_netmsg(data, 0xfffffffe, 0x04, (unsigned short)len, 0, 0);
				memcpy(data+20, &tty_p->delay_data[tty_p->delay_header], len);
	
                		if(net_send(tty_p->skt_fd, data, 20 + len) == 20 + len)
				{
                 	   		tty_p->wr_datalen = len;
				}
				else
				{
					ADV_LOGMSG("MPT[%d] send delay data : net_send not equ receive\n", devidx);
					break;
				}
				tty_p->delay_header = (tty_p->delay_header + len) % DELAY_LEN;  
            		}

		}*/
	}
        break;
    case 0x0800: {
        int movedata_len;
        int tmp1;
        int tmp2;
	//ADV_LOGMSG("MPT[%d]:  advttyp receives data from edgserver, length=%d\n", devidx, datalen);                   
        if((movedata_len = datalen) + QueueUsedSize(tty_p) >= CF_DATABUFLEN)
        {
            movedata_len = CF_DATABUFLEN - QueueUsedSize(tty_p) - 1;
            char msg1[500];
           sprintf(msg1, "MPT[%d]: receive datalen=%d, data trunked movedata_len = %d\n", devidx, datalen, movedata_len);
           log_msg(msg1);
         }
        if (movedata_len <= 0)
        {
        	sendQueueFreeSize(tty_p, 0);
        	break;
        }        	
            
        if(tty_p->rd_tailer + movedata_len < CF_DATABUFLEN) {
            memcpy(&tty_p->rd_data[tty_p->rd_tailer],
                   &tty_p->net_rxmsg[20], movedata_len);
            tty_p->rd_tailer += movedata_len;
            
        }
        else if(tty_p->rd_tailer + movedata_len == CF_DATABUFLEN) {
            memcpy(&tty_p->rd_data[tty_p->rd_tailer],
                   &tty_p->net_rxmsg[20], movedata_len);
            tty_p->rd_tailer = 0;
            
        }
        else{
            tmp1 = CF_DATABUFLEN - tty_p->rd_tailer;
            memcpy(&tty_p->rd_data[tty_p->rd_tailer],
                   &tty_p->net_rxmsg[20], tmp1);
            tmp2 = movedata_len - tmp1;
            memcpy(&tty_p->rd_data[0],
                   &tty_p->net_rxmsg[20 + tmp1], tmp2);
            tty_p->rd_tailer = tmp2;
            
        }
        
        
            
        //tty_p->rd_datalen += movedata_len;
        
        sendQueueFreeSize(tty_p, 0);	
        break;}
    }
    
    switch(rsp_type) {
    case 1:
        tty_rsp[0] = DREQ_DEVICE_CMDSET;
        tty_rsp[1] = rsp_cmd;
        tty_rsp[2] = 'O';
        tty_rsp[3] = 'K';
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 4), tty_rsp);
        user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 4);        
	break;
    case 2:
        tty_rsp[0] = DREQ_DEVICE_CMDSET;
        tty_rsp[1] = rsp_cmd;
        tty_rsp[2] = 'N';
        tty_rsp[3] = 'G';
        //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_RESPONSE, 4), tty_rsp);
        user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_RESPONSE, tty_rsp, 4);
        break;
    }
    
    L_EXIT:    
    tty_p->net_rxlen = 0;
    return;
}

static void dev_init(TTYCTRL_P tty_p)
{
    if(tty_p->skt_fd >= 0) {
        close(tty_p->skt_fd);
        tty_p->skt_fd = -1;
    }
    tty_p->status = 0;
	tty_p->wr_datalen = 0;
    tty_p->rd_datalen = tty_p->rd_header = tty_p->rd_tailer = 0;
    tty_p->net_rxlen = 0;

    tty_p->usendok=1;//now can send data to kernel
    tty_p->bufferflags= 0;
    tty_p->detechtime = 0;
    tty_p->isstatus = 0;
    tty_p->delay_cnt =  tty_p->delay_header = tty_p->delay_tailer = 0;
    tty_p->throttle = 0;
    return;
}

static void cleanup(TTYCTRL_P tty_p)
{
    if(tty_p->skt_fd >= 0) {
        if(tty_p->status & STATUS_NWREADY) {
            //ioctl(tty_p->mpt_fd, ADV_IOC(IOCMD_DISCONNECTED, 0), 0);
	    	user_to_kernel((devidx << 8)+NETLINK_DEAL_COMD, IOCMD_DISCONNECTED, NULL, 0);
	   
	    tty_p->status &= ~STATUS_NWREADY;
        }
        close(tty_p->skt_fd);
        tty_p->skt_fd = -1;
    }
    
    if(skfd >= 0) {
        close(skfd);
        skfd = -1;
    }
    return;
}

static void shutdown_handle()
{
    _running = 0;
    return;
}

static int log_open(char * log_name)
{
	return __log_fd = open(log_name,
	                       O_WRONLY | O_CREAT | O_APPEND | O_NDELAY, 0666);
}

static void log_close(void)
{
	if(__log_fd >= 0) {
	    close(__log_fd);
	    __log_fd = -1;
	}
	return;
}

static void log_msg(const char * msg)
{
	time_t t;
	struct tm * lt;
	char msg_buf[1024];
	
	if(__log_fd < 0)
	    return;
	    
	t = time(0);
	lt = localtime(&t);	
    sprintf(msg_buf, "%02d-%02d-%4d %02d:%02d:%02d  ",
		    lt->tm_mon + 1, lt->tm_mday, lt->tm_year + 1900,
			lt->tm_hour, lt->tm_min, lt->tm_sec);
    strcat(msg_buf, msg);
    
    if(flock(__log_fd, LOCK_EX) < 0)
        return;
    write(__log_fd, msg_buf, strlen(msg_buf));
    flock(__log_fd,  LOCK_UN);
    return;
}

static int hexstr(char * strp)
{
    int i, ch, val;
    for(i = val = 0; (ch = *(strp + i)) != '\0'; ++i) {
        if(ch >= '0' && ch <= '9') {
            val = 16 * val + ch - '0';
        }
        else if((ch >= 'a' && ch <= 'f') || (ch >= 'A' && ch <= 'F')) {
            ch = toupper(ch);
            val = 16 * val + ch - 'A' + 10;
        }
        else
            return 0;
    }
    return val;
}

static int net_send(int skt_fd, unsigned char * msg, int len)
{
    int txret;
    int txidx;
    //ADV_LOGMSG("net_send MPT[%d]", devidx);
    if(skt_fd < 0)
        return skt_fd;
    for(txidx = 0; txidx < len; ) {
        if((txret = send(skt_fd, msg + txidx, len - txidx, 0)) <= 0)
            return txret;
        txidx += txret;
    }
    //ADV_LOGMSG(" send : %d\n", txidx);
    return txidx;
}

static int net_recv(int skt_fd, unsigned char * msg, int len)
{
    int rxret;
    int rxidx;
    //ADV_LOGMSG("net_recv MPT[%d]", devidx);    
    if(skt_fd < 0)
        return skt_fd;
    for(rxidx = 0; rxidx < len; ) {
        if((rxret = recv(skt_fd, msg + rxidx, len - rxidx, 0)) <= 0)
            return rxret;
        rxidx += rxret;
    }
    //ADV_LOGMSG(" recev : %d\n", rxidx);
    return rxidx;
}

static __inline u_int buad_idx2val(int baudidx)
{
    u_int baudval[] = {300, 600, 1200, 2400,
                        4800, 7200, 9600, 19200,
                        38400, 57600, 115200, 230400,
                        460800, 921600, 150, 134,
                        110, 75, 50};
                        
    if((baudidx < 0) ||
       (baudidx > sizeof(baudval) / sizeof(u_int) - 1))
       return 0;
    return(baudval[baudidx]);
}

static __inline void init_netmsg(unsigned char * msg, unsigned int tid,
                                 unsigned short cmd, unsigned short len,
                                 unsigned int p1, unsigned int p2)
{
    msg[0] = (unsigned char)0x4d;
    msg[1] = (unsigned char)0x41;
    msg[2] = (unsigned char)0x44;
    msg[3] = (unsigned char)0x41;
    *((unsigned int *)&msg[4]) = tid;
    *((unsigned short *)&msg[8]) = htons(cmd);
    *((unsigned short *)&msg[10]) = htons(len);
    *((unsigned int *)&msg[12]) = htonl(p1);
    *((unsigned int *)&msg[16]) = htonl(p2);
    return;
}

static __inline int curr_state(TTYCTRL_P tty_p)
{
    return tty_p->curr_state;
}

static __inline int next_state(TTYCTRL_P tty_p, int next_state)
{
    return tty_p->curr_state = next_state;
}
